unit module_implementations;

{

   COPYRIGHT 2013 .. 2015 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Classes, System.Math, System.UITypes, Vcl.Dialogs,

  Globals, module_defs, module_matrix, KnobsUtils, KnobsConversions, midi_reader, windowing,

  WaveUtils, WaveStorage, scala, {$ifdef win32} freereverb, {$endif} fftwgen, fftw_, Talkie;

const

  DELAY_BUF_SIZE  = 65536;   // main delay buffer size - 65536 samples on 44100 Hz means 1.49 s.
  MAX_XTAB        = 20480;   // crossfade lookup table size for pitch change
  DEF_FFT_SIZE    = 1024;    // Default bin size for FFT modules

type

  TARState         = ( arIdle, arA, arR, arWait);
  TAHDState        = ( ahdIdle, ahdA, ahdH, ahdD, ahdWait);
  TADSRState       = ( adsrIdle, adsrA, adsrD, adsrS, adsrR, adsrWait);
  TEnvRange        = ( erSlow, erMedium, erFast);
  TEnvShape        = ( esLog, esLin, esExp);
  TLfoRange        = ( lsSlow, lsMedium, lsFast, lsBPM, lsLin);
  TDelaySize       = ( dsShort, dsMedium, dsLong);
  TRWState         = ( trwIdle, trwFired);
  TTrigOScState    = ( toIdle, toRunning, toWaiting, toDone);
  TSeqInState      = ( seqClkLow, seqClkHigh);
  TSeqStepState    = ( sssIdle, sssActive, sssWaiting);
  TPulsesState     = ( psIdle, psPulsing, psWaiting);
  TMidiPlayerState = ( msActive, msDone);

  TLongWord = record
    Lo, Hi : Word;
  end;

  TFadeTable = array[ 0 .. MAX_XTAB - 1] of TSignal;
  PFadeTable = ^TFadeTable;
  TEuclids   = set of Byte;


  TModConstant = class( TMod)
  private
    FMode : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    Mode: Integer read FMode write FMode;
  end;

  TModPad = class( TMod)
  private
    FMode : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    Mode: Integer read FMode write FMode;
  end;

  TModInverter = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModAdder = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMultiplier = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMultiMult = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModNoteToDelay = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModZeroCounter = class( TMod)
  private
    FOldSign : Boolean;
    FTime    : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMedian = class( TMod)
  private
    FData      : TMedianCalculator;
    FOldTrig   : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModCompare = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModGate = class( TMod)
  private
    FFunction : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModDivider = class( TMod)
  private
    FLastInput : Boolean;
    FCounter   : Integer;
    FOldCount  : Integer;
    FState     : Boolean;
    FOldReset  : Boolean;
  private
    procedure   PerformReset;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDFlipFlop = class( TMod)
  private
    FClockState : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModRSFlipFlop = class( TMod)
  private
    FSetState   : Boolean;
    FResetState : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModCounter = class( TMod)
  private
    FState   : Word;
    FMode    : Integer;
    FOldRes  : Boolean;
    FOldTrig : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModAdc = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModDac = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TPDState = 1 .. 12;

  TModPhaseDetect = class( TMod)
  private
    FOldIn1 : Boolean;
    FOldIn2 : Boolean;
    FState  : TPDState;
    FAlpha  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModNot = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModAmplifier = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModScaler = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRatio = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRotator = class( TMod)
  private
    FAngle: TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModRectifier = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModMinMax = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModIntDif = class( TMod)
  private
    FSum        : TSignal;
    FlastSample : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModBaseOsc = class( TMod)
  private
    FPhase      : TSignal;
    FPhaseDelta : TSignal;
    FPosition   : TSignal;
    FOldSync    : Boolean;
    FShape      : Integer;
    FBandLimit  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModBaseTrigOsc = class( TModBaseOsc)
  private
    FState   : TTrigOScState;
    FCounter : Integer;
  public
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModOsc = class( TModBaseOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModTrigOsc = class( TModBaseTrigOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModSquare = class( TModBaseOsc)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSquareTrig = class( TModBaseTrigOsc)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModBaseLfo = class( TMod)
  private
    FPhase    : TSignal;
    FPosition : TSignal;
    FOldRes   : Boolean;
    FShape    : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModBaseTrigLfo = class( TModBaseLfo)
  private
    FState   : TTrigOScState;
    FCounter : Integer;
    FDidRun  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
  end;

  TModCounterLfo = class( TMod)
  private
    FPhase : TSignal;
  protected
    procedure   Wrapped;                                                                          virtual;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
  end;

  TModLfo = class( TModBaseLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModLfoTrig = class( TModBaseTrigLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModSquareLfo = class( TModBaseLfo)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSquareLfoTrig = class( TModBaseTrigLfo)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModRandSig = class( TModCounterLfo)
  private
    FStart  : TSignal;
    FTarget : TSignal;
  protected
    procedure   Wrapped;                                                                         override;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPulses = class( TMod)
  private
    FPhase         : TSignal;
    FState         : TPulsesState;
    FPulseCounter  : Integer;
    FPeriodCounter : Integer;
    FLedCounter    : Integer;
    FOldFrequency  : TSignal;
    FOldLambda     : TSignal;
    FOldDist       : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModRandomWalkLfo = class( TModBaseLfo)
  private
    FState : TRWState;
    FX     : TSignal;
    FY     : TSignal;
    FZ     : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoise = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseTrig = class( TMod)
  private
    FLastTrig: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseLfo = class( TMod)
  private
    FOutValue : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseLfoTrig = class( TMod)
  private
    FLastTrig : Boolean;
    FOutValue : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModAverage = class( TMod)
  private
    FAlpha : TSignal;  // Alpha = System_Rate / ( 2 * PI * Freq + System_Rate) ... Fc := (log(b1))/(-2.0 * M_PI) * Fs; {b1 < 0.0432139182637722579772}
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSVF = class( TMod)
  private
    FFreq  : TSignal;
    FDamp  : TSignal;
    FDrive : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPinkFilter = class( TMod)
  private
    FB0 : TSignal;
    FB1 : TSignal;
    FB2 : TSignal;
    FB3 : TSignal;
    FB4 : TSignal;
    FB5 : TSignal;
    FB6 : TSignal;
    FG  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModResonator = class( TMod)
  // http://www.katjaas.nl/complexintegrator/complexresonator.html
  private
    FOldRe : TSignal;
    FOldIm : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModFreqShifter = class( TMod)
  // https://github.com/swh/ladspa/blob/master/bode_shifter_1431.xml
  private
    FDelay : array[ 0 .. 255] of TSignal;
    FDPtr  : Integer;
    FPhase : TSignal;
    FDelta : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvAR = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvARRetrig = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FArmed        : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvAHD = class( TMod)
  private
    FData      : TSignalArray;
    FState     : TAHDState;
    FTime      : Integer;
    FDuration  : Integer;
    FArmed     : Boolean;
    FLastLevel : TSignal;
    FValue     : TSignal;
    FEnvValue  : TSignal;
    FTrigFlag  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvADSR = class( TMod)
  private
    FData      : TSignalArray;
    FState     : TADSRState;
    FTime      : Integer;
    FDuration  : Integer;
    FArmed     : Boolean;
    FLastLevel : TSignal;
    FValue     : TSignal;
    FEnvValue  : TSignal;
    FTrigFlag  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FSteps    : Integer;
    FCounter  : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModSeqSeq = class( TMod)
  private
    FState           : Word;
    FSteps           : Integer;
    FSequenceCounter : Integer;
    FPeriodCounter   : Integer;
    FOldRes          : Boolean;
    FOldTrig         : Boolean;
    FClkCnt          : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModSeq16 = class( TMod)
  private
    FOutValue   : TSignal;
    FReset      : Boolean;
    FNextValue  : TSignal;
    FForwards   : Boolean;
    FActive     : Boolean;
    FChainValue : TSignal;
    FSteps      : Integer;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FOldRes     : Boolean;
    FOldActive  : Boolean;
    FTicks      : Integer;
    FPeriodTime : Integer;
  private
    procedure   SampleInputs;
    procedure   MakeNextValue;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModTextSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FSteps    : Integer;
    FCounter  : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    procedure   SetStepValues( const aValues: TSignals);
  end;

  TModSeqStep = class( TMod)
  private
    FTime    : TSignal;
    FState   : TSeqStepState;
    FOldRes  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModSeqClockStep = class( TMod)
  private
    FCount      : Integer;
    FState      : TSeqStepState;
    FOldRes     : Boolean;
    FOldClock   : Boolean;
    FTrigInFwd  : Boolean;
    FTrigInBwd  : Boolean;
    FDir        : Boolean;
    FTrigOutFwd : Boolean;
    FTrigOutBwd : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModCsvData = class( TMod)
  private
    FMemory   : Array[ 0 .. 15] of TSignals;
    FComments : TStringList;
    FFileName : string;
    FChannels : Integer;
    FCounter  : Integer;
    FSteps    : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
    FMode     : Integer;
    FStopped  : Boolean;
  private
    procedure   LoadFile;
    procedure   SetFileName( const aValue: string);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   DoTick;                                                                          override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;

  TModKleeBit = class( TMod)
  private
    FPattern    : Byte;
    FOldStep    : Boolean;
    FOldLoad    : Boolean;
    FBus1       : Boolean;
    FBus2       : Boolean;
    FBus3       : Boolean;
    FAnalog     : TSignal;
    FLoopMode   : Boolean;
    FInvertMode : Boolean;
  private
    procedure   LoadPreset;
    procedure   Step;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModSeqRandom = class( TMod)
  private
    FValues       : array[ Byte] of TMidiSeptet;
    FPattern      : array[ Byte] of TMidiSeptet;
    FCount        : Integer;
    FScale        : Integer;
    FBaseNote     : TBaseNote;
    FLowNote      : TMidiSeptet;
    FHighNote     : TMidiSeptet;
    FNoteCount    : Integer;
    FScaleNotes   : array[ TMidiSeptet] of TMidiSeptet;
    FOldTrig      : Boolean;
    FOldOne       : Boolean;
    FPosition     : Byte;
    FDistribution : TSignal;
  private
    function    Transpose( aScale: TScale; anAmount: Integer): TScale;
    function    Complete ( aScale: TScale): TScale;
    procedure   MakeScale;
    function    RandomNote( aPosition: Byte; aDistribution: TSignal): TMidiSeptet;
    procedure   ChangeScale( aScale: Integer);
    procedure   ChangeOneNote( aPosition: Byte);
    procedure   ChangeAllNotes;
    procedure   Step;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModMorse = class( TMod)
  private
    FMessage     : string;
    FSymbols     : string;
    FSymbolIndex : Integer;
    FOldReset    : Boolean;
    FOldClock    : Boolean;
  private
    function    CharToSymbols( aChar: Char): string;
    procedure   MessageToSymbols;
    procedure   SetMessage( const aValue: string);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  public
    property    Message: string read FMessage write SetMessage;
  end;

  TModGateSeq = class( TMod)
  private
    FOutValue   : TSignal;
    FSteps      : Integer;
    FOldRes     : Boolean;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FForwards   : Boolean;
    FReset      : Boolean;
  private
    procedure   SampleInputs;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModMemory = class( TMod)
  private
    FMemoryL : TSignals;
    FMemoryR : TSignals;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    function    Convolve( aSample: TSignal): TSignalPair;
  end;

  TModDelay = class( TModMemory)
  private
    FMemLength : TSignal;
    FInPointer : Integer;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModClockedDelay = class( TModMemory)
  private
    FMemLength : TSignal;
    FInPointer : Integer;
    FOldClock  : Boolean;
  public
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModWave = class( TModMemory)
  private
    FFileName    : string;
    FDuration    : string;
    FIsStereo    : Boolean;
  private
    procedure   SetDuration( aValue: Cardinal);
    procedure   SwapMemoryFor( aMemory: TSignalArray);
    procedure   LoadFile;
    procedure   SetFileName( const aValue: string);
  public
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;

  TModWavePlayer = class( TModWave)
  private
    FPlayPointer : TSignal;
    FOldRes      : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;


  TModSampleAndHold = class( TMod)
  private
    FLastTrig: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMultiSAndH = class( TMod)
  private
    FLastTrig: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMux = class( TMod)
  private
    FState : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDemux = class( TMod)
  private
    FState : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModTrackAndHold = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSwitch2to1 = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModXFade = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPan = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixM4to1 = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixS2to1 = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixT1to1 = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixM6to6 = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMixAudio = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModQuantize = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModHardClip = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModTransInv = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModScala = class( TMod)
  private
    FMappingFileName : string;
    FScaleFileName   : string;
    FScale           : TScalaScale;
    FLastNote        : Integer;
    FlastOut         : TSignal;
  private
    procedure   SetMappingFileName( const aValue: string);
    procedure   SetScaleFileName  ( const aValue: string);
    function    Lookup( aValue: TSignal): TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    MappingFileName : string read FMappingFileName write SetMappingFileName;
    property    ScaleFileName   : string read FScaleFileName   write SetScaleFileName;
  end;

  TModTerraGen = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPluck = class( TMod)
  private
    FMemory      : TSignals;
    FMemLength   : TSignal;
    FInPointer   : Integer;
    FOutValue    : TSignal;
    FOldFeedback : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModDisplay = class( TMod)
  private
    FMax1: TSignal;
    FMax2: TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModStateSpace = class( TMod)
  private
    FA        : TSignalMatrix;
    FB        : TSignalMatrix;
    FC        : TSignalMatrix;
    FD        : TSignalMatrix;
    FX        : TSignalMatrix;
    FNewX     : TSignalMatrix;
    FOut      : TSignalMatrix;
    FIn       : TSignalMatrix;
    FRunnable : Boolean;
  private
    function    Consistent: Boolean;
    procedure   ClearState;
  protected
    procedure   Setup;                                                                  virtual; abstract;
    procedure   Run;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMidiIn = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
  end;

  TModMidiNoteIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FLastNote    : Byte;
    FLastChannel : Byte;
    FOldChannel  : Byte;
    FGate        : Boolean;
    FNote        : TSignal;
    FVelocityOn  : TSignal;
    FVelocityOff : TSignal;
  public
    procedure   NotesOff;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                           override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMidiCCIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FCC          : Byte;
    FGate        : Boolean;
    FValue       : TSignal;
    FCounter     : Integer;
    FLedCounter  : Integer;
  public
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                           override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModMidiOut = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
  end;

  TModMidiNoteOut = class( TModMidiOut)
  private
    FLastNote    : Byte;
    FLastChannel : Byte;
    FGate        : Boolean;
    FDidSend     : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
  end;

  TModMidiCCOut = class( TModMidiOut)
  private
    FGate  : Boolean;
  public
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
  end;

  TModMidiPlayer = class( TMod)
  private
    FState           : TMidiPlayerState;
    FReader          : TWrenMidiReader;
    FEvent           : TWrenMidiTimedEvent;
    FFileName        : string;
    FReading         : Boolean;
    FCounter         : Integer;
    FChannel         : Byte;
    FOldTrig         : Boolean;
    FOldReset        : Boolean;
    FGateCount       : Integer;
    FNote            : TSignal;
    FVelocityOn      : TSignal;
    FVelocityOff     : TSignal;
    FDoneCounter     : Integer;
    FDoneLedCounter  : Integer;
    FResetLedCounter : Integer;
    FChOffset        : Integer;
  private
    procedure   SetFileName( const aValue: string);
    function    GetNextEvent: TWrenMidiTimedEvent;
  public
    procedure   NotesOff;
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   DoSlowTick;                                                                      override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;

  {$ifdef win32}
  TModReverb = class( TMod)
  // http://members.chello.nl/t.koning8/kolindex.htm
  private
    FReverb : TReverb;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;
  {$endif}

  TModEnvelope = class( TMod)
  // http://www.musicdsp.org/showone.php?id=97
  // Modified to have an inverted output
  private
    FValue     : TSignal;
    FAttack    : TSignal;
    FRelease   : TSignal;
    FPeakValue : TSignal;
    FPeakDecay : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModCompressor = class( TMod)
  private
    const
      COMP_MAX_GAIN =  64.0;
      COMP_MIN_GAIN = -32.0;
  private
    FValue       : TSignal;
    FAttack      : TSignal;
    FRelease     : TSignal;
    FRefLevel    : TSignal;
    FPreGain     : TSignal;
    FGain        : TSignal;
    FBypass      : Integer;
    FOutL        : TSignal;
    FOutR        : TSignal;
    FLoopAmp     : TSignal;
    FMaxLoopAmp  : TSignal;
    FOldAttack   : TSignal;
    FOldRelease  : TSignal;
    FMinDbVal    : TSignal;
    FMaxDbVal    : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModFormant = class( TMod)
  // http://www.musicdsp.org/showone.php?id=110
  private
    FCoeffStore    : array[ 0 .. 10] of TSignals;
    FCoeffiecients : array[ 0 .. 10] of TSignal;
    FMemory        : TSignalArray;
    FVowel         : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModFlangeChorus = class( TMod)
  // This is a copy from some archaic older FX code I had (originally Turbo Pascal era)
  private
    FDryMix   : TSignal;
    FWetMix   : TSignal;
    FFeedBack : TSignal;
    FRate     : TSignal;
    FSweep    : TSignal;
    FDelay    : TSignal;
    FFrozen   : Boolean;
  private
    fp        : Integer;
    ep1       : Integer;
    ep2       : Integer;
    step      : Integer;
    depth     : Integer;
    idelay    : Integer;
    min_sweep : Integer;
    max_sweep : Integer;
    wlSweep   : TLongWord;
    Buf       : array[ 0 .. DELAY_BUF_SIZE] of TSignal;
  private
    procedure   Setup;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;


  TModPitchChange = class( TMod)
  // This is a copy from some archaic older FX code I had (originally Turbo Pascal era)
  private
    FDryMix   : TSignal;
    FWetMix   : TSignal;
    FFeedback : TSignal;
    FRate     : TSignal;
    FDepth    : TSignal;
    FDelay    : TSignal;
    FFrozen   : Boolean;
  private
    fp         : Integer;
    ep1        : Integer;
    ep2        : Integer;
    ep3        : Integer;
    ep4        : Integer;
    idepth     : Integer;
    idelay     : Integer;
    min_sweep  : Integer;
    max_sweep  : Integer;
    step       : Word;
    xfade      : Integer;
    xfade_cnt  : Integer;
    active     : Integer;
    active_cnt : Integer;
    sweep_up   : Boolean;
    chanA      : Boolean;
    blendA     : TSignal;
    blendB     : TSignal;
    fadeA      : PFadeTable;
    fadeB      : PFadeTable;
    wlSweep    : TLongWord;
    Buf        : array[ 0 .. DELAY_BUF_SIZE] of TSignal;
    Fade_Out   : TFadeTable;
    Fade_In    : TFadeTable;
  private
    procedure   Setup;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   SampleRateChanged;                                                               override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEuclids = class( TMod)
  private
    FPattern  : TEuclids;
    FSteps    : Integer;
    FPulses   : Integer;
    FOldTrig  : Boolean;
    FPosition : Integer;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModXYScope = class( TMod)
  private
    FHistory   : Integer;
    FDataQueue : TSignalFifo;
  private
    procedure   SetHistory( aValue: Integer);
    procedure   AddPoint( anX, anY: TSignal);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    History: Integer read FHistory write SetHistory;
  end;

  TFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FIndex        : Integer;
    FFTWInBuffer  : Pfftw_real;
    FFTWOutBuffer : Pfftw_complex;
    FFTWPlan      : fftw_plan;
  public
    procedure   Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    procedure   Finalize;
    procedure   Tick( anInValue: TSignal; var anOutValue: TComplexSignal);
  end;
  TFFTDataArray = array of TFFTData;

  TFFTHandler = class
  private
    FDestroying : Boolean;
    FBinCount   : Integer;
    FWindow     : TFFTWindow;
    FData       : TFFTDataArray;
  private
    function    GetWindowCount: Integer;
  public
    constructor Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    destructor  Destroy;                                                                         override;
    procedure   Tick( anInValue: TSignal; var anOutValue: TComplexSignal);
  public
    property    WindowCount: Integer read GetWindowCount;
  end;

  TIFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FIndex        : Integer;
    FFTWInBuffer  : Pfftw_complex;
    FFTWOutBuffer : Pfftw_real;
    FFTWPlan      : fftw_plan;
  public
    procedure   Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    procedure   Finalize;
    procedure   Tick( anInValue: TComplexSignal; var anOutValue: TSignal);
  end;
  TIFFTDataArray = array of TIFFTData;

  TIFFTHandler = class
  private
    FDestroying : Boolean;
    FBinCount   : Integer;
    FWindow     : TFFTWindow;
    FData       : TIFFTDataArray;
  private
    function    GetCount: Integer;
  public
    constructor Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    destructor  Destroy;                                                                         override;
    procedure   Tick( anInValue: TComplexSignal; var anOutValue: TSignal);
  public
    property    Count: Integer read GetCount;
  end;

  TModBaseFFT = class( TMod)
  private
    FBinCount    : Integer;
    FWindowCount : Integer;
    FWindowType  : TFFTWindowClass;
    FHandler     : TFFTHandler;
  private
    procedure   SetNewHandler( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    procedure   SetBinCount( aValue: Integer);
  public
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    BinCount : Integer read FBinCount   write SetBinCount;
  end;

  TModBaseIFFT = class( TMod)
  private
    FBinCount    : Integer;
    FWindowCount : Integer;
    FWindowType  : TFFTWindowClass;
    FHandler     : TIFFTHandler;
  private
    procedure   SetNewHandler( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    procedure   SetBinCount( aValue: Integer);
  public
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  end;

  TModBaseFFTIFFT = class( TMod)
  private
    FBinCount    : Integer;
    FWindowCount : Integer;
    FWindowType  : TFFTWindowClass;
    FInHandler   : TFFTHandler;
    FOutHandler  : TIFFTHandler;
  private
    procedure   SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    procedure   SetBinCount( aValue: Integer);
  public
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    function    InternalName( anIndex: Integer): string;                                         override;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  end;

  TModFFT = class( TModBaseFFT)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModIFFT = class( TModBaseIFFT)
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModVocoder = class( TMod)
  private
    FBinCount         : Integer;
    FWindowCount      : Integer;
    FModulatorHandler : TFFTHandler;
    FCarrierHandler   : TFFTHandler;
    FOutHandler       : TIFFTHandler;
  private
    procedure   SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    procedure   SetBinCount( aValue: Integer);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  end;

  TModFilterBank = class( TMod)
  private
    FBinCount     : Integer;
    FWindowCount  : Integer;
    FInHandler    : TFFTHandler;
    FOutHandler   : TIFFTHandler;
    FCoefficients : TSignalArray;
    FCurrCoeff    : Integer;
    FFrequencies  : TSignalArray;
  private
    procedure   SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    procedure   SetBinCount( aValue: Integer);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  end;

  TModTalkie = class( TMod)
  private
    FTalkie     : TTalkie;
    FDestroying : Boolean;
    FStart      : Boolean;
    FOldBankSel : TSignal;
    FBank       : TSignal;
    FPhrase     : TSignal;
    FSampleRate : TSignal;
    FFrameRate  : TSignal;
    FSpeaking   : Boolean;
    FDidRun     : Boolean;
  private
    procedure   TalkStarted( aSender: TObject);
    procedure   TalkDone   ( aSender: TObject);
    procedure   Say;
    procedure   NewFramRate  ( aValue: TSignal);
    procedure   NewSampleRate( aValue: TSignal);
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                              override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;



  function  SignalToEnvRange ( aValue: TSignal): TEnvRange;
  function  SignalToEnvShape ( aValue: TSignal): TEnvShape;
  function  SignalToLfoRange ( aValue: TSignal): TLfoRange;
  function  SignalToDelaySize( aValue: TSignal): TDelaySize;


implementation

var

  IssuedFftwWarning : Boolean = False;

  procedure FftwWarning( const aModuleType: string);
  begin
    if not IssuedFftwWarning
    then begin
      IssuedFftwWarning := True;
      MessageDlg(
        Format(
          'Could not initialize the %s module, the required FFTW DLL may not be present or can not be found.'^M^M +
          'This is a one time warning for all FFT based module types, they will all not work.'^M^M +
          'Please check the debug log (F5) after closing this window',
          [ aModuleType],
          AppLocale
        ),
        mtWarning,
        [ mbOK],
        0
      );
    end;
  end;


  function SignalToEnvRange( aValue: TSignal): TEnvRange;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := erFast
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := erMedium
    else Result := erSlow;
  end;

  function SignalToEnvShape( aValue: TSignal): TEnvShape;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := esLog
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := esLin
    else Result := esExp;
  end;

  function SignalToLfoRange( aValue: TSignal): TLfoRange;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := lsFast
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := lsMedium
    else if AlmostEqual( aValue, 2, 0.45)
    then Result := lsSlow
    else if AlmostEqual( aValue, 3, 0.45)
    then Result := lsBPM
    else Result := lsLin;
  end;

  function SignalToDelaySize( aValue: TSignal): TDelaySize;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := dsShort
    else if AlmostEqual( aValue, 1, 0.45)
    then Result := dsMedium
    else Result := dsLong;
  end;

  function EuclideanRhythm( aPulses, aSteps: Integer): TEuclids;
  //
  // Idea from the wiki Bresenham and Eclidean RHythm pages,
  // Changed to let the pattern start with a one always,
  // the original is:
  //
  // function EuclideanRhythm( aPulses, aSteps: Integer): string;
  // var
  //   Error : Single;
  //   Slope : Single;
  //   i     : Integer;
  // begin
  //   Result := '';
  //   if aSteps > 0
  //   then begin
  //     Error := 0;
  //     Slope := aPulses / aSteps;
  //     for i := 0 to aSteps - 1
  //     do begin
  //       Error := Error + Slope;
  //       if Error >= 0.5
  //       then begin
  //         Result := Result + '1';
  //         Error  := Error  -  1;
  //       end
  //       else Result := Result + '0';
  //     end;
  //   end;
  // end;
  //
  // Then the integer version became:
  //
  // function EuclideanRhythm2( aPulses, aSteps: Integer): string;
  // var
  //   Error : Integer;
  //   i     : Integer;
  // begin
  //   Result := '';
  //   if aSteps > 0
  //   then begin
  //     Error := 0;
  //     for i := 0 to aSteps - 1
  //     do begin
  //       Error := Error + aPulses;
  //       if Error >= aSteps div 2
  //       then begin
  //         Result := Result + '1';
  //         Error  := Error - aSteps;
  //       end
  //       else Result := Result + '0';
  //     end;
  //   end;
  // end;
  //
  // Then finally the modification to make the pattern start with a one
  // always (and uses a result set of Byte instead of a string to avoid memory allocations):
  //
  //
  //
  // done : test a more compact one from : http://electro-music.com/forum/viewtopic.php?p=406480#406480
  // also see : http://www.pdpatchrepo.info/hurleur/euclid.pd (do not test for now)
  //
  /// * generates euclideian patterns
  //    c: current step number
  //    k: hits per bar
  //    n: bar length
  //    r: rotation
  //    returns 1 or 0 according to the euclidean pattern*/
  //
  // fun int euclide( int c, int k, int n, int r ) {
  //   return (((c + r) * k) % n) < k;
  //}
  //
  // 0 => int i;
  // while(i < 12) {
  //   euclide(i, 7, 12, 4) => int e;
  //   //declaring b just to print e
  //   <<<e>>> => int b;
  //   1 +=> i;
  //   100::ms => now;
  //}
  var
    Error : Integer;
    i     : Integer;
  begin
    Result := [];
    if aSteps > 0
    then begin
      Error  := 0;
      for i := 0 to aSteps - 1
      do begin
        Error := Error + aPulses;
        if Error > 0
        then begin
          Result := Result + [ i];
          Error  := Error - aSteps;
        end;
      end;
    end;
  end;


{ ========
  TModConstant = class( TMod)
  public
}

const

  constant_value      = 0;
  constant_mute       = 1;
  constant_mode       = 2;

  constant_out        = 0;

  constant_mode_lin44 = 0;
  constant_mode_db4   = 1;
  constant_mode_lin11 = 2;
  constant_mode_db0   = 3;
  constant_mode_lin04 = 4;
  constant_mode_lin01 = 5;
  constant_mode_note  = 6;


    constructor TModConstant.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModConstant.SetDefaults; // override;
    begin
      FDezipperMap := [
        constant_value,
        constant_mute
      ];
    end;

    procedure   TModConstant.DoSlowTick; // override;
    begin
      if SignalToInt( FInputs[ constant_mode]) = constant_mode_note
      then FOutputs[ constant_out] := ( FInputs[ constant_value]) - NoteNumberToUnits( MiddleNote)
      else FOutputs[ constant_out] := FInputs[ constant_value] * SignalToMute( FInputs[ constant_mute]);
    end;

    function    TModConstant.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        constant_value :  Result := 'value';
        constant_mute  :  Result := 'mute';
        constant_mode  :  Result := 'mode';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModConstant.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        constant_out :  Result := 'out';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPad = class( TMod)
  public
}

const

  pad_padx       = 0;
  pad_pady       = 1;
  pad_outputtype = 2;

  pad_x          = 0;
  pad_y          = 1;


    constructor TModPad.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModPad.SetDefaults; // override;
    begin
      FDezipperMap := [
        pad_padx,
        pad_pady
      ];
    end;

    procedure   TModPad.DoSlowTick; // override;
    var
      aType : Integer;
    begin
      aType := SignalToInt( FInputs[ pad_outputtype]);
      FOutputs[ pad_x] := SignalForOutputType( FInputs[ pad_padx], aType);
      FOutputs[ pad_y] := SignalForOutputType( FInputs[ pad_pady], aType);
    end;

    function    TModPad.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pad_padx       :  Result := 'padx';
        pad_pady       :  Result := 'pady';
        pad_outputtype :  Result := 'outputtype';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModPad.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pad_x :  Result := 'x';
        pad_y :  Result := 'y';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModInverter = class( TMod)
  public
}

const

  inverter_in1  = 0; // Keep these consecutive
  inverter_in2  = 1;
  inverter_in3  = 2;
  inverter_in4  = 3;

  inverter_out1 = 0; // Keep these consecutive
  inverter_out2 = 1;
  inverter_out3 = 2;
  inverter_out4 = 3;

    constructor TModInverter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModInverter.DoTick; // override;
    var
      i : Integer;
    begin
      for i := inverter_in1 to inverter_in4
      do FOutputs[ i - inverter_in1 + inverter_out1] := - FInputs[ i];
    end;

    function    TModInverter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        inverter_in1 : Result := 'in1';
        inverter_in2 : Result := 'in2';
        inverter_in3 : Result := 'in3';
        inverter_in4 : Result := 'in4';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModInverter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        inverter_out1 : Result := 'out1';
        inverter_out2 : Result := 'out2';
        inverter_out3 : Result := 'out3';
        inverter_out4 : Result := 'out4';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAdder = class( TMod)
  public
}

const

  adder_mute = 0;
  adder_in1  = 1;
  adder_in2  = 2;
  adder_in3  = 3;
  adder_in4  = 4;

  adder_out  = 0;

    constructor TModAdder.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModAdder.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := adder_in1 to adder_in4          // Make sure unused inputs are set to 0, so it'll 'work'
      do FInputs[ i] := 0;
      FDezipperMap := [
        adder_mute
      ];
    end;

    procedure   TModAdder.DoTick; // override;
    var
      i : Integer;
      R : TSignal;
    begin
      R := 0.0;
      for i := adder_in1 to adder_in4
      do R := R + FInputs[ i];
      FOutputs[ adder_out] := R * SignalToMute( FInputs[ adder_mute]);
    end;

    function    TModAdder.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        adder_mute : Result := 'mute';
        adder_in1  : Result := 'in1';
        adder_in2  : Result := 'in2';
        adder_in3  : Result := 'in3';
        adder_in4  : Result := 'in4';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModAdder.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        adder_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMultiplier = class( TMod)
  public
}

const

  multiplier_mute = 0;
  multiplier_in1  = 1;
  multiplier_in2  = 2;
  multiplier_in3  = 3;
  multiplier_in4  = 4;

  multiplier_out  = 0;

    constructor TModMultiplier.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMultiplier.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := multiplier_in1 to multiplier_in4          // Make sure unused inputs are set to 1, so it'll 'work'
      do FInputs[ i] := 1;
      FDezipperMap := [
        multiplier_mute
      ];
    end;

    procedure   TModMultiplier.DoTick; // override;
    var
      i : Integer;
      R : TSignal;
    begin
      R := 1.0;
      for i := multiplier_in1 to multiplier_in4
      do R := R * FInputs[ i];
      FOutputs[ multiplier_out] := R * SignalToMute( FInputs[ multiplier_mute]);
    end;

    function    TModMultiplier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        multiplier_mute : Result := 'mute';
        multiplier_in1  : Result := 'in1';
        multiplier_in2  : Result := 'in2';
        multiplier_in3  : Result := 'in3';
        multiplier_in4  : Result := 'in4';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModMultiplier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        multiplier_out : Result := 'out';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMultiMult = class( TMod)
  public
}

const

  multimult_mult = 0;
  multimult_in1  = 1;
  multimult_in2  = 2;
  multimult_in3  = 3;
  multimult_in4  = 4;
  multimult_in5  = 5;
  multimult_in6  = 6;
  multimult_in7  = 7;
  multimult_in8  = 8;

  multimult_out1 = 0;
  multimult_out2 = 1;
  multimult_out3 = 2;
  multimult_out4 = 3;
  multimult_out5 = 4;
  multimult_out6 = 5;
  multimult_out7 = 6;
  multimult_out8 = 7;

    constructor TModMultiMult.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMultiMult.SetDefaults; // override;
    begin
      FInputs[ multimult_mult] := 1; // set module transparent with unconnected multiplier
    end;

    procedure   TModMultiMult.DoTick; // override;
    begin
      FOutputs[ multimult_out1] := FInputs[ multimult_mult] * FInputs[ multimult_in1];
      FOutputs[ multimult_out2] := FInputs[ multimult_mult] * FInputs[ multimult_in2];
      FOutputs[ multimult_out3] := FInputs[ multimult_mult] * FInputs[ multimult_in3];
      FOutputs[ multimult_out4] := FInputs[ multimult_mult] * FInputs[ multimult_in4];
      FOutputs[ multimult_out5] := FInputs[ multimult_mult] * FInputs[ multimult_in5];
      FOutputs[ multimult_out6] := FInputs[ multimult_mult] * FInputs[ multimult_in6];
      FOutputs[ multimult_out7] := FInputs[ multimult_mult] * FInputs[ multimult_in7];
      FOutputs[ multimult_out8] := FInputs[ multimult_mult] * FInputs[ multimult_in8];
    end;

    function    TModMultiMult.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        multimult_mult : Result := 'mult';
        multimult_in1  : Result := 'in1';
        multimult_in2  : Result := 'in2';
        multimult_in3  : Result := 'in3';
        multimult_in4  : Result := 'in4';
        multimult_in5  : Result := 'in5';
        multimult_in6  : Result := 'in6';
        multimult_in7  : Result := 'in7';
        multimult_in8  : Result := 'in8';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModMultiMult.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        multimult_out1 : Result := 'out1';
        multimult_out2 : Result := 'out2';
        multimult_out3 : Result := 'out3';
        multimult_out4 : Result := 'out4';
        multimult_out5 : Result := 'out5';
        multimult_out6 : Result := 'out6';
        multimult_out7 : Result := 'out7';
        multimult_out8 : Result := 'out8';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoteToDelay = class( TMod)
  public
}

const

  notetodelay_in  = 0;

  notetodelay_out = 0;

    constructor TModNoteToDelay.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModNoteToDelay.DoSlowTick; // override;
    begin
      FOutputs[ notetodelay_out] := 1 / LookupFrequency( FInputs[ notetodelay_in]);
    end;

    function    TModNoteToDelay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        notetodelay_in : Result := 'in';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModNoteToDelay.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        notetodelay_out : Result := 'out';
        else              Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModZeroCounter = class( TMod)
  private
    FOldSign : Boolean;
    FTime    : Integer;
  public
}

const

  zerocounter_in        = 0;
  zerocounter_threshold = 1;

  zerocounter_period    = 0;
  zerocounter_gate      = 1;
  zerocounter_note      = 2;

    constructor TModZeroCounter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModZeroCounter.SetDefaults; // override;
    begin
      FDezipperMap := [
        zerocounter_threshold
      ];
    end;

    procedure   TModZeroCounter.DoTick; // override;
    var
      aValue : TSignal;
      aLevel : TSignal;
      aSign  : Boolean;
      aTime  : TSignal;
      aFreq  : TSignal;
      aNote  : TSignal;
    begin
      Inc( FTime);
      aValue := FInputs[ zerocounter_in];
      aLevel := Abs( aValue);
      if aLevel > FInputs[ zerocounter_threshold]
      then begin
        FOutputs[ zerocounter_gate] := 1;
        aSign := SignalToLogic( aValue);
        if aSign <> FOldSign
        then begin
          FOldSign := aSign;
          FOutputs[ zerocounter_period] := LogicToSignal( aSign);
          if not aSign
          then begin
            aTime := FTime / System_Rate;
            if aTime > 0
            then begin
              aFreq := 1 / aTime;
              aNote := FrequencyToUnits( aFreq) - NoteNumberToUnits( MiddleNote);
              FOutputs[ zerocounter_note] := aNote;
            end;
            FTime := 0;
          end;
        end;
      end
      else FOutputs[ zerocounter_gate] := -1;
    end;

    procedure   TModZeroCounter.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, FOutputs[ zerocounter_gate]], AppLocale));
    end;

    function    TModZeroCounter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        zerocounter_in        : Result := 'in';
        zerocounter_threshold : Result := 'threshold';
        else                    Result := NO_INPUT;
      end;
    end;

    function    TModZeroCounter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        zerocounter_period : Result := 'period';
        zerocounter_gate   : Result := 'gate';
        zerocounter_note   : Result := 'note';
        else                 Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModMedian = class( TMod)
  private
    FData      : TSignalFixedQueue;
    FOldTrig   : Boolean;
  public
}

const

  median_in      = 0;
  median_trig    = 1;

  median_out     = 0;
  median_high    = 1;
  median_low     = 2;

//  public

    constructor TModMedian.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FData   := TMedianCalculator.Create( 0);
      FIsFast := True;
      FIsSlow := False;
    end;

    destructor  TModMedian.Destroy; // override;
    begin
      FreeAndNil( FData);
      inherited;
    end;

    procedure   TModMedian.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      aBinSize : Integer;
      NewData  : TMedianCalculator;
      OldData  : TMedianCalculator;
      i        : Integer;
    begin
      if SameText( aName, 'binsize')
      then begin
        aBinSize := SignalToInt( aValue);
        if not Assigned( FData) or ( aBinSize <> FData.Count)
        then begin
          NewData := TMedianCalculator.Create( aBinSize);
          for i := 0 to FData.Count - 1
          do NewData.Append( FData.GetData);
          OldData := FData;
          FData   := NewData;
          FreeAndnil( OldData);
        end;
      end
      else inherited;
    end;

    function    TModMedian.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'binsize')
      then begin
        if Assigned( FData)
        then Result := FData.Count
        else Result := 0;
      end
      else Result := inherited;
    end;

    procedure   TModMedian.DoTick; // override;
    var
      aTrig : Boolean;
    begin
      aTrig := SignalToLogic( FInputs[ median_trig]);
      if ( aTrig <> FOldTrig) and aTrig and Assigned( FData) and ( FData.Count > 0)
      then begin
        FData.Append( FInputs[ median_in]);
        FOutputs[ median_out ] := FData.Median;
        FOutputs[ median_high] := FData.Highest;
        FOutputs[ median_low ] := FData.Lowest;
      end;
      FOldTrig := aTrig;
    end;

    procedure   TModMedian.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ median_trig]], AppLocale));
    end;

    function    TModMedian.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        median_in   : Result := 'in';
        median_trig : Result := 'trig';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModMedian.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        median_out  : Result := 'out';
        median_high : Result := 'high';
        median_low  : Result := 'low';
        else          Result := NO_OUTPUT;
      end;
    end;

    function    TModMedian.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'binsize';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModCompare = class( TMod)
  public
}

const

  compare_in1      = 0;
  compare_in2      = 1;
  compare_range    = 2;

  compare_outless  = 0;
  compare_outequal = 1;
  compare_outmore  = 2;

    constructor TModCompare.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModCompare.DoTick; // override;
    var
      in1      : TSignal;
      in2      : TSignal;
      outless  : TSignal;
      outequal : TSignal;
      outmore  : TSignal;
    begin
      in1      := FInputs[ compare_in1];
      in2      := FInputs[ compare_in2];

      outless  := LogicToSignal( False);
      outequal := LogicToSignal( False);
      outmore  := LogicToSignal( False);

      if AlmostEqual( in1, in2, FInputs[ compare_range])
      then outequal := LogicToSignal( True)
      else if in1 < in2
      then outless := LogicToSignal( True)
      else outmore  := LogicToSignal( True);

      FOutputs[ compare_outless ] := outless;
      FOutputs[ compare_outequal] := outequal;
      FOutputs[ compare_outmore ] := outmore;
    end;

    procedure   TModCompare.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/less/%f' , [ aPrefix, Name, FOutputs[ compare_outless ]], AppLocale));
      aCallback( Self, Format( '%s/%s/equal/%f', [ aPrefix, Name, FOutputs[ compare_outequal]], AppLocale));
      aCallback( Self, Format( '%s/%s/more/%f' , [ aPrefix, Name, FOutputs[ compare_outmore ]], AppLocale));
    end;

    function    TModCompare.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        compare_in1   : Result := 'in1';
        compare_in2   : Result := 'in2';
        compare_range : Result := 'range';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModCompare.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        compare_outless  : Result := 'outless';
        compare_outequal : Result := 'outequal';
        compare_outmore  : Result := 'outmore';
        else               Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModLogic = class( TMod)
  public
}

const

  gate_function = 0;
  gate_width    = 1;
  gate_in1      = 2;
  gate_in2      = 3;
  gate_in3      = 4;
  gate_in4      = 5;

  gate_out      = 0;

  gatefunc_and  = 0;
  gatefunc_nand = 1;
  gatefunc_or   = 2;
  gatefunc_nor  = 3;
  gatefunc_xor  = 4; // odd parity, true when an odd number of inputs is active
  gatefunc_xnor = 5; // even pariry, true when an even number of inputs is active
  gatefunc_one  = 6; // true when exactly one input is active
  gatefunc_none = 7; // false when exactly one input is activee

    constructor TModGate.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModGate.SetDefaults; // override;
    var
      i : Integer;
    begin
      case SignalToInt( FInputs[ gate_function]) of
        gatefunc_and :
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 1, so it'll 'work'
            do FInputs[ i] := 1;
          end;
        gatefunc_nand :
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 1, so it'll 'work'
            do FInputs[ i] := 1;
          end;
        gatefunc_or :
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        gatefunc_nor :
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        gatefunc_xor :
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        gatefunc_xnor :
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        gatefunc_one :
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
        else // none
          begin
            for i := gate_in1 to gate_in4         // Make sure unused inputs are set to 0, so it'll 'work'
            do FInputs[ i] := 0;
          end;
      end;
    end;

    procedure   TModGate.DoTick; // override;
    var
      i         : Integer;
      R         : TBinary;
      OneCounts : Integer;
      BitCount  : Integer;
      aFunction : Integer;
    begin
      aFunction := SignalToInt( FInputs[ gate_function]);
      if aFunction <> FFunction
      then begin
        FFunction := aFunction;
        SetDefaults;
      end;
      BitCount := SignalToInt( FInputs[ gate_width]) + 1;

      case aFunction of
        gatefunc_and :
          begin
            if BitCount = 1
            then begin
              FOutputs[ gate_out] :=
                LogicToSignal(
                  SignalToLogic( FInputs[ gate_in1]) and
                  SignalToLogic( FInputs[ gate_in2]) and
                  SignalToLogic( FInputs[ gate_in3]) and
                  SignalToLogic( FInputs[ gate_in4])
                );
            end
            else begin
              R := -1;
              for i := gate_in1 to gate_in4
              do R := R and SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ gate_out] := BinaryToSignal( R, BitCount);
            end;
          end;
        gatefunc_nand :
          begin
            if BitCount = 1
            then begin
              FOutputs[ gate_out] :=
                LogicToSignal(
                  Not(
                    SignalToLogic( FInputs[ gate_in1]) and
                    SignalToLogic( FInputs[ gate_in2]) and
                    SignalToLogic( FInputs[ gate_in3]) and
                    SignalToLogic( FInputs[ gate_in4])
                  )
                );
            end
            else begin
              R := -1;
              for i := gate_in1 to gate_in4
              do R := R and SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ gate_out] := BinaryToSignal( not R, BitCount);
            end;
          end;
        gatefunc_or :
          begin
            if BitCount = 1
            then begin
              FOutputs[ gate_out] :=
                LogicToSignal(
                  SignalToLogic( FInputs[ gate_in1]) or
                  SignalToLogic( FInputs[ gate_in2]) or
                  SignalToLogic( FInputs[ gate_in3]) or
                  SignalToLogic( FInputs[ gate_in4])
                );
            end
            else begin
              R := 0;
              for i := gate_in1 to gate_in4
              do R := R or SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ gate_out] := BinaryToSignal( R, BitCount);
            end;
          end;
        gatefunc_nor :
          begin
            if BitCount = 1
            then begin
              FOutputs[ gate_out] :=
                LogicToSignal(
                  Not(
                    SignalToLogic( FInputs[ gate_in1]) or
                    SignalToLogic( FInputs[ gate_in2]) or
                    SignalToLogic( FInputs[ gate_in3]) or
                    SignalToLogic( FInputs[ gate_in4])
                  )
                );
            end
            else begin
              R := 0;
              for i := gate_in1 to gate_in4
              do R := R or SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ gate_out] := BinaryToSignal( not R, BitCount);
            end;
          end;
        gatefunc_xor :
          begin
            if BitCount = 1
            then begin
              FOutputs[ gate_out] :=
                LogicToSignal(
                  SignalToLogic( FInputs[ gate_in1]) xor
                  SignalToLogic( FInputs[ gate_in2]) xor
                  SignalToLogic( FInputs[ gate_in3]) xor
                  SignalToLogic( FInputs[ gate_in4])
                );
            end
            else begin
              R := 0;
              for i := gate_in1 to gate_in4
              do R := R xor SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ gate_out] := BinaryToSignal( R, BitCount);
            end;
          end;
        gatefunc_xnor :
          begin
            if BitCount = 1
            then begin
              FOutputs[ gate_out] :=
                LogicToSignal(
                  Not(
                    SignalToLogic( FInputs[ gate_in1]) xor
                    SignalToLogic( FInputs[ gate_in2]) xor
                    SignalToLogic( FInputs[ gate_in3]) xor
                    SignalToLogic( FInputs[ gate_in4])
                  )
                );
            end
            else begin
              R := 0;
              for i := gate_in1 to gate_in4
              do R := R xor SignalToBinary( FInputs[ i], BitCount);
              FOutputs[ gate_out] := BinaryToSignal( not R, BitCount);
            end;
          end;
        gatefunc_one :
          begin
            OneCounts := 0;
            for i := gate_in1 to gate_in4
            do begin
              if SignalToLogic( FInputs[ i])
              then Inc( OneCounts);
            end;
            FOutputs[ gate_out] := LogicToSignal( OneCounts = 1);
          end;
        else // none
          begin
            OneCounts := 0;
            for i := gate_in1 to gate_in4
            do begin
              if SignalToLogic( FInputs[ i])
              then Inc( OneCounts);
            end;
            FOutputs[ gate_out] := LogicToSignal( OneCounts <> 1);
          end;
      end;
    end;

    procedure   TModGate.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutputs[ gate_out]], AppLocale));
    end;

    function    TModGate.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        gate_function : Result := 'function';
        gate_width    : Result := 'width';
        gate_in1      : Result := 'in1';
        gate_in2      : Result := 'in2';
        gate_in3      : Result := 'in3';
        gate_in4      : Result := 'in4';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModGate.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        gate_out : Result := 'out';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDivider = class( TMod)
  private
    FLastInput : Boolean;
    FCounter   : Integer;
    FOldCount  : Integer;
    FState     : Boolean;
    FOldReset  : Boolean;
  private
}

const

  divider_count = 0;
  divider_trig  = 1;
  divider_reset = 2;

  divider_out   = 0;

const

  DividerStates : array[ Boolean] of TSignal = ( -1, 1);

    procedure   TModDivider.PerformReset;
    begin
      FCounter   := Max( 0, FOldCount - 1);
      FState     := True;
      FLastInput := True;
      FOutputs[ divider_out] := DividerStates[ FState];
    end;

//  public

    constructor TModDivider.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModDivider.DoTick; // override;
    var
      Inp      : Boolean;
      Res      : Boolean;
      NewCount : Integer;
    begin
      Res      := SignalToLogic( FInputs[ divider_reset]);
      NewCount := SignalToInt  ( FInputs[ divider_count]);

      if ResetFlag or ( Res and not FOldReset) or ( FOldCount <> NewCount)
      then begin
        FOldCount := NewCount;
        PerformReset
      end
      else begin
        Inp := SignalToLogic( FInputs[ divider_trig]);
        if FLastInput xor Inp
        then begin
          if FCounter <= 0
          then begin
            FCounter := Max( 0, FOldCount - 1);
            FState   := not FState;
            FOutputs[ divider_out] := DividerStates[ FState];
          end
          else Dec( FCounter);
        end;
        FLastInput := Inp;
      end;
      ResetFlag := False;
      FOldReset := Res;
    end;

    procedure   TModDivider.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f'  , [ aPrefix, Name, FOutputs[ divider_out  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInputs [ divider_trig ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInputs [ divider_reset]], AppLocale));
    end;

    function    TModDivider.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        divider_count : Result := 'count';
        divider_trig  : Result := 'trig';
        divider_reset : Result := 'reset';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModDivider.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        divider_out : Result := 'out';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDFlipFlop = class( TMod)
  private
    FClockState : Boolean;
  public
}

const

  dflipflop_d    = 0;
  dflipflop_clk  = 1;

  dflipflop_q    = 0;
  dflipflop_notq = 1;


    constructor TModDFlipFlop.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModDFlipFlop.DoTick; // override;
    var
      CLK : Boolean;
      Q   : Boolean;
    begin
      if ResetFlag
      then begin
        ResetFlag              := False;
        FOutputs[ dflipflop_q] := LogicToSignal( False);
        Q                      := False;
      end
      else begin
        CLK := SignalToLogic( FInputs[ dflipflop_clk]);
        if not FClockState and CLK
        then begin
          Q                      := SignalToLogic( FInputs[ dflipflop_d]);
          FOutputs[ dflipflop_q] := LogicToSignal( Q);
        end
        else Q := SignalToLogic( FOutputs[ dflipflop_q]);
        FClockState := CLK;
      end;
      FOutputs[ dflipflop_notq] := LogicToSignal( not Q);
    end;

    procedure   TModDFlipFlop.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/q/%f'   , [ aPrefix, Name, FOutputs[ dflipflop_q   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/notq/%f', [ aPrefix, Name, FOutputs[ dflipflop_notq]], AppLocale));
    end;

    function    TModDFlipFlop.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        dflipflop_d   : Result := 'd';
        dflipflop_clk : Result := 'clk';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModDFlipFlop.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        dflipflop_q    : Result := 'q';
        dflipflop_notq : Result := 'notq';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRSFlipFlop = class( TMod)
  private
    FSetState   : Boolean;
    FResetState : Boolean;
  public
}

const

  rsflipflop_s        = 0;
  rsflipflop_r        = 1;
  rsflipflop_trigmode = 2;

  rsflipflop_q        = 0;
  rsflipflop_notq     = 1;

  rsff_mode_edge      = 0;
  rsff_mode_level     = 1;

    constructor TModRSFlipFlop.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModRSFlipFlop.DoTick; // override;
    var
      S    : Boolean;
      R    : Boolean;
      Q    : Boolean;
      Mode : Integer;
    begin
      Mode := SignalToInt( FInputs[ rsflipflop_trigmode]);
      S    := SignalToLogic( FInputs [ rsflipflop_s]);
      R    := SignalToLogic( FInputs [ rsflipflop_r]) or ResetFlag;
      Q    := SignalToLogic( FOutputs[ rsflipflop_q]);
      if (( not FResetState) and R) or (( Mode = rsff_mode_level) and R)
      then Q := False
      else if (( not FSetState) and S) or (( Mode = rsff_mode_level) and S)
      then Q := True;
      FOutputs[ rsflipflop_q   ] := LogicToSignal(     Q);
      FOutputs[ rsflipflop_notq] := LogicToSignal( not Q);
      FSetState   := S;
      FResetState := R;
      ResetFlag   := False;
    end;

    procedure   TModRSFlipFlop.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/q/%f'   , [ aPrefix, Name, FOutputs[ rsflipflop_q   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/notq/%f', [ aPrefix, Name, FOutputs[ rsflipflop_notq]], AppLocale));
    end;

    function    TModRSFlipFlop.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rsflipflop_s        : Result := 's';
        rsflipflop_r        : Result := 'r';
        rsflipflop_trigmode : Result := 'trigmode';
        else                  Result := NO_INPUT;
      end;
    end;

    function    TModRSFlipFlop.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rsflipflop_q    : Result := 'q';
        rsflipflop_notq : Result := 'notq';
        else              Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModCounter = class( TMod)
  private
    FState   : Word;
    FMode    : Integer;
    FOldRes  : Boolean;
    FOldTrig : Boolean;
  public
}

const

  counter_trig  =  0;
  counter_reset =  1;
  counter_dir   =  2;
  counter_mode  =  3;

  counter_out0  =  0;
  counter_out1  =  1;
  counter_out2  =  2;
  counter_out3  =  3;
  counter_out4  =  4;
  counter_out5  =  5;
  counter_out6  =  6;
  counter_out7  =  7;
  counter_out8  =  8;
  counter_out9  =  9;
  counter_out10 = 10;
  counter_out11 = 11;
  counter_out12 = 12;
  counter_out13 = 13;
  counter_out14 = 14;
  counter_out15 = 15;

  counter_bin   =  0;
  counter_seq   =  1;


    constructor TModCounter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModCounter.SetDefaults; // override;
    begin
      FInputs[ counter_dir] := 1; // Default direction is up
    end;

    procedure   TModCounter.DoTick; // override;
    const
      MaxCount = 16;
    var
      Direction : Boolean;
      Trig      : Boolean;
      Res       : Boolean;
      Mode      : Integer;
      i         : Integer;
      p         : Integer;
    begin
      Direction := SignalToLogic( FInputs[ counter_dir  ]);
      Mode      := SignalToInt  ( FInputs[ counter_mode ]);
      Trig      := SignalToLogic( FInputs[ counter_trig ]);
      Res       := SignalToLogic( FInputs[ counter_reset]);

      if (( not FOldRes) and Res) or ResetFlag or ( Mode <> FMode)
      then begin
        FMode    := Mode;
        FOldTrig := True;
        if Direction
        then FState := 0               // Forwards, reset to first state
        else begin
          if FMode = counter_bin       // Backwards, reset to last state
          then FState := $ffff         // binary
          else FState := MaxCount - 1; // sequential (or ring)
        end;
      end
      else begin
        if Trig and not FOldTrig
        then begin
          if Direction
          then begin                     // Forwards, go to nexxt state
          {$Q-R-}
            Inc( FState);                // binary counter will oveflow properly
          {$Q+R+}
            if FMode = counter_seq       // ring counter needs a check
            then FState := FState mod MaxCount;
          end
          else begin                     // Backwards, go to previous state
            if FState > 0
            then Dec( FState)
            else begin
              if FMode = counter_bin
              then FState := $ffff
              else FState := MaxCount - 1;
            end;
          end;
        end;
        FOldTrig := Trig;
      end;
      FOldRes   := Res;
      ResetFlag := False;


      if FMode = counter_bin
      then begin
        p := 1;
        for i := 0 to MaxCount - 1
        do begin
          if ( FState and p) <> 0
          then FOutputs[ counter_out0 + i] := LogicToSignal( True )
          else FOutputs[ counter_out0 + i] := LogicToSignal( False);
          p := p shl 1;
        end;
      end
      else begin
        for i := 0 to MaxCount - 1
        do begin
          if i = FState
          then FOutputs[ counter_out0 + i] := LogicToSignal( True )
          else FOutputs[ counter_out0 + i] := LogicToSignal( False);
        end;
      end;
    end;

    procedure   TModCounter.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := counter_out0 to counter_out15
      do aCallback( Self, Format( '%s/%s/out%d/%f', [ aPrefix, Name, 1 + i - counter_out1, FOutputs[ i]], AppLocale));

      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInputs[ counter_trig ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInputs[ counter_reset]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'  , [ aPrefix, Name, FInputs[ counter_dir  ]], AppLocale));
    end;

    function    TModCounter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        counter_trig  : Result := 'trig';
        counter_reset : Result := 'reset';
        counter_dir   : Result := 'dir';
        counter_mode  : Result := 'mode';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModCounter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        counter_out0  : Result := 'out0';
        counter_out1  : Result := 'out1';
        counter_out2  : Result := 'out2';
        counter_out3  : Result := 'out3';
        counter_out4  : Result := 'out4';
        counter_out5  : Result := 'out5';
        counter_out6  : Result := 'out6';
        counter_out7  : Result := 'out7';
        counter_out8  : Result := 'out8';
        counter_out9  : Result := 'out9';
        counter_out10 : Result := 'out19';
        counter_out11 : Result := 'out11';
        counter_out12 : Result := 'out12';
        counter_out13 : Result := 'out13';
        counter_out14 : Result := 'out14';
        counter_out15 : Result := 'out15';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAdc = class( TMod)
  public
}

const

  adc_in    =  0;

  adc_out0  =  0;
  adc_out1  =  1;
  adc_out2  =  2;
  adc_out3  =  3;
  adc_out4  =  4;
  adc_out5  =  5;
  adc_out6  =  6;
  adc_out7  =  7;
  adc_out8  =  8;
  adc_out9  =  9;
  adc_out10 = 10;
  adc_out11 = 11;
  adc_out12 = 12;
  adc_out13 = 13;
  adc_out14 = 14;
  adc_out15 = 15;

  adc_bits  = 16;

    constructor TModAdc.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModAdc.DoTick; // override;
    var
      aValue : Integer;
      i      : Integer;
    begin
      aValue := SignalToBinary( FInputs[ adc_in], adc_out15 - adc_out0 + 1);
      for i := adc_out0 to adc_out15
      do begin
        FOutputs[ i] := LogicToSignal( aValue and 1 = 1);
        aValue := aValue shr 1;
      end;
    end;

    procedure   TModAdc.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := adc_out0 to adc_out15
      do aCallback( Self, Format( '%s/%s/out%d/%f', [ aPrefix, Name, i - adc_out0, FOutputs[ i]], AppLocale));
    end;

    function    TModAdc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        adc_in : Result := 'in';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModAdc.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        adc_out0  : Result := 'out0';
        adc_out1  : Result := 'out1';
        adc_out2  : Result := 'out2';
        adc_out3  : Result := 'out3';
        adc_out4  : Result := 'out4';
        adc_out5  : Result := 'out5';
        adc_out6  : Result := 'out6';
        adc_out7  : Result := 'out7';
        adc_out8  : Result := 'out8';
        adc_out9  : Result := 'out9';
        adc_out10 : Result := 'out10';
        adc_out11 : Result := 'out11';
        adc_out12 : Result := 'out12';
        adc_out13 : Result := 'out13';
        adc_out14 : Result := 'out14';
        adc_out15 : Result := 'out15';
        else        Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDac = class( TMod)
  public
}

const

  dac_in0  =  0;
  dac_in1  =  1;
  dac_in2  =  2;
  dac_in3  =  3;
  dac_in4  =  4;
  dac_in5  =  5;
  dac_in6  =  6;
  dac_in7  =  7;
  dac_in8  =  8;
  dac_in9  =  9;
  dac_in10 = 10;
  dac_in11 = 11;
  dac_in12 = 12;
  dac_in13 = 13;
  dac_in14 = 14;
  dac_in15 = 15;

  dac_out  =  0;

  dac_bits = 16;

    constructor TModDac.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModDac.DoTick; // override;
    var
      aValue : Integer;
      i      : Integer;
    begin
      aValue := 0;
      for i := dac_in0 to dac_in15
      do begin
        if SignalToLogic( FInputs[ i])
        then aValue := ( aValue shr 1) or ( 1 shl ( dac_bits - 1))
        else aValue := aValue shr 1;
      end;
      FOutputs[ dac_out] := BinaryToSignal( aValue, dac_bits);
    end;

    procedure   TModDac.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := dac_in0 to dac_in15
      do aCallback( Self, Format( '%s/%s/in%d/%f', [ aPrefix, Name, i - dac_in0, FInputs[ i]], AppLocale));
    end;

    function    TModDac.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        dac_in0  : Result := 'in0';
        dac_in1  : Result := 'in1';
        dac_in2  : Result := 'in2';
        dac_in3  : Result := 'in3';
        dac_in4  : Result := 'in4';
        dac_in5  : Result := 'in5';
        dac_in6  : Result := 'in6';
        dac_in7  : Result := 'in7';
        dac_in8  : Result := 'in8';
        dac_in9  : Result := 'in9';
        dac_in10 : Result := 'in10';
        dac_in11 : Result := 'in11';
        dac_in12 : Result := 'in12';
        dac_in13 : Result := 'in13';
        dac_in14 : Result := 'in14';
        dac_in15 : Result := 'in15';
        else       Result := NO_INPUT;
      end;
    end;

    function    TModDac.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        dac_out  : Result := 'out';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TPDState = 1 .. 12;

  TModPhaseDetect = class( TMod)
  private
    FOldIn1 : Boolean;
    FOldIn2 : Boolean;
    FState  : TPDState;
    FAlpha  : TSignal;
  public
}

const

  phasedetect_in1         = 0;
  phasedetect_in2         = 1;
  phasedetect_alpha       = 2;
  phasedetect_alphamod    = 3;
  phasedetect_alphamodamt = 4;
  phasedetect_type        = 5;
  phasedetect_typesel     = 6;

  phasedetect_out         = 0;
  phasedetect_lock        = 1;

    constructor TModPhaseDetect.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModPhaseDetect.SetDefaults; // override;
    begin
      FState := 1;
      FInputs[ phasedetect_typesel] := -1.0;
      FDezipperMap := [
        phasedetect_alpha,
        phasedetect_alphamodamt
      ];
    end;

    procedure   TModPhaseDetect.DoSlowTick; // override;
    begin
      FAlpha := Normalize( ExpAttack( FInputs[ phasedetect_alpha] + FInputs[ phasedetect_alphamod] * FInputs[ phasedetect_alphamodamt], 1, 0, 1));
    end;

    procedure   TModPhaseDetect.DoTick; // override;
    var
      NewIn1   : Boolean;
      NewIn2   : Boolean;
      In1Up    : Boolean;
      In1Down  : Boolean;
      In2Up    : Boolean;
      In2Down  : Boolean;
      Inp      : TSignal;
      Outp     : TSignal;
      aType    : Integer;
      aTypeSel : Boolean;
    begin
      NewIn1   := SignalToLogic( FInputs[ phasedetect_in1    ]);
      NewIn2   := SignalToLogic( FInputs[ phasedetect_in2    ]);
      aType    := SignalToInt  ( FInputs[ phasedetect_type   ]);
      aTypeSel := SignalToLogic( FInputs[ phasedetect_typesel]);

      if aTypeSel
      then aType := 1 - aType;

      In1Up   := not FOldIn1 and     NewIn1;
      In1Down :=     FOldIn1 and not NewIn1;
      In2Up   := not FOldIn2 and     NewIn2;
      In2Down :=     FOldIn2 and not NewIn2;

      case FState of
         1 : if in1up   then FState :=  3 else if in2up   then FState :=  2;
         2 : if In1Up   then FState :=  6 else if In2Down then FState :=  4;
         3 : if In1Down then FState :=  5 else if In2Up   then FState :=  6;
         4 : if In1Up   then FState :=  8 else if In2Up   then FState :=  2;
         5 : if In1Up   then FState :=  3 else if In2Up   then FState :=  7;
         6 : if In1Down then FState :=  7 else if in2down then FState :=  8;
         7 : if In1Up   then FState :=  9 else if In2Down then FState :=  1;
         8 : if In1Down then FState :=  1 else if In2Up   then FState := 10;
         9 : if In1Down then FState := 11 else if In2Down then FState :=  3;
        10 : if In1Down then FState :=  2 else if In2Down then FState := 12;
        11 : if In1Up   then FState :=  9 else if In2Down then FState :=  5;
        12 : if In1Down then FState :=  4 else if In2Up   then FState := 10;
      end;

      Outp := FOutputs[ phasedetect_out];

      if aType = 0
      then Inp := LogicToSignal( NewIn1 xor NewIn2)
      else begin
        if FState in [ 3, 5, 9, 11]
        then Inp := 1.0                                      // Input frequency is higher
        else if FState in [ 2, 4, 10, 12]
        then Inp := - 1.0                                    // Input frequency is higher
        else Inp := Outp;                                    // Input frequency is the same
      end;

      Outp  := Outp + FAlpha * ( Inp - Outp);

      FOutputs[ phasedetect_out ] := Normalize( Outp);
      FOutputs[ phasedetect_lock] := LogicToSignal( FState in [ 1, 6, 7, 8]);

      FOldIn1 := NewIn1;
      FOldIn2 := NewIn2;
    end;

    procedure   TModPhaseDetect.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/up/%f'  , [ aPrefix, Name, LogicToSignal( FState in [ 3, 5,  9, 11])], AppLocale));
      aCallback( Self, Format( '%s/%s/lock/%f', [ aPrefix, Name, LogicToSignal( FState in [ 1, 6,  7,  8])], AppLocale));
      aCallback( Self, Format( '%s/%s/down/%f', [ aPrefix, Name, LogicToSignal( FState in [ 2, 4, 10, 12])], AppLocale));
    end;

    function    TModPhaseDetect.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        phasedetect_in1         : Result := 'in1';
        phasedetect_in2         : Result := 'in2';
        phasedetect_alpha       : Result := 'alpha';
        phasedetect_alphamod    : Result := 'alphamod';
        phasedetect_alphamodamt : Result := 'alphamodamt';
        phasedetect_type        : Result := 'type';
        phasedetect_typesel     : Result := 'typesel';
        else                      Result := NO_INPUT;
      end;
    end;

    function    TModPhaseDetect.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        phasedetect_out  : Result := 'out';
        phasedetect_lock : Result := 'lock';
        else               Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNot = class( TMod)
  public
}

const

  not_in1  = 0;
  not_in2  = 1;
  not_in3  = 2;
  not_in4  = 3;

  not_out1 = 0;
  not_out2 = 1;
  not_out3 = 2;
  not_out4 = 3;

    constructor TModNot.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModNot.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := not_in1 to not_in4          // Set unused inputs to 1 - just to keep the lights off
      do FInputs[ i] := 1;
    end;

    procedure   TModNot.DoTick; // override;
    var
      i : Integer;
    begin
      for i := not_in1 to not_in4
      do FOutputs[ i] := LogicToSignal( not SignalToLogic( FInputs[ i]));
    end;

    procedure   TModNot.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := not_out1 to not_out4
      do aCallback( Self, Format( '%s/%s/out%d/%f', [ aPrefix, Name, i + 1, FOutputs[ i]], AppLocale));
    end;

    function    TModNot.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        not_in1 : Result := 'in1';
        not_in2 : Result := 'in2';
        not_in3 : Result := 'in3';
        not_in4 : Result := 'in4';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModNot.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        not_out1 : Result := 'out1';
        not_out2 : Result := 'out2';
        not_out3 : Result := 'out3';
        not_out4 : Result := 'out4';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAmplifier = class( TMod)
  public
}

const

  amplifier_gain = 0;
  amplifier_mute = 1;
  amplifier_in1  = 2; // keep in1 .. in4 consecutive
  amplifier_in2  = 3;
  amplifier_in3  = 4;
  amplifier_in4  = 5;

  amplifier_out1 = 0; // keep out1 .. out4 consecutive
  amplifier_out2 = 1;
  amplifier_out3 = 2;
  amplifier_out4 = 3;

    constructor TModAmplifier.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModAmplifier.SetDefaults; // override;
    begin
      FDezipperMap := [
        amplifier_gain,
        amplifier_mute
      ];
    end;

    procedure   TModAmplifier.DoTick; // override;
    var
      i : Integer;
      G : TSignal;
    begin
      G := FInputs[ amplifier_gain] * SignalToMute( FInputs[ amplifier_mute]);
      for i := amplifier_in1 to amplifier_in4
      do FOutputs[ i - amplifier_in1 + amplifier_out1] := FInputs[ i] * G;
    end;

    function    TModAmplifier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        amplifier_gain : Result := 'gain';
        amplifier_mute : Result := 'mute';
        amplifier_in1  : Result := 'in1';
        amplifier_in2  : Result := 'in2';
        amplifier_in3  : Result := 'in3';
        amplifier_in4  : Result := 'in4';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModAmplifier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        amplifier_out1: Result := 'out1';
        amplifier_out2: Result := 'out2';
        amplifier_out3: Result := 'out3';
        amplifier_out4: Result := 'out4';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModScaler = class( TMod)
  public
}

const

  scaler_gain   = 0;
  scaler_offset = 1;
  scaler_mode   = 2;
  scaler_mute   = 3;
  scaler_in1    = 4;
  scaler_in2    = 5;
  scaler_in3    = 6;
  scaler_in4    = 7;

  scaler_out1   = 0;
  scaler_out2   = 1;
  scaler_out3   = 2;
  scaler_out4   = 3;
  scaler_sum    = 4;


    constructor TModScaler.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModScaler.SetDefaults; // override;
    begin
      FDezipperMap := [
        scaler_gain,
        scaler_offset,
        scaler_mute
      ];
    end;

    procedure   TModScaler.DoTick; // override;
    var
      i : Integer;
      G : TSignal;
      O : TSignal;
      S : TSignal;
      T : TSignal;
    begin
      G := FInputs[ scaler_gain  ] * SignalToMute( FInputs[ scaler_mute]);
      O := FInputs[ scaler_offset] * SignalToMute( FInputs[ scaler_mute]);
      S := 0.0;
      for i := scaler_in1 to scaler_in4
      do begin
        T :=  G * FInputs[ i] + O;
        FOutputs[ i - scaler_in1 + scaler_out1] := T;
        S := S + T;
      end;
      FOutputs[ scaler_sum] := S;
    end;

    function    TModScaler.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        scaler_gain   : Result := 'gain';
        scaler_offset : Result := 'offset';
        scaler_mode   : Result := 'mode';
        scaler_mute   : Result := 'mute';
        scaler_in1    : Result := 'in1';
        scaler_in2    : Result := 'in2';
        scaler_in3    : Result := 'in3';
        scaler_in4    : Result := 'in4';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModScaler.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        scaler_out1 : Result := 'out1';
        scaler_out2 : Result := 'out2';
        scaler_out3 : Result := 'out3';
        scaler_out4 : Result := 'out4';
        scaler_sum  : Result := 'sum';   // Sum output
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRatio = class( TMod)
  public
}

const

  ratio_multiplier = 0;
  ratio_divider    = 1;
  ratio_mute       = 2;
  ratio_in1        = 3;
  ratio_in2        = 4;
  ratio_in3        = 5;
  ratio_in4        = 6;

  ratio_out1       = 0;
  ratio_out2       = 1;
  ratio_out3       = 2;
  ratio_out4       = 3;


    constructor TModRatio.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModRatio.SetDefaults; // override;
    begin
      FDezipperMap := [
        ratio_multiplier,
        ratio_divider,
        ratio_mute
      ];
    end;

    procedure   TModRatio.DoTick; // override;
    var
      i      : Integer;
      MulDiv : TSignal;
      RDiv   : Integer;
    begin
      RDiv := Round( FInputs[ ratio_divider]);
      if RDiv = 0
      then Muldiv := 1
      else Muldiv := Round( SignalToMute( FInputs[ ratio_mute]) * FInputs[ ratio_multiplier]) div RDiv;
      for i := ratio_in1 to ratio_in4
      do FOutputs[ i - ratio_in1 + ratio_out1] := MulDiv * FInputs[ i];
    end;

    function    TModRatio.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        ratio_multiplier : Result := 'multiplier';
        ratio_divider    : Result := 'divider';
        ratio_mute       : Result := 'mute';
        ratio_in1        : Result := 'in1';
        ratio_in2        : Result := 'in2';
        ratio_in3        : Result := 'in3';
        ratio_in4        : Result := 'in4';
        else               Result := NO_INPUT;
      end;
    end;

    function    TModRatio.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        ratio_out1 : Result := 'out1';
        ratio_out2 : Result := 'out2';
        ratio_out3 : Result := 'out3';
        ratio_out4 : Result := 'out4';
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRotator = class( TMod)
  public
}

const

  rotator_angle = 0;
  rotator_in1   = 1;
  rotator_in2   = 2;

  rotator_out1  = 0;
  rotator_out2  = 1;


    constructor TModRotator.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModRotator.SetDefaults; // override;
    begin
      FInputs[ rotator_angle] := 0;
      FInputs[ rotator_in1  ] := 0;
      FInputs[ rotator_in2  ] := 0;
    end;

    procedure   TModRotator.DoTick; // override;
    var
      aCos    : TSignal;
      aSin    : TSignal;
      in1     : TSignal;
      in2     : TSignal;
    begin
      FAngle := FInputs[ rotator_angle] * TWO_PI;
      SinCos( FAngle, aSin, aCos);
      in1 := FInputs[ rotator_in1];
      in2 := FInputs[ rotator_in2];
      FOutputs[ rotator_out1] := aCos * in1 - aSin * in2;
      FOutputs[ rotator_out2] := aSin * in1 + aCos * in2;
    end;

    function    TModRotator.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rotator_angle : Result := 'angle';
        rotator_in1   : Result := 'in1';
        rotator_in2   : Result := 'in2';
        else Result := NO_INPUT;
      end;
    end;

    function    TModRotator.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rotator_out1 : Result := 'out1';
        rotator_out2 : Result := 'out2';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRectifier = class( TMod)
  public
}

const

  rectifier_mode   = 0;
  rectifier_in     = 1;

  rectifier_out    = 0;
  rectifier_outinv = 1;


    constructor TModRectifier.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModRectifier.SetDefaults; // override;
    begin
      FInputs[ rectifier_in] := 0;
    end;

    procedure   TModRectifier.DoTick; // override;
    var
      Mode : Integer;
      Inp  : TSignal;
    begin
      Mode := SignalToInt( FInputs[ rectifier_mode]);
      Inp  :=              FInputs[ rectifier_in  ] ;

      if Mode = 0                                      // Half wave
      then begin
        if SignalToLogic( Inp)
        then FOutputs[ rectifier_out] := Inp
        else FOutputs[ rectifier_out] := 0;
      end
      else if Mode = 1                                 // Full wave
      then FOutputs[ rectifier_out] := Abs( Inp)
      else FOutputs[ rectifier_out] :=      Inp;      // transparent

      FOutputs[ rectifier_outinv] := - FOutputs[ rectifier_out];
    end;

    function    TModRectifier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rectifier_mode : Result := 'mode';
        rectifier_in   : Result := 'in';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModRectifier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        rectifier_out    : Result := 'out';
        rectifier_outinv : Result := 'outinv';
        else               Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMinMax = class( TMod)
  public
}

const

  minmax_in1 = 0;
  minmax_in2 = 1;

  minmax_min = 0;
  minmax_max = 1;


    constructor TModMinMax.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMinMax.SetDefaults; // override;
    begin
      FInputs[ minmax_in1] := 0;
      FInputs[ minmax_in2] := 0;
    end;

    procedure   TModMinMax.DoTick; // override;
    var
      in1 : TSignal;
      in2 : TSignal;
    begin
      in1                   := FInputs[ minmax_in1];
      in2                   := FInputs[ minmax_in2];
      FOutputs[ minmax_min] := Min( in1, in2);
      FOutputs[ minmax_max] := Max( in1, in2);
    end;

    function    TModMinMax.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        minmax_in1 : Result := 'in1';
        minmax_in2 : Result := 'in2';
        else         Result := NO_INPUT;
      end;
    end;

    function    TModMinMax.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        minmax_min : Result := 'min';
        minmax_max : Result := 'max';
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModIntDif = class( TMod)
  private
    FSum        : TSignal;
    FlastSample : TSignal;
  public
}

const

  intdif_in   = 0;
  intdif_mode = 1;

  intdif_int  = 0;
  intdif_dif  = 1;


    constructor TModIntDif.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModIntDif.SetDefaults; // override;
    begin
      FInputs[ intdif_in] := 0;
    end;

    procedure   TModIntDif.DoTick; // override;
    var
      Inp  : TSignal;
      Dif  : TSignal;
      Mode : Integer;
    begin
      Inp := FInputs[ intdif_in];

      if ResetFlag
      then begin
        ResetFlag   := False;
        FSum        := 0;
        FlastSample := Inp;
      end
      else FSum := Clip( FSum + Inp, -1000, 1000);

      Mode := SignalToInt( FInputs[ intdif_mode]);
      Dif  := Inp - FlastSample;

      if Mode = 0                                      // Half wave
      then begin
        if SignalToLogic( Dif)
        then FOutputs[ intdif_dif] := Dif
        else FOutputs[ intdif_dif] := 0;
      end
      else if Mode = 1                                 // Full wave
      then FOutputs[ intdif_dif] := Abs( Dif)
      else FOutputs[ intdif_dif] :=      Dif;         // Transparent

      FOutputs[ intdif_int] := FSum;
      FlastSample            := Inp;
    end;

    function    TModIntDif.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        intdif_in   : Result := 'in';
        intdif_mode : Result := 'mode';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModIntDif.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        intdif_int : Result := 'int';
        intdif_dif : Result := 'dif';
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModBaseOsc = class( TMod)
  private
    FPhase      : TSignal;
    FPhaseDelta : TSignal;
    FPosition   : TSignal;
    FOldSync    : Boolean;
    FShape      : Integer;
    FBandLimit  : TSignal;
  public
}

const

  baseosc_frequency  =  0;
  baseosc_fm         =  1;
  baseosc_phase      =  2;
  baseosc_sync       =  3;
  baseosc_mute       =  4;
  baseosc_shape      =  5;
  baseosc_pmlevel    =  6;
  baseosc_fmlevel    =  7;
  baseosc_cents      =  8;
  baseosc_bandlimit  =  9;
  baseosc_pastlastin = 10;

  baseosc_out        = 0;


    constructor TModBaseOsc.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModBaseOsc.SetDefaults; // override;
    begin
      FDezipperMap := [
        baseosc_frequency,
        baseosc_mute,
        baseosc_pmlevel,
        baseosc_fmlevel,
        baseosc_cents
      ];
      FInputs[ baseosc_bandlimit] := 3;
    end;

    procedure   TModBaseOsc.DoTick; // override;
    var
      Trunced: Integer;
    begin
      if ( not FOldSync) and SignalToLogic( FInputs[ baseosc_sync])
      then FPhase  := 0.0;

      FOldSync    := SignalToLogic( FInputs[ baseosc_sync]);
      FPhaseDelta := LookupPhaseDelta( FInputs[ baseosc_frequency] + FInputs[ baseosc_fm] * FInputs[ baseosc_fmlevel] + FInputs[ baseosc_cents]);

      FPhase      := FPhase + FPhaseDelta;
      Trunced     := Trunc( FPhase);
      FPhase      := FPhase - Trunced;
      if FPhase < 0
      then FPhase := FPhase + 1;

      FPosition   := FPhase + Clip( FInputs[ baseosc_phase] * FInputs[ baseosc_pmlevel], -1, 1);
      Trunced     := Trunc( FPosition);
      FPosition   := FPosition - Trunced;
      if FPosition < 0
      then FPosition := FPosition + 1;

      FShape      := SignalToInt( FInputs[ baseosc_shape    ]);
      FBandLimit  :=              FInputs[ baseosc_bandlimit];
    end;

    function    TModBaseOsc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        baseosc_frequency : Result := 'frequency';
        baseosc_fm        : Result := 'fm';
        baseosc_phase     : Result := 'phase';
        baseosc_sync      : Result := 'sync';
        baseosc_mute      : Result := 'mute';
        baseosc_shape     : Result := 'shape';
        baseosc_pmlevel   : Result := 'pmlevel';
        baseosc_fmlevel   : Result := 'fmlevel';
        baseosc_cents     : Result := 'cents';
        baseosc_bandlimit : Result := 'bandlimit';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModBaseOsc.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        baseosc_out : Result := 'out';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModBaseTrigOsc = class( TModBaseOsc)
  private
    FState   : TTrigOscState;
    FCounter : TSignal;
  public
}

const

  basetrigosc_trig  = baseosc_phase;
  basetrigosc_count = baseosc_sync;


    procedure   TModbaseTrigOsc.DoTick; // override;
    var
      FOldCounter : Integer;
      Trunced     : Integer;
    begin

      case FState of

        toIdle :
          begin
            FPosition := 0.0;
            FCounter  := 0;
            if SignalToLogic( FInputs[ basetrigosc_trig])
            then FState := toRunning;
          end;

        toRunning,
        toWaiting :
          begin
            // This is almost the inherited code - except - it ignores phase

            FPhaseDelta := LookupPhaseDelta( FInputs[ baseosc_frequency] + FInputs[ baseosc_fm] * FInputs[ baseosc_fmlevel] + FInputs[ baseosc_cents]);
            FPosition   := FPosition + FPhaseDelta;
            Trunced     := Trunc( FPosition);
            FPosition   := FPosition - Trunced;
            if FPosition < 0
            then FPosition := FPosition + 1;
            FShape := SignalToInt( FInputs[ baseosc_shape]);

            if FState = toRunning
            then begin
              FOldCounter := FCounter;
              if AlmostEqual( FPosition, 0.5, FPhaseDelta)
              then FCounter := FCounter + 1
              else if AlmostEqual( FPosition, 1.0, FPhaseDelta)
              then FCounter := FCounter + 1;
              if FCounter <> FOldCounter
              then begin
                if FCounter >= FInputs[ basetrigosc_count]
                then FState := toDone
                else FState := toWaiting;
              end;
            end
            else begin
              if
                not AlmostEqual( FPosition, 0.5, FPhaseDelta) and
                not AlmostEqual( FPosition, 1.0, FPhaseDelta)
              then FState := toRunning;
            end;

          end;

        toDone :
          begin
            FPosition := 0.0;
            if FInputs[ basetrigosc_trig] <= 0.0
            then FState := toIdle;
          end;
      end;
    end;

    function    TModBaseTrigOsc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        basetrigosc_trig  : Result := 'trig';
        basetrigosc_count : Result := 'count';
        else                Result := inherited;
      end;
    end;


{ ========
  TModOsc = class( TModBaseOsc)
  public
}

    procedure   TModOSc.DoTick; // override;
    var
      aValue : TSignal;
    begin
      inherited;
      case FShape of
        0  : // sine
          FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * LookupSine( FPosition);
        1  : // tri
          begin
            if FPosition < 0.5
            then aValue :=   1
            else aValue := - 1;
            aValue := 4 * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FMod( FPosition + 0.5, 1), FBandLimit, FPhaseDelta));
            FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * ( FPhaseDelta * aValue + ( 1 - FPhaseDelta) * FOutputs[ baseosc_out]);
          end;
        2  : // saw
          FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * (( 2.0 * FPosition - 1) - PolyBLEP( FPosition, FBandLimit, FPhaseDelta));
        3  : // square
          begin
            if FPosition < 0.5
            then aValue :=   1
            else aValue := - 1;
            FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FMod( FPosition + 0.5, 1), FBandLimit, FPhaseDelta));
          end;
      end;
    end;


{ ========
  TModTrigOsc = class( TModBaseTrigOsc)
  public
}

    procedure   TModTrigOsc.DoTick; // override;
    var
      aValue : TSignal;
    begin
      inherited;
      case FShape of
        0  : // sine
          FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * LookupSine( FPosition);
        1  : // tri
          begin
            if FPosition < 0.5
            then aValue :=   1
            else aValue := - 1;
            aValue := 4 * aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FMod( FPosition + 0.5, 1), FBandLimit, FPhaseDelta);
            FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * ( FPhaseDelta * aValue + ( 1 - FPhaseDelta) * FOutputs[ baseosc_out]);
          end;
        2  : // saw
          FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * (( 2.0 * FPosition - 1) - PolyBLEP( FPosition, FBandLimit, FPhaseDelta));
        3  : // square
          begin
            if FPosition < 0.5
            then aValue :=   1
            else aValue := - 1;
            FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FMod( FPosition + 0.5, 1), FBandLimit, FPhaseDelta));
          end;
      end;
    end;


{ ========
  TModSquare = class( TModBaseOsc)
  public
}

const

  square_pwm      =  baseosc_pastlastin + 0;
  square_pwmlevel =  baseosc_pastlastin + 1;


    procedure   TModSquare.SetDefaults;
    begin
      inherited;
      FDezipperMap := FDezipperMap + [
        square_pwmlevel
      ];
    end;

    procedure   TModSquare.DoTick; // override;
    var
      pwmp   : TSignal;
      aValue : TSignal;
    begin
      inherited;                                // For phase calculations
      pwmp := Clip( 0.5 * ( 1.0 - FInputs[ square_pwm] * FInputs[ square_pwmlevel]), 0.01, 0.99);
      if FPosition < pwmp
      then aValue :=   1
      else aValue := - 1;
      FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FMod( FPosition + ( 1 - pwmp), 1), FBandLimit, FPhaseDelta));
    end;

    function    TModSquare.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         square_pwm      : Result := 'pwm';
         square_pwmlevel : Result := 'pwmlevel';
        else               Result := inherited;
      end;
    end;


{ ========
  TModSquareTrig = class( TModBaseTrigOsc)
  public
}

const

  squaretrig_pwm      =  baseosc_pastlastin + 0;
  squaretrig_pwmlevel =  baseosc_pastlastin + 1;


    procedure   TModSquareTrig.SetDefaults;
    begin
      inherited;
      FDezipperMap := FDezipperMap + [
        squaretrig_pwmlevel
      ];
    end;

    procedure   TModSquareTrig.DoTick; // override;
    var
      pwmp   : TSignal;
      aValue : TSignal;
    begin
      inherited;                                // For phase calculations
      pwmp := Clip( 0.5 * ( 1.0 - FInputs[ square_pwm] * FInputs[ square_pwmlevel]), 0.01, 0.99);
      if FPosition < pwmp
      then aValue :=   1
      else aValue := - 1;
      FOutputs[ baseosc_out] := SignalToMute( FInputs[ baseosc_mute]) * ( aValue + PolyBLEP( FPosition, FBandLimit, FPhaseDelta) - PolyBLEP( FMod( FPosition + ( 1 - pwmp), 1), FBandLimit, FPhaseDelta));
    end;

    function    TModSquareTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         squaretrig_pwm      : Result := 'pwm';
         squaretrig_pwmlevel : Result := 'pwmlevel';
        else                   Result := inherited;
      end;
    end;


{ ========
  TModBaseLfo = class( TMod)
  private
    FPhase    : TSignal;
    FPosition : TSignal;
    FOldRes   : Boolean;
    FShape    : Integer;
  public
}

const

  baselfo_frequency  = 0;
  baselfo_fm         = 1;
  baselfo_phase      = 2;
  baselfo_res        = 3;
  baselfo_mute       = 4;
  baselfo_fmlevel    = 5;
  baselfo_pmlevel    = 6;
  baselfo_shape      = 7;
  baselfo_range      = 8;
  baselfo_outputtype = 9;

  baselfo_out        = 0;
  baselfo_sync       = 1;


    constructor TModBaseLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModBaseLfo.SetDefaults; // override;
    begin
      FDezipperMap := [
        baselfo_frequency,
        baselfo_mute,
        baselfo_fmlevel,
        baselfo_pmlevel
      ];
    end;

    procedure   TModBaseLfo.DoSlowTick; // override;
    var
      Delta   : TSignal;
      Trunced : Integer;
    begin
      if ( not FOldRes) and SignalToLogic( FInputs[ baselfo_res]) or ResetFlag
      then FPhase := 0.0;
      ResetFlag := False;
      FOldRes   := SignalToLogic( FInputs[ baselfo_res  ]);
      FShape    := SignalToInt  ( FInputs[ baselfo_shape]);

      case SignalToInt( FInputs[ baselfo_range]) of
        0  : Delta := LookupLfoFastPhaseDelta  ( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
        1  : Delta := LookupLfoMediumPhaseDelta( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
        2  : Delta := LookupLfoSlowPhaseDelta  ( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
        3  : Delta := LookupLfoBPMPhaseDelta   ( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
        else Delta := LookupLfoLinPhaseDelta   ( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
      end;

      FPhase      := FPhase + Delta;
      Trunced     := Trunc( FPhase);
      FPhase      := FPhase - Trunced;
      if FPhase < 0
      then FPhase := FPhase + 1;

      FPosition   := FPhase + Clip( FInputs[ baselfo_phase] * FInputs[ baselfo_pmlevel], -1, 1);
      Trunced     := Trunc( FPosition);
      FPosition   := FPosition - Trunced;
      if FPosition < 0
      then FPosition := FPosition + 1;

      FOutputs[ baselfo_sync] := LogicToSignal( FPosition < 0.01);
    end;

    procedure   TModBaseLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/res/%f' , [ aPrefix, Name, FInputs [ baselfo_res]         ], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, LogicToSignal( FPosition < 0.5)], AppLocale));
    end;

    function    TModBaseLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        baselfo_frequency  : Result := 'frequency';
        baselfo_fm         : Result := 'fm';
        baselfo_phase      : Result := 'phase';
        baselfo_res        : Result := 'res';
        baselfo_mute       : Result := 'mute';
        baselfo_fmlevel    : Result := 'fmlevel';
        baselfo_pmlevel    : Result := 'pmlevel';
        baselfo_shape      : Result := 'shape';
        baselfo_range      : Result := 'range';
        baselfo_outputtype : Result := 'outputtype';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModBaseLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        baselfo_out  : Result := 'out';
        baselfo_sync : Result := 'sync';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModBaseTrigLfo = class( TModLfo)
  private
    FState   : TTrigOScState;
    FCounter : Integer;
    FDidRun  : Boolean;
  public
}

const

  basetriglfo_trig  = baselfo_res;
  basetriglfo_count = baselfo_phase;

  basetriglfo_ready = baselfo_sync;


    constructor TModBaseTrigLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModBaseTrigLfo.DoSlowTick; // override;
    var
      FOldCounter : Integer;
      Delta       : TSignal;
      Trunced     : Integer;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState    := toIdle;
        FDidRun   := False;
      end;

      case FState of

        toIdle :
          begin
            FPosition := 0.0;
            FCounter  := 0;
            if SignalToLogic( FInputs[ basetriglfo_trig])
            then begin
              FState  := toRunning;
              FDidRun := True;
            end;
          end;

        toRunning,
        toWaiting :
          begin
            // This is almost the inherited code - except - it ignores phase

            FShape := SignalToInt( FInputs[ baselfo_shape]);

            case SignalToInt( FInputs[ baselfo_range]) of
              0  : Delta := LookupLfoFastPhaseDelta  ( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
              1  : Delta := LookupLfoMediumPhaseDelta( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
              2  : Delta := LookupLfoSlowPhaseDelta  ( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
              3  : Delta := LookupLfoBPMPhaseDelta   ( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
              else Delta := LookupLfoLinPhaseDelta   ( FInputs[ baselfo_frequency] + FInputs[ baselfo_fm] * FInputs[ baselfo_fmlevel]);
            end;

            FPosition   := FPosition + Delta;
            Trunced     := Trunc( FPosition);
            FPosition   := FPosition - Trunced;
            if FPosition < 0
            then FPosition := FPosition + 1;

            if FState = toRunning
            then begin
              FOldCounter := FCounter;
              if AlmostEqual( FPosition, 0.5, Delta)
              then FCounter := FCounter + 1
              else if AlmostEqual( FPosition, 1.0, Delta)
              then FCounter := FCounter + 1;
              if FCounter <> FOldCounter
              then begin
                if FCounter >= FInputs[ basetriglfo_count]
                then FState := toDone
                else FState := toWaiting;
              end;
            end
            else begin
              if
                not AlmostEqual( FPosition, 0.5, Delta) and
                not AlmostEqual( FPosition, 1.0, Delta)
              then FState := toRunning;
            end;
          end;

        toDone :
          begin
            FPosition := 0.0;
            if not SignalToLogic( FInputs[ basetriglfo_trig])
            then FState := toIdle;
          end;
      end;
      FOutputs[ basetriglfo_ready] := LogicToSignal(( FState = toIdle) and FDidRun);
    end;

    function    TModBaseTrigLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        basetriglfo_trig  : Result := 'trig';
        basetriglfo_count : Result := 'count';
        else                Result := inherited;
      end;
    end;

    function    TModBaseTrigLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        basetriglfo_ready : Result := 'ready';
        else                Result := inherited;
      end;
    end;

    procedure   TModBaseTrigLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInputs[ basetriglfo_trig  ]                     ], AppLocale));
      aCallback( Self, Format( '%s/%s/ready/%f', [ aPrefix, Name, FOutputs[ basetriglfo_ready]                     ], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f'  , [ aPrefix, Name, LogicToSignal( FState in [ toRunning, toWaiting])], AppLocale));
    end;


{ ========
  TModCounterLfo = class( TMod)
  private
    FPhase : TSignal;
  protected
}

const

  counterlfo_frequency  = 0;
  counterlfo_fm         = 1;
  counterlfo_fmlevel    = 2;
  counterlfo_range      = 3;
  counterlfo_mute       = 4;
  counterlfo_outputtype = 5;

    procedure   TModCounterLfo.Wrapped; // virtual;
    begin
      // To be overridden
    end;

//  public

    constructor TModCounterLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModCounterLfo.DoSlowTick; // override;
    var
      Delta   : TSignal;
      DidWrap : Boolean;
    begin
      if ResetFlag
      then FPhase := 0;
      ResetFlag := False;

      case SignalToInt( FInputs[ counterlfo_range]) of
        0  : Delta := LookupLfoFastPhaseDelta  ( FInputs[ counterlfo_frequency] + FInputs[ counterlfo_fm] * FInputs[ counterlfo_fmlevel]);
        1  : Delta := LookupLfoMediumPhaseDelta( FInputs[ counterlfo_frequency] + FInputs[ counterlfo_fm] * FInputs[ counterlfo_fmlevel]);
        2  : Delta := LookupLfoSlowPhaseDelta  ( FInputs[ counterlfo_frequency] + FInputs[ counterlfo_fm] * FInputs[ counterlfo_fmlevel]);
        3  : Delta := LookupLfoBPMPhaseDelta   ( FInputs[ counterlfo_frequency] + FInputs[ counterlfo_fm] * FInputs[ counterlfo_fmlevel]);
        else Delta := LookupLfoLinPhaseDelta   ( FInputs[ counterlfo_frequency] + FInputs[ counterlfo_fm] * FInputs[ counterlfo_fmlevel]);
      end;

      FPhase  := FPhase - Delta;
      DidWrap := False;

      while FPhase < 0
      do begin
        if not DidWrap
        then begin
          Wrapped;
          DidWrap := True;
        end;
        FPhase := FPhase + 1.0;
      end;
    end;

    procedure   TModCounterLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/active/%f', [ aPrefix, Name, LogicToSignal( FPhase >= 0.5)], AppLocale));
    end;

    function    TModCounterLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        counterlfo_frequency : Result := 'frequency';
        counterlfo_fm        : Result := 'fm';
        counterlfo_fmlevel   : Result := 'fmlevel';
        counterlfo_range     : Result := 'range';
        else                   Result := NO_INPUT;
      end;
    end;


{ ========
  TModLfo = class( TModBaseLfo)
  public
}

    procedure   TModLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      case FShape of
        0  : // sine
          FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( LookupSine( FPosition), SignalToInt( FInputs[ baselfo_outputtype]));
        1  : // tri
          if FPosition <= 0.5
          then FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( 4.0 *         FPosition  - 1, SignalToInt( FInputs[ baselfo_outputtype]))
          else FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( 4.0 * ( 1.0 - FPosition) - 1, SignalToInt( FInputs[ baselfo_outputtype]));
        2  : // saw
          FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( 2.0 * FPosition - 1, SignalToInt( FInputs[ baselfo_outputtype]));
        3  : // square
          if FPosition < 0.5
          then FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType(  1.0, SignalToInt( FInputs[ baselfo_outputtype]))
          else FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( -1.0, SignalToInt( FInputs[ baselfo_outputtype]));
      end;
    end;


{ ========
  TModLfoTrig = class( TModbaseTrigLfo)
  public
}

    procedure   TModLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      case FShape of
        0  : // sine
          FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( LookupSine( FPosition), SignalToInt( FInputs[ baselfo_outputtype]));
        1  : // tri
          if FPosition <= 0.5
          then FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( 4.0 *         FPosition  - 1, SignalToInt( FInputs[ baselfo_outputtype]))
          else FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( 4.0 * ( 1.0 - FPosition) - 1, SignalToInt( FInputs[ baselfo_outputtype]));
        2  : // saw
          FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( 2.0 * FPosition - 1, SignalToInt( FInputs[ baselfo_outputtype]));
        3  : // square
          if FPosition < 0.5
          then FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType(  1.0, SignalToInt( FInputs[ baselfo_outputtype]));
          else FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( -1.0, SignalToInt( FInputs[ baselfo_outputtype]));
      end;
    end;


{ ========
  TModSquareLfo = class( TModLfo)
  public
}

const

  squarelfo_pwm      = 10;
  squarelfo_pwmlevel = 11;


    procedure   TModSquareLfo.SetDefaults;
    begin
      inherited;
      FDezipperMap := [
        baselfo_frequency,
        baselfo_fmlevel,
        baselfo_pmlevel,
        squarelfo_pwmlevel
      ];
    end;

    procedure   TModSquareLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if FPosition < Clip( 0.5 * ( 1 - FInputs[ squarelfo_pwm] * FInputs[ squarelfo_pwmlevel]), 0.01, 0.99)
      then FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType(  1.0, SignalToInt( FInputs[ baselfo_outputtype]))
      else FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( -1.0, SignalToInt( FInputs[ baselfo_outputtype]));
    end;

    function    TModSquareLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         squarelfo_pwm      : Result := 'pwm';
         squarelfo_pwmlevel : Result := 'pwmlevel';
        else                  Result := inherited;
      end;
    end;


{ ========
  TModTrigSquareLfo = class( TModBaseTrigLfo)
  public
}

const

  squarelfotrig_pwm      = 10;
  squarelfotrig_pwmlevel = 11;


    procedure   TModSquareLfoTrig.SetDefaults;
    begin
      inherited;
      FDezipperMap := [
        baselfo_frequency,
        baselfo_fmlevel,
        baselfo_pmlevel,
        squarelfotrig_pwmlevel
      ];
    end;

    procedure   TModSquareLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if FPosition < Clip( 0.5 * ( 1.0 - FInputs[ squarelfotrig_pwm] * FInputs[ squarelfotrig_pwmlevel]), 0.01, 0.99)
      then FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType(  1.0, SignalToInt( FInputs[ baselfo_outputtype]))
      else FOutputs[ baselfo_out] := SignalToMute( FInputs[ baselfo_mute]) * SignalForOutputType( -1.0, SignalToInt( FInputs[ baselfo_outputtype]));
    end;

    function    TModSquareLfoTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         squarelfo_pwm      : Result := 'pwm';
         squarelfo_pwmlevel : Result := 'pwmlevel';
        else                  Result := inherited;
      end;
    end;


{ ========
  TModRandSig = class( TModCounterLfo)
  private
    FStart  : TSignal;
    FTarget : TSignal;
  protected
}

const

  // lower input numbers are inherited

  randsig_mute       =  4;
  randsig_outputtype =  5;
  randsig_dist       =  6;
  randsig_distmod    =  7;
  randsig_distmodamt =  8;
  randsig_lambda     =  9;
  randsig_mode       = 10;

  randsig_out        =  0;

  randsig_mode_lin   =  0;
  randsig_mode_exp   =  1;

    procedure   TModRandSig.Wrapped; // override;
    var
      Dist   : TSignal;
      Lambda : TSignal;
      Mode   : Integer;
    begin
      Dist    := Clip( FInputs[ randsig_dist] + FInputs[ randsig_distmod] * FInputs[ randsig_distmodamt], 0, 1);
      Mode    := SignalToInt( FInputs[ randsig_mode]);
      FStart  := FTarget;
      if Mode = randsig_mode_exp
      then begin
        Lambda  := Max( FInputs[ randsig_lambda], 1.0000001);
        FTarget := Clip( Dist * ( RandomExpo( Lambda, False)) + ( 1 - Dist) * FStart, -1, 1);
      end
      else FTarget := Clip( Dist * ( 1 - 2 * Random) + ( 1 - Dist) * FStart, -1, 1);
    end;

//  public

    procedure   TModRandSig.SetDefaults; // override;
    begin
      FTarget := 2 * Random - 1;
      Wrapped; // Start with a wrap to make it start moving from the start.
      FDezipperMap :=
        FDezipperMap +
        [
          randsig_mute,
          randsig_dist,
          randsig_distmodamt
        ];
    end;

    procedure   TModRandSig.DoSlowTick; // override;
    var
      Value : TSignal;
    begin
      inherited;
      Value := FTarget + FPhase * ( FStart - FTarget); // Phase goes from 1 -> 0
      FOutputs[ randsig_out] := SignalToMute( FInputs[ randsig_mute]) * SignalForOutputType( Value, SignalToInt( FInputs[ randsig_outputtype]));
    end;

    function    TModRandSig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        randsig_mute       : Result := 'mute';
        randsig_outputtype : Result := 'outputtype';
        randsig_dist       : Result := 'dist';
        randsig_distmod    : Result := 'distmod';
        randsig_distmodamt : Result := 'distmodamt';
        randsig_lambda     : Result := 'lambda';
        randsig_mode       : Result := 'mode';
        else                 Result := inherited;
      end;
    end;

    function    TModRandSig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         randsig_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPulses = class( TMod)
  private
    FPhase         : TSignal;
    FState         : TPulsesState;
    FPulseCounter  : Integer;
    FPeriodCounter : Integer;
    FOldFrequency  : TSignal;
    FOldLambda     : TSignal;
    FOldDist       : TSignal;
  public
}

const

  pulses_frequency      =  0;
  pulses_range          =  1;
  pulses_dist           =  2;
  pulses_distmod        =  3;
  pulses_distmodamt     =  4;
  pulses_lambda         =  5;
  pulses_rndmode        =  6;
  pulses_syncmode       =  7;
  pulses_outputtype     =  8;

  pulses_out            =  0;

  pulses_rndmode_lin    =  0;
  pulses_rndmode_exp    =  1;

  pulses_syncmode_async =  0;
  pulses_syncmode_sync  =  1;


    constructor TModPulses.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModPulses.SetDefaults; // override;
    begin
      FDezipperMap :=
        [
          pulses_dist      ,
          pulses_distmodamt,
          pulses_lambda
        ];
    end;

    procedure   TModPulses.DoSlowTick; // override;
    var
      Delta      : TSignal;
      Frequency  : TSignal;
      Range      : Integer;
      Dist       : TSignal;
      Lambda     : TSignal;
      RndMode    : Integer;
      SyncMode   : Integer;
      OutputType : Integer;
      Wrap       : Boolean;
    begin
      Frequency  := Max( 1e-6, FInputs[ pulses_frequency]);
      Range      := SignalToInt( FInputs[ pulses_range]);
      Dist       := FInputs[ pulses_dist];
      Lambda     := Max( FInputs[ pulses_lambda], 1.0000001);
      RndMode    := SignalToInt( FInputs[ pulses_rndmode]);
      SyncMode   := SignalToInt( FInputs[ pulses_syncmode]);
      OutputType := SignalToInt( FInputs[ pulses_outputtype]);

      if
        ResetFlag or
        ( FOldFrequency <> Frequency) or
        ( FOldLambda    <> Lambda   ) or
        ( FOldDist      <> Dist     )
      then begin
        ResetFlag := False;
        FPhase    := 0.0;
        FState    := psIdle;
      end;
      FOldFrequency := Frequency;
      FOldLambda    := Lambda;
      FOldDist      := Dist;

      Dist := Clip( Dist + FInputs[ pulses_distmod] * FInputs[ pulses_distmodamt], 0, 1);

      case Range of
        0  : Delta := LookupLfoFastPhaseDelta  ( Frequency);
        1  : Delta := LookupLfoMediumPhaseDelta( Frequency);
        2  : Delta := LookupLfoSlowPhaseDelta  ( Frequency);
        3  : Delta := LookupLfoBPMPhaseDelta   ( Frequency);
        else Delta := LookupLfoLinPhaseDelta   ( Frequency);
      end;

      FPhase := FPhase + Delta;
      Wrap   := False;
      while FPhase > 1.0
      do begin
        FPhase := FPhase - 1.0;
        Wrap   := True;
      end;

      if ( SyncMode = pulses_syncmode_async) or Wrap
      then begin
        case FState of

          psIdle : begin
            FPulseCounter := Trunc( SlowTimeToSampleCount( 1e-3)); // Pulse for    1 ms
            FLedCounter   := Trunc( SlowTimeToSampleCount( 1e-1)); // Led on for 100 ms
            if RndMode = pulses_rndmode_exp
            then begin
              if SyncMode = pulses_syncmode_async
              then FPeriodCounter := Max( 1, Trunc(( Control_Rate / Frequency ) * Dist * RandomExpo( Lambda, True)))
              else FPeriodCounter := Trunc( 25 * Dist * RandomExpo( Lambda, True));
            end
            else begin
              if SyncMode = pulses_syncmode_async
              then FPeriodCounter := Max( 1, Trunc(( Control_Rate / Frequency ) * Dist * Random))
              else FPeriodCounter := Trunc( 25 * Dist * Random);
            end;
            FState := psPulsing;
          end;

          else begin
            if FPeriodCounter > 0
            then Dec( FPeriodCounter)
            else FState := psIdle;
          end;
        end;
      end;

      if FLedCounter > 0
      then Dec( FLedCounter);

      if FState = psPulsing
      then begin
        if FPulseCounter > 0
        then Dec( FPulseCounter)
        else FState := psWaiting;
      end;

      FOutputs[ pulses_out] := SignalForOutputType( LogicToSignal( FState = psPulsing), OutputType);
    end;

    procedure   TModPulses.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, LogicToSignal( FLedCounter > 0)], AppLocale));
    end;

    function    TModPulses.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pulses_frequency  : Result := 'frequency';
        pulses_range      : Result := 'range';
        pulses_dist       : Result := 'dist';
        pulses_distmod    : Result := 'distmod';
        pulses_distmodamt : Result := 'distmodamt';
        pulses_lambda     : Result := 'lambda';
        pulses_rndmode    : Result := 'rndmode';
        pulses_syncmode   : Result := 'syncmode';
        pulses_outputtype : Result := 'syncmode';
        else                Result := inherited;
      end;
    end;

    function    TModPulses.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pulses_out : Result := 'out';
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModRandomWalkLfo = class( TModLfo)
  private
    FState : TRWState;
    FX     : TSignal;
    FY     : TSignal;
    FZ     : TSignal;
  public
}

const

  randomwalklfo_range       =  0;
  randomwalklfo_frequency   =  1;
  randomwalklfo_fm          =  2;
  randomwalklfo_fmlevel     =  3;
  randomwalklfo_distance    =  4;
  randomwalklfo_mute        =  5;
  randomwalklfo_color       =  6;
  randomwalklfo_colormod    =  7;
  randomwalklfo_colormodamt =  8;
  randomwalklfo_outputtype  =  9;
  randomwalklfo_lambda      = 10;
  randomwalklfo_mode        = 11;

  randomwalklfo_outx        =  0;
  randomwalklfo_outy        =  1;
  randomwalklfo_outz        =  2;

  randomwalklfo_mode_lin    =  0;
  randomwalklfo_mode_exp    =  0;


    constructor TModRandomWalkLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModRandomWalkLfo.SetDefaults; // override;
    begin
      FDezipperMap := [
        randomwalklfo_frequency,
        randomwalklfo_fmlevel,
        randomwalklfo_distance,
        randomwalklfo_mute,
        randomwalklfo_color,
        randomwalklfo_colormodamt
      ];
      FInputs[ randomwalklfo_fm      ] := 0;
      FInputs[ randomwalklfo_colormod] := 0;
    end;

    procedure   TModRandomWalkLfo.DoSlowTick; // override;
    var
      Distance : TSignal;
      Delta    : TSignal;
      Color    : TSignal;
      Mute     : TSignal;
      OutType  : Integer;
      Lambda   : TSignal;
      Mode     : Integer;
    begin

      case SignalToInt( FInputs[ randomwalklfo_range]) of
        0  : Delta := LookupLfoFastPhaseDelta  ( FInputs[ randomwalklfo_frequency] + Clip( FInputs[ randomwalklfo_fm] * FInputs[ randomwalklfo_fmlevel], -1, 1));
        1  : Delta := LookupLfoMediumPhaseDelta( FInputs[ randomwalklfo_frequency] + Clip( FInputs[ randomwalklfo_fm] * FInputs[ randomwalklfo_fmlevel], -1, 1));
        2  : Delta := LookupLfoSlowPhaseDelta  ( FInputs[ randomwalklfo_frequency] + Clip( FInputs[ randomwalklfo_fm] * FInputs[ randomwalklfo_fmlevel], -1, 1));
        3  : Delta := LookupLfoBPMPhaseDelta   ( FInputs[ randomwalklfo_frequency] + Clip( FInputs[ randomwalklfo_fm] * FInputs[ randomwalklfo_fmlevel], -1, 1));
        else Delta := LookupLfoLinPhaseDelta   ( FInputs[ randomwalklfo_frequency] + Clip( FInputs[ randomwalklfo_fm] * FInputs[ randomwalklfo_fmlevel], -1, 1));
      end;

      FPosition := FMod( FPosition + Delta, 1.0);

      case FState of
        trwIdle :
          begin
            if FPosition >= 0.5
            then begin
              FState   := trwFired;
              Lambda   := Max( FInputs[ randomwalklfo_lambda], 1.0000001);
              Distance := FInputs[ randomwalklfo_distance];
              Color    := Clip( FInputs[ randomwalklfo_color] + FInputs[ randomwalklfo_colormod] * FInputs[ randomwalklfo_colormodamt], 0, 1);
              Mute     := SignalToMute( FInputs[ randomwalklfo_mute]);
              OutType  := SignalToInt( FInputs[ randomwalklfo_outputtype]);
              Mode     := SignalToInt( FInputs[ randomwalklfo_mode]);

              if Mode = randomwalklfo_mode_exp
              then begin
                FX := ( 1 - Color) * FX + Color * ( FX + RandomExpo( Lambda, False) * Distance);
                FY := ( 1 - Color) * FY + Color * ( FY + RandomExpo( Lambda, False) * Distance);
                FZ := ( 1 - Color) * FZ + Color * ( FZ + RandomExpo( Lambda, False) * Distance);
              end
              else begin
                FX := ( 1 - Color) * FX + Color * ( FX + ( 1.0 - 2.0 * Random) * Distance);
                FY := ( 1 - Color) * FY + Color * ( FY + ( 1.0 - 2.0 * Random) * Distance);
                FZ := ( 1 - Color) * FZ + Color * ( FZ + ( 1.0 - 2.0 * Random) * Distance);
              end;

              if FX < -1 then FX := - 2 - FX else if FX > 1 then FX := 2 - FX;
              if FY < -1 then FY := - 2 - FY else if FY > 1 then FY := 2 - FY;
              if FZ < -1 then FZ := - 2 - FZ else if FZ > 1 then FZ := 2 - FZ;

              FOutputs[ randomwalklfo_outx] := Mute * SignalForOutputType( FX, OutType);
              FOutputs[ randomwalklfo_outy] := Mute * SignalForOutputType( FY, OutType);
              FOutputs[ randomwalklfo_outz] := Mute * SignalForOutputType( FZ, OutType);
            end;
          end;

        trwFired :
          begin
            if FPosition < 0.5
            then FState := trwIdle;
          end;
      end;
    end;

    procedure   TModRandomWalkLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/outx/%f', [ aPrefix, Name, FX], AppLocale));
      aCallback( Self, Format( '%s/%s/outy/%f', [ aPrefix, Name, FY], AppLocale));
      aCallback( Self, Format( '%s/%s/outz/%f', [ aPrefix, Name, FZ], AppLocale));
    end;

    function    TModRandomWalkLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        randomwalklfo_range       : Result := 'range';
        randomwalklfo_frequency   : Result := 'frequency';
        randomwalklfo_fm          : Result := 'fm';
        randomwalklfo_fmlevel     : Result := 'fmlevel';
        randomwalklfo_distance    : Result := 'distance';
        randomwalklfo_mute        : Result := 'mute';
        randomwalklfo_color       : Result := 'color';
        randomwalklfo_colormod    : Result := 'colormod';
        randomwalklfo_colormodamt : Result := 'colormodamt';
        randomwalklfo_outputtype  : Result := 'outputtype';
        randomwalklfo_lambda      : Result := 'lambda';
        randomwalklfo_mode        : Result := 'mode';
        else                        Result := NO_INPUT;
      end;
    end;

    function    TModRandomWalkLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        randomwalklfo_outx : Result := 'outx';
        randomwalklfo_outy : Result := 'outy';
        randomwalklfo_outz : Result := 'outz';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoise = class( TMod)
  public
}

const

  noise_color       = 0;
  noise_colormod    = 1;
  noise_colormodamt = 2;
  noise_mute        = 3;

  noise_out         = 0;


    constructor TModNoise.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModNoise.SetDefaults; // override;
    begin
      FDezipperMap := [
        noise_color,
        noise_colormodamt,
        noise_mute
      ];
    end;

    procedure   TModNoise.DoTick; // override;
    var
      Color : TSignal;
    begin
      Color := Clip( FInputs[ noise_color] + FInputs[ noise_colormod] * FInputs[ noise_colormodamt], 0, 1);
      FOutputs[ noise_out] := SignalToMute( FInputs[ noise_mute]) * (( 1 - Color) * FOutputs[ noise_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
    end;

    function    TModNoise.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noise_color       : Result := 'color';
        noise_colormod    : Result := 'colormod';
        noise_colormodamt : Result := 'colormodamt';
        noise_mute        : Result := 'mute';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModNoise.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noise_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseTrig = class( TMod)
  private
    FLastInput: TSignal;
  public
}

const

  noisetrig_color       = 0;
  noisetrig_colormod    = 1;
  noisetrig_colormodamt = 2;
  noisetrig_trig        = 3;
  noisetrig_mute        = 4;

  noisetrig_out         = 0;


    constructor TModNoiseTrig.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModNoiseTrig.SetDefaults; // override;
    begin
      FDezipperMap := [
        noisetrig_color,
        noisetrig_colormodamt,
        noisetrig_mute
      ];
    end;

    procedure   TModNoiseTrig.DoTick; // override;
    var
      Color : TSignal;
      Trig  : Boolean;
    begin
      Trig  := SignalToLogic( FInputs[ noisetrig_trig]);
      if not FLastTrig and Trig
      then begin
        Color := Clip( FInputs[ noisetrig_color] + FInputs[ noisetrig_colormod] * FInputs[ noisetrig_colormodamt], 0, 1);
        FOutputs[ noisetrig_out] := SignalToMute( FInputs[ noisetrig_mute]) * (( 1 - Color) * FOutputs[ noisetrig_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
        FLastTrig := Trig;
      end
      else if FLastTrig and not Trig
      then FLastTrig := Trig;
    end;

    procedure   TModNoiseTrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ noisetrig_trig]], AppLocale));
    end;

    function    TModNoiseTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noisetrig_color       : Result := 'color';
        noisetrig_colormod    : Result := 'colormod';
        noisetrig_colormodamt : Result := 'colormodamt';
        noisetrig_trig        : Result := 'trig';
        noisetrig_mute        : Result := 'mute';
        else                    Result := NO_INPUT;
      end;
    end;

    function    TModNoiseTrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noisetrig_out : Result := 'out';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseLfo = class( TMod)
  private
    FOutValue : TSignal;
  public
}

const

  noiselfo_color       = 0;
  noiselfo_colormod    = 1;
  noiselfo_colormodamt = 2;
  noiselfo_mute        = 3;
  noiselfo_outputtype  = 4;

  noiselfo_out         = 0;


    constructor TModNoiseLfo.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModNoiseLfo.SetDefaults; // override;
    begin
      FDezipperMap := [
        noiselfo_color,
        noiselfo_colormodamt,
        noiselfo_mute
      ];
    end;

    procedure   TModNoiseLfo.DoSlowTick; // override;
    var
      Color : TSignal;
    begin
      Color := Clip( FInputs[ noiselfo_color] + FInputs[ noiselfo_colormod] * FInputs[ noiselfo_colormodamt], 0, 1);
      FOutValue := (( 1 - Color) * FOutputs[ noiselfo_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
      FOutputs[ noiselfo_out] := SignalToMute( FInputs[ noiselfo_mute]) * SignalForOutputType( FOutValue, SignalToInt( FInputs[ noiselfo_outputtype]));
    end;

    procedure   TModNoiseLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutValue], AppLocale));
    end;

    function    TModNoiseLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noiselfo_color       : Result := 'color';
        noiselfo_colormod    : Result := 'colormod';
        noiselfo_colormodamt : Result := 'colormodamt';
        noiselfo_mute        : Result := 'mute';
        noiselfo_outputtype  : Result := 'outputtype';
        else                   Result := NO_INPUT;
      end;
    end;

    function    TModNoiseLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noiselfo_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseLfoTrig = class( TMod)
  private
    FLastTrig : Boolean;
    FOutValue : TSignal;
  public
}

const

  noiselfotrig_color       = 0;
  noiselfotrig_colormod    = 1;
  noiselfotrig_colormodamt = 2;
  noiselfotrig_trig        = 3;
  noiselfotrig_mute        = 4;
  noiselfotrig_outputtype  = 5;

  noiselfotrig_out         = 0;


    constructor TModNoiseLfoTrig.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModNoiseLfoTrig.SetDefaults; // override;
    begin
      FDezipperMap := [
        noiselfotrig_color,
        noiselfotrig_colormodamt,
        noiselfotrig_mute
      ];
    end;

    procedure   TModNoiseLfoTrig.DoSlowTick; // override;
    var
      Color : TSignal;
      Trig  : Boolean;
    begin
      Trig  := SignalToLogic( FInputs[ noiselfotrig_trig]);
      if not FLastTrig and Trig
      then begin
        Color := Clip( FInputs[ noiselfotrig_color] + FInputs[ noiselfotrig_colormod] * FInputs[ noiselfotrig_colormodamt], 0, 1);
        FOutValue := (( 1 - Color) * FOutputs[ noiselfotrig_out] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
        FOutputs[ noiselfotrig_out] :=  SignalToMute( FInputs[ noiselfotrig_mute]) * SignalForOutputType( FOutValue, SignalToInt( FInputs[ noiselfotrig_outputtype]));
        FLastTrig := Trig;
      end
      else if FLastTrig and not Trig
      then FLastTrig := Trig;
    end;

    procedure   TModNoiseLfoTrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs [ noiselfotrig_trig]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutValue                   ], AppLocale));
    end;

    function    TModNoiseLfoTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noiselfotrig_color       : Result := 'color';
        noiselfotrig_colormod    : Result := 'colormod';
        noiselfotrig_colormodamt : Result := 'colormodamt';
        noiselfotrig_trig        : Result := 'trig';
        noiselfotrig_mute        : Result := 'mute';
        noiselfotrig_outputtype  : Result := 'outputtype';
        else                       Result := NO_INPUT;
      end;
    end;

    function    TModNoiseLfoTrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        noiselfotrig_out : Result := 'out';
        else Result      := NO_OUTPUT;
      end;
    end;


{ ========
  TModAverage = class( TMod)
  private
    FAlpha : TSignal;  // Alpha = System_Rate / ( 2 * PI * Freq + System_Rate) ... Fc := (log(b1))/(-2.0 * M_PI) * Fs; (*b1 < 0.0432139182637722579772*)
  public
}

const

  average_in          = 0;
  average_alpha       = 1;
  average_alphamod    = 2;
  average_alphamodamt = 3;

  average_lp          = 0;
  average_hp          = 1;


    constructor TModAverage.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModAverage.SetDefaults; // override;
    begin
      FDezipperMap := [
        average_alpha,
        average_alphamodamt
      ];
    end;

    procedure   TModAverage.DoSlowTick; // override;
    begin
      FAlpha := Normalize( ExpAttack( FInputs[ average_alpha] + FInputs[ average_alphamod] * FInputs[ average_alphamodamt], 1, 0, 1));
    end;

    procedure   TModAverage.DoTick; // override;
    var
      Inp   : TSignal;
      Outp  : TSignal;
      Delta : TSignal;
    begin
      Inp   := FInputs [ average_in];
      Outp  := FOutputs[ average_lp];
      Delta := Inp - Outp;
      Outp  := Outp + FAlpha * Delta;
      FOutputs[ average_lp] := Normalize( Outp );
      FOutputs[ average_hp] := Normalize( Delta);
    end;

    function    TModAverage.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        average_in          : Result := 'in';
        average_alpha       : Result := 'alpha';
        average_alphamod    : Result := 'alphamod';
        average_alphamodamt : Result := 'alphamodamt';
        else                  Result := NO_INPUT;
      end;
    end;

    function    TModAverage.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        average_lp : Result := 'lp';
        average_hp : Result := 'hp';
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSVF = class( TMod)
  private
    FFreq  : TSignal;
    FDamp  : TSignal;
    FDrive : TSignal;
  public
}

    (*
      from    : http://www.musicdsp.org/archive.php?classid=3#92
      also see: http://www.music.mcgill.ca/~ich/classes/FiltersChap2.pdf

      input  = input buffer;
      output = output buffer;
      fs     = sampling frequency;
      fc     = cutoff frequency normally something like:
               440.0*pow(2.0, (midi_note - 69.0)/12.0);                 :: We use signals rather, Frequency = Signal * SampleRate / 2
      res    = resonance 0 to 1;
      drive  = internal distortion 0 to 0.1
      freq   = 2.0*sin(PI*MIN(0.25, fc/(fs*2)));  // the fs*2 is because it's double sampled
      damp   = MIN(2.0*(1.0 - pow(res, 0.25)), MIN(2.0, 2.0/freq - freq*0.5));
      notch  = notch output
      low    = low pass output
      high   = high pass output
      band   = band pass output
      peak   = peaking output = low - high
      --
      double sampled svf loop:
      for (i=0; i<numSamples; i++)
      {
        in    = input[i];
        notch = in - damp*band;
        low   = low + freq*band;
        high  = notch - low;
        band  = freq*high + band - drive*band*band*band;
        out   = 0.5*(notch or low or high or band or peak);
        notch = in - damp*band;
        low   = low + freq*band;
        high  = notch - low;
        band  = freq*high + band - drive*band*band*band;
        out  += 0.5*(same out as above);
        output[i] = out;
      }
    *)

const

  svf_in         = 0;
  svf_frequency  = 1;
  svf_q          = 2;
  svf_distortion = 3;
  svf_fm         = 4;
  svf_fmlevel    = 5;
  svf_inlevel    = 6;

  svf_lp         = 0;
  svf_bp         = 1;
  svf_hp         = 2;
  svf_br         = 3;


    constructor TModSVF.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModSVF.SetDefaults; // override;
    begin
      FDezipperMap := [
        svf_frequency,
        svf_q,
        svf_distortion,
        svf_fmlevel,
        svf_inlevel
      ];
    end;

    procedure   TModSVF.DoSlowTick; // override;
    begin
      // NOTE: this assumes an SVF sampled at twice the system rate

      FFreq := LookupSvfFrequency( FInputs[ svf_frequency] + FInputs[ svf_fm] * FInputs[ svf_fmlevel]);
      FDamp := 1 - Clip( FInputs[ svf_q], 0.01, 1);

      // see: http://www.earlevel.com/main/2003/03/02/the-digital-state-variable-filter/
      // instability when FFreq aproaches 1

      if FFreq > 0.99
      then FFreq := 0.99;


//      else if FFreq <= 0.0001
//      then FDamp := Min( 2.0 * ( 1.0 - Power( Q, 0.25)), 2.0)
//      else FDamp := Min( 2.0 * ( 1.0 - Power( Q, 0.25)), Min( 2.0, 2.0 / FFreq - FFreq * 0.5));

      FDrive := Clip( FInputs[ svf_distortion] / 10, 0.0, 0.1);
    end;

    procedure   TModSVF.DoTick; // override;
    var
      INP : TSignal;
      HP  : TSignal;
      BP  : TSignal;
      LP  : TSignal;
      BR  : TSignal;
    begin
      INP := FInputs[ svf_in] * FInputs[ svf_inlevel];
      LP  := FOutputs[ svf_lp];
      BP  := FOutputs[ svf_bp];

      BR := INP - FDamp * BP;
      LP := LP  + FFreq * BP;
      HP := BR  - LP;
      BP := FFreq * HP + BP - FDrive * BP * BP * BP; // with 0 <= Drive <= 0.1
      BR := INP - FDamp * BP;
      LP := LP  + FFreq * BP;
      HP := BR  - LP;
      BP := FFreq * HP + BP - FDrive * BP * BP * BP;

      FOutputs[ svf_lp] := Normalize( LP);
      FOutputs[ svf_bp] := Normalize( BP);
      FOutputs[ svf_hp] := Normalize( HP);
      FOutputs[ svf_br] := Normalize( BR);
    end;

    function    TModSVF.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        svf_in         : Result := 'in';
        svf_frequency  : Result := 'frequency';
        svf_q          : Result := 'q';
        svf_distortion : Result := 'distortion';
        svf_fm         : Result := 'fm';
        svf_fmlevel    : Result := 'fmlevel';
        svf_inlevel    : Result := 'inlevel';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModSVF.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        svf_lp : Result := 'lp';
        svf_bp : Result := 'bp';
        svf_hp : Result := 'hp';
        svf_br : Result := 'br';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPinkFilter = class( TMod)
  private
    FB0 : TSignal;
    FB1 : TSignal;
    FB2 : TSignal;
    FB3 : TSignal;
    FB4 : TSignal;
    FB5 : TSignal;
    FB6 : TSignal;
    FG  : TSignal;
  public
}

const

  pinkfilter_in   = 0;
  pinkfilter_mode = 1;
  pinkfilter_gain = 2;

  pinkfilter_out  = 0;



    constructor TModPinkFilter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModPinkFilter.SetDefaults; // override;
    begin
      FDezipperMap := [
        pinkfilter_gain
      ];
      FInputs[ pinkfilter_in] := 0;
      FB0 := 0.0;
      FB1 := 0.0;
      FB2 := 0.0;
      FB3 := 0.0;
      FB4 := 0.0;
      FB5 := 0.0;
      FB6 := 0.0;
      SampleRateChanged;
    end;

    procedure   TModPinkFilter.DoTick; // override;
    // http://www.musicdsp.org/files/pink.txt
    var
      White : TSignal;
    begin
      White := FInputs[ pinkfilter_in];
      case SignalToInt( FInputs[ pinkfilter_mode]) of
        0 : begin
          FB0 := 0.99886 * FB0 + White * 0.0555179;
          FB1 := 0.99332 * FB1 + White * 0.0750759;
          FB2 := 0.96900 * FB2 + White * 0.1538520;
          FB3 := 0.86650 * FB3 + White * 0.3104856;
          FB4 := 0.55000 * FB4 + White * 0.5329522;
          FB5 := -0.7616 * FB5 - White * 0.0168980;
          FOutputs[ pinkfilter_out] := FInputs[ pinkfilter_gain] * FG * ( FB0 + FB1 + FB2 + FB3 + FB4 + FB5 + FB6 + White * 0.5362);
          FB6 := White * 0.115926;
        end;
        else begin
          FB0 := 0.99765 * FB0 + White * 0.0990460;
          FB1 := 0.96300 * FB1 + White * 0.2965164;
          FB2 := 0.57000 * FB2 + White * 1.0526913;
          FOutputs[ pinkfilter_out] := FInputs[ pinkfilter_gain] * FG * ( FB0 + FB1 + FB2 + White * 0.1848);
        end
      end;
    end;

    procedure   TModPinkFilter.SampleRateChanged; // override;
    begin
      FG  := dBToSignal( -13) * ( 44100 / System_Rate);
    end;

    function    TModPinkFilter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pinkfilter_in   : Result := 'in';
        pinkfilter_mode : Result := 'mode';
        pinkfilter_gain : Result := 'gain';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModPinkFilter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pinkfilter_out : Result := 'out';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModResonator = class( TMod)
  // http://www.katjaas.nl/complexintegrator/complexresonator.html
  private
    FOldRe : TSignal;
    FOldIm : TSignal;
  public
}

const

  resonator_freq      =  0;
  resonator_decay     =  1;
  resonator_in        =  2;
  resonator_fm        =  3;
  resonator_fmlevel   =  4;
  resonator_frequency =  5;
  resonator_cents     =  6;
  resonator_pm        =  7;
  resonator_pmlevel   =  8;
  resonator_dm        =  9;
  resonator_dmlevel   = 10;
  resonator_click     = 11;

  resonator_out       =  0;

    constructor TModResonator.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModResonator.SetDefaults; // override;
    begin
      FDezipperMap := [
        resonator_decay    ,
        resonator_fmlevel  ,
        resonator_frequency,
        resonator_cents    ,
        resonator_pmlevel  ,
        resonator_dmlevel  ,
        resonator_click
      ];
    end;

    procedure   TModResonator.DoTick; // override;
    var
      aFreq   : TSignal;
      anAngle : TSignal;
      aCos    : TSignal;
      aSin    : TSignal;
      inRe    : TSignal;
      inIm    : TSignal;
      aDecay  : TSignal;
      aRadius : TSignal;
    begin
      aFreq   := LookupFrequency( FInputs[ resonator_freq] + FInputs[ resonator_frequency] + FInputs[ resonator_fm] * FInputs[ resonator_fmlevel] + FInputs[ resonator_cents]);
      anAngle := TWO_PI * ( aFreq / System_Rate + FInputs[ resonator_pm] * FInputs[ resonator_pmlevel]);
      aDecay  := FInputs[ resonator_decay] * ( 1 + FInputs[ resonator_dm] * FInputs[ resonator_dmlevel]);
      if aDecay <= 0
      then aDecay := 10e-6;                                             // decay time is seconds - so that's 10 s minimal time
      aRadius := exp( - 1 / ( System_Rate * aDecay));
      SinCos( anAngle, aSin, aCos);
      inRe    := Normalize( aRadius * FOldRe + FInputs[ resonator_in]); // Use Re. for input
      inIm    := Normalize( aRadius * FOldIm);
      FOldRe  := aCos * inRe - aSin * inIm;
      FOldIm  := aSin * inRe + aCos * inIm;
      FOutputs[ resonator_out] := FOldIm + FInputs[ resonator_click] * ( FOldRe - FOldIm);
    end;

    function    TModResonator.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        resonator_freq      : Result := 'freq';
        resonator_decay     : Result := 'decay';
        resonator_in        : Result := 'in';
        resonator_fm        : Result := 'fm';
        resonator_fmlevel   : Result := 'fmlevel';
        resonator_frequency : Result := 'frequency';
        resonator_cents     : Result := 'cents';
        resonator_pm        : Result := 'pm';
        resonator_pmlevel   : Result := 'pmlevel';
        resonator_dm        : Result := 'dm';
        resonator_dmlevel   : Result := 'dmlevel';
        resonator_click     : Result := 'click';
        else                  Result := NO_INPUT;
      end;
    end;

    function    TModResonator.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        resonator_out : Result := 'out';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModFreqShifter = class( TMod)
  // https://github.com/swh/ladspa/blob/master/bode_shifter_1431.xml
  private
    FDelay : array[ 0 .. 255] of TSignal;
    FDPtr  : Integer;
    FPhase : TSignal;
    FDelta : TSignal;
  public
}

const

  freqshifter_in         = 0;
  freqshifter_shift      = 1;
  freqshifter_shiftmod   = 2;
  freqshifter_shiftdepth = 3;

  freqshifter_up    = 0;
  freqshifter_down  = 1;

  XCoeffs : array[ 0 .. 99] of TSignal = (
    +0.0008103736, +0.0008457886, +0.0009017196, +0.0009793364,
    +0.0010798341, +0.0012044365, +0.0013544008, +0.0015310235,
    +0.0017356466, +0.0019696659, +0.0022345404, +0.0025318040,
    +0.0028630784, +0.0032300896, +0.0036346867, +0.0040788644,
    +0.0045647903, +0.0050948365, +0.0056716186, +0.0062980419,
    +0.0069773575, +0.0077132300, +0.0085098208, +0.0093718901,
    +0.0103049226, +0.0113152847, +0.0124104218, +0.0135991079,
    +0.0148917649, +0.0163008758, +0.0178415242, +0.0195321089,
    +0.0213953037, +0.0234593652, +0.0257599469, +0.0283426636,
    +0.0312667947, +0.0346107648, +0.0384804823, +0.0430224431,
    +0.0484451086, +0.0550553725, +0.0633242001, +0.0740128560,
    +0.0884368322, +0.1090816773, +0.1412745301, +0.1988673273,
    +0.3326528346, +0.9997730178, -0.9997730178, -0.3326528346,
    -0.1988673273, -0.1412745301, -0.1090816773, -0.0884368322,
    -0.0740128560, -0.0633242001, -0.0550553725, -0.0484451086,
    -0.0430224431, -0.0384804823, -0.0346107648, -0.0312667947,
    -0.0283426636, -0.0257599469, -0.0234593652, -0.0213953037,
    -0.0195321089, -0.0178415242, -0.0163008758, -0.0148917649,
    -0.0135991079, -0.0124104218, -0.0113152847, -0.0103049226,
    -0.0093718901, -0.0085098208, -0.0077132300, -0.0069773575,
    -0.0062980419, -0.0056716186, -0.0050948365, -0.0045647903,
    -0.0040788644, -0.0036346867, -0.0032300896, -0.0028630784,
    -0.0025318040, -0.0022345404, -0.0019696659, -0.0017356466,
    -0.0015310235, -0.0013544008, -0.0012044365, -0.0010798341,
    -0.0009793364, -0.0009017196, -0.0008457886, -0.0008103736
  );

    constructor TModFreqShifter.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModFreqShifter.SetDefaults; // override;
    begin
      FDezipperMap := [
        freqshifter_shift
      ];
    end;

    procedure   TModFreqShifter.DoSlowTick; // override;
    begin
      FDelta := Clip( LookupFrequencyCubed( FInputs[ freqshifter_shift]) + LookupFrequencyLin( FInputs[ freqshifter_shiftmod] * FInputs[ freqshifter_shiftdepth]), 0.0, 10000.0)  / System_Rate;
    end;

    procedure   TModFreqShifter.DoTick; // override;
    var
      i       : Integer;
      d       : Word;
      hilb    : TSignal;
      rm1     : TSignal;
      rm2     : TSignal;
      Trunced : Integer;
    begin
      FDelay[ FDPtr] := Normalize( FInputs[ freqshifter_in]);

      // hilb := 0.0;
      // for i := Low( XCoeffs) to High( XCoeffs)
      // do hilb := hilb + XCoeffs[ i] * FDelay[ ( FDPtr - 2 * i) and 255];

      hilb := 0;
      d    := FDPtr;
      for i := Low( XCoeffs) to High( XCoeffs)
      do begin
        hilb := hilb + XCoeffs[ i] * FDelay[ d];
        d    := ( d + 254) and 255;
      end;

      FPhase  := FPhase + FDelta;
      Trunced := Trunc( FPhase);
      FPhase  := FPhase - Trunced;
      if FPhase < 0
      then FPhase := FPhase + 1;

      rm1 := hilb * 0.63661978 * LookupSine( FPhase);
      rm2 := FDelay[ ( FDPtr - 99) and 255] * LookupCosine( FPhase);

      FOutputs[ freqshifter_down] := ( rm2 - rm1) * 0.5;
      FOutputs[ freqshifter_up  ] := ( rm2 + rm1) * 0.5;

      FDPtr := ( FDPtr + 1) and 255;
    end;

    function    TModFreqShifter.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        freqshifter_in         : Result := 'in';
        freqshifter_shift      : Result := 'shift';
        freqshifter_shiftmod   : Result := 'shiftmod';
        freqshifter_shiftdepth : Result := 'shiftdepth';
        else                     Result := NO_INPUT;
      end;
    end;

    function    TModFreqShifter.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        freqshifter_up   : Result := 'up';
        freqshifter_down : Result := 'down';
        else               Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvAR = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
  public
}

const

  envar_trig    = 0;
  envar_attack  = 1;
  envar_release = 2;
  envar_gain    = 3;
  envar_sigin   = 4;
  envar_mute    = 5;

  envar_out     = 0;
  envar_sigout  = 1;


    constructor TModEnvAR.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModEnvAR.SetDefaults; // override;
    begin
      inherited;
      FInputs[ envar_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FDezipperMap := [
        envar_attack,
        envar_release
      ];
    end;

    procedure   TModEnvAR.DoTick; // override;
    var
      R : TSignal;
    begin
      if ResetFlag
      then begin
        ResetFlag            := False;
        FState               := arIdle;
        FValue               := 0;
        FOutputs[ envar_out] := 0;
      end;

      case FState of

        arIdle :

          if SignalToLogic( FInputs[ envar_trig])
          then begin
            if FInputs[ envar_attack] <= 0
            then FCoeffAttack := 1
            else FCoeffAttack := 1 / Max( 1, TimeToSampleCount( FInputs[ envar_attack])); // Linear attack
            if FInputs[ envar_release] <= 0
            then FCoeffRelease := -1
            else FCoeffRelease := ( Ln( 0.001) - Ln( 1.000)) / Max( 1, TimeToSampleCount( FInputs[ envar_release])); // Exp release
            FState := arA;
          end;

        arA :
          begin
            R := FValue;
            if R < 0.001
            then R := 0.001;
            R := R + FCoeffAttack;
            if R > 0.999
            then begin
              FValue := 1.0;
              FState := arR;
            end
            else FValue := R;
            case SignalToInt( FInputs[ envar_mute]) of
              1  : FOutputs[ envar_out] := 0;
              2  : FOutputs[ envar_out] := 1;
              else FOutputs[ envar_out] := FValue * FInputs[ envar_gain];
            end;
          end;

        arR :
          begin
            R := FValue;
            R := R * ( 1 + FCoeffRelease);
            if R < 0.001
            then begin
              FValue := 0.0;
              FState := arWait;
            end
            else FValue := R;
            case SignalToInt( FInputs[ envar_mute]) of
              1  : FOutputs[ envar_out] := 0;
              2  : FOutputs[ envar_out] := 1;
              else FOutputs[ envar_out] := FValue * FInputs[ envar_gain];
            end;
          end;

        arWait :
          if not SignalToLogic( FInputs[ envar_trig])
          then FState := arIdle;

      end;
      case SignalToInt( FInputs[ envar_mute]) of
        1  : FOutputs[ envar_sigout] := 0;
        2  : FOutputs[ envar_sigout] := FInputs[ envar_sigin];
        else FOutputs[ envar_sigout] := FInputs[ envar_sigin] * FOutputs[ envar_out];
      end;
    end;

    procedure   TModEnvAR.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ envar_trig]], AppLocale));
    end;

    function    TModEnvAR.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envar_trig    : Result := 'trig';
        envar_attack  : Result := 'attack';
        envar_release : Result := 'release';
        envar_gain    : Result := 'gain';
        envar_sigin   : Result := 'sigin';
        envar_mute    : Result := 'mute';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModEnvAR.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envar_out    : Result := 'out';
        envar_sigout : Result := 'sigout';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvARRetrig = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FArmed        : Boolean;
  public
}

    constructor TModEnvARRetrig.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModEnvARRetrig.SetDefaults; // override;
    begin
      inherited;
      FInputs[ envar_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FDezipperMap := [
        envar_attack,
        envar_release
      ];
    end;

    procedure   TModEnvARRetrig.DoTick; // override;
    var
      R : TSignal;
    begin
      if ResetFlag
      then begin
        ResetFlag            := False;
        FState               := arIdle;
        FValue               := 0;
        FOutputs[ envar_out] := 0;
      end;

      if FArmed and SignalToLogic( FInputs[ envar_trig])
      then FState := arIdle;

      case FState of

        arIdle :

          if SignalToLogic( FInputs[ envar_trig])
          then begin
            FCoeffAttack  := 1 / Max( 1, TimeToSampleCount( FInputs[ envar_attack])); // Linear attack
            FCoeffRelease := ( Ln( 0.001) - Ln( 1.000)) / Max( 1, TimeToSampleCount( FInputs[ envar_release])); // Exp release
            FState        := arA;
            FArmed        := False;
          end;

        arA :
          begin
            if not FArmed
            then FArmed := not SignalToLogic( FInputs[ envar_trig]);
            R := FValue;
            if R < 0.001
            then R := 0.001;
            R := R + FCoeffAttack;
            if R > 0.999
            then begin
              FValue := 1.0;
              FState := arR;
            end
            else FValue := R;
            case SignalToInt( FInputs[ envar_mute]) of
              1 :  FOutputs[ envar_out] := 0;
              2 :  FOutputs[ envar_out] := 1;
              else FOutputs[ envar_out] := FValue * FInputs[ envar_gain];
            end;
          end;

        arR :
          begin
            if not FArmed
            then FArmed := not SignalToLogic( FInputs[ envar_trig]);
            R := FValue;
            R := R * ( 1 + FCoeffRelease);
            if R < 0.001
            then begin
              FValue := 0.0;
              FState := arWait;
            end
            else FValue := R;
            case SignalToInt( FInputs[ envar_mute]) of
              1 :  FOutputs[ envar_out] := 0;
              2 :  FOutputs[ envar_out] := 1;
              else FOutputs[ envar_out] := FValue * FInputs[ envar_gain];
            end;
          end;

        arWait :
          if not SignalToLogic( FInputs[ envar_trig])
          then FState := arIdle;

      end;
      case SignalToInt( FInputs[ envar_mute]) of
        1 :  FOutputs[ envar_sigout] := 0;
        2 :  FOutputs[ envar_sigout] := FInputs[ envar_sigin];
        else FOutputs[ envar_sigout] := FInputs[ envar_sigin] * FOutputs[ envar_out];
      end;
    end;

    procedure   TModEnvARRetrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ envar_trig]], AppLocale));
    end;

    function    TModEnvARRetrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envar_trig    : Result := 'trig';
        envar_attack  : Result := 'attack';
        envar_release : Result := 'release';
        envar_gain    : Result := 'gain';
        envar_sigin   : Result := 'sigin';
        envar_mute    : Result := 'mute';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModEnvARRetrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envar_out    : Result := 'out';
        envar_sigout : Result := 'sigout';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvAHD = class( TMod)
  private
    FData      : TSignalArray;
    FState     : TAHDState;
    FTime      : Integer;
    FDuration  : Integer;
    FArmed     : Boolean;
    FLastLevel : TSignal;
    FValue     : TSignal;
    FEnvValue  : TSignal;
    FTrigFlag  : Boolean;
  public
}

const

  envahd_trig       =  0;
  envahd_mode       =  1;
  envahd_ashape     =  2;
  envahd_dshape     =  3;
  envahd_attack     =  4;
  envahd_hold       =  5;
  envahd_decay      =  6;
  envahd_gain       =  7;
  envahd_sigin      =  8;
  envahd_mute       =  9;
  envahd_arange     = 10;
  envahd_hrange     = 11;
  envahd_drange     = 12;
  envahd_amod       = 13;
  envahd_hmod       = 14;
  envahd_dmod       = 15;
  envahd_amodneg    = 16;
  envahd_hmodneg    = 17;
  envahd_dmodneg    = 18;
  envahd_amodlevel  = 19;
  envahd_hmodlevel  = 20;
  envahd_dmodlevel  = 21;
  envahd_outputtype = 22;

  envahd_out        =  0;
  envahd_sigout     =  1;

    constructor TModEnvAHD.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      SetLength( FData, 67);       // Should match the viewer width in the designer
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModEnvAHD.SetDefaults; // override;
    begin
      FInputs[ envahd_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FDezipperMap := [
        envahd_attack,
        envahd_hold,
        envahd_decay,
        envahd_amodlevel,
        envahd_hmodlevel,
        envahd_dmodlevel
      ];
    end;

    procedure   TModEnvAHD.DoSlowTick; // override;
    var
      Trig : Boolean;
      AMod : TSignal;
      HMod : TSignal;
      DMod : TSignal;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState    := ahdIdle;
        FValue    := 0;
      end;

      Trig      := FTrigFlag;
      FTrigFlag := False;

      if FArmed and Trig and ( SignalToInt( FInputs[ envahd_mode]) = 1)
      then FState := ahdIdle;

      case FState of

        ahdIdle :

          if Trig
          then begin
            FArmed     := False;
            FLastLevel := FValue;
            AMod       := ( FInputs[ envahd_amod] - FInputs[ envahd_amodneg]) * FInputs[ envahd_amodlevel];
            FTime      := 0;
            FDuration  := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( FInputs[ envahd_attack] * ( 1 + AMod))));
            FState     := ahdA;
          end;

        ahdA :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then begin
              FTime     := 0;
              HMod      := ( FInputs[ envahd_hmod] - FInputs[ envahd_hmodneg]) * FInputs[ envahd_hmodlevel];
              FDuration := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( FInputs[ envahd_hold] * ( 1 + HMod))));
              FState    := ahdH;
            end
            else begin
              case SignalToEnvShape( FInputs[ envahd_ashape]) of
                esLog  : FValue := LogAttack( FTime, FDuration, FLastLevel, 1);
                esLin  : FValue := LinAttack( FTime, FDuration, FLastLevel, 1);
                else     FValue := ExpAttack( FTime, FDuration, FLastLevel, 1);
              end;
            end;
          end;

        ahdH :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then begin
              FLastLevel := FValue;
              FTime      := 0;
              DMod       := ( FInputs[ envahd_dmod] - FInputs[ envahd_dmodneg]) * FInputs[ envahd_dmodlevel];
              FDuration  := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( FInputs[ envahd_decay] * ( 1 + DMod))));
              FState     := ahdD;
            end;
          end;

        ahdD :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then FState := ahdWait
            else begin
              case SignalToEnvShape( FInputs[ envahd_dshape]) of
                esLog : FValue := LogDecay( FTime, FDuration, FLastLevel, 0);
                esLin : FValue := LinDecay( FTime, FDuration, FLastLevel, 0);
                else    FValue := ExpDecay( FTime, FDuration, FLastLevel, 0);
              end;
            end;
          end;

        ahdWait :
          if not Trig or FArmed
          then FState := ahdIdle;
      end;

      case SignalToInt( FInputs[ envahd_mute]) of
        1  : FEnvValue := Normalize(  EnvelopeSmoothing * FEnvValue                                   / ( 1 + EnvelopeSmoothing));
        2  : FEnvValue := Normalize(( EnvelopeSmoothing * FEnvValue + 1)                              / ( 1 + EnvelopeSmoothing));
        else FEnvValue := Normalize(( EnvelopeSmoothing * FEnvValue + FValue * FInputs[ envahd_gain]) / ( 1 + EnvelopeSmoothing));
      end;

      FOutputs[ envahd_out] := SignalForEnvOutputType( FEnvValue, SignalToInt( FInputs[ envahd_outputtype]));
    end;

    procedure   TModEnvAHD.DoTick; // override;
    begin
      if SignalToLogic( FInputs[ envahd_trig])
      then FTrigFlag := True;
      FOutputs[ envahd_sigout] := FInputs[ envahd_sigin] * FOutputs[ envahd_out]; // sigout := sigin * out
    end;

    procedure   TModEnvAHD.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, LogicToSignal( not ( FState in [ ahdIdle, ahdWait]))], AppLocale));
      aCallback( Self, Format( '%s/%s/trigin/%f', [ aPrefix, Name, FInputs[ envahd_trig]                               ], AppLocale));
    end;

    procedure   TModEnvAHD.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      ATime  : TSignal;
      HTime  : TSignal;
      DTime  : TSignal;
      TTime  : TSignal;
      i      : Integer;
      p      : Integer;
      AShape : TEnvShape;
      DShape : TEnvShape;
      R      : TSignal;
    begin
      ATime  := FInputs[ envahd_attack];
      HTime  := FInputs[ envahd_hold  ];
      DTime  := FInputs[ envahd_decay ];
      TTime  := ATime + HTime + DTime;

      if TTIme = 0
      then Exit;

      ATime  := ATime * 67 / TTime;
      HTime  := HTime * 67 / TTime;
      DTime  := DTime * 67 / TTime;
      AShape := SignalToEnvShape( FInputs[ envahd_ashape]);
      DShape := SignalToEnvShape( FInputs[ envahd_dshape]);

      i := 0;
      while i < ATime
      do begin
        case AShape of
          esLog  : R := LogAttack( i, ATime, 0, 1);
          esLin  : R := LinAttack( i, ATime, 0, 1);
          else     R := ExpAttack( i, ATime, 0, 1);
        end;
        FData[ i] :=  SignalForEnvOutputType( R, SignalToInt( FInputs[ envahd_outputtype]));
        Inc( i);
      end;
      while i < ATime + HTime
      do begin
        FData[ i] := SignalForEnvOutputType( 1, SignalToInt( FInputs[ envahd_outputtype]));
        Inc( i);
      end;
      p := 0;
      while i < 67
      do begin
        case DShape of
          esLog : R := LogDecay( p, DTime, 1, 0);
          esLin : R := LinDecay( p, DTime, 1, 0);
          else    R := ExpDecay( p, DTime, 1, 0);
        end;
        FData[ i] := SignalForEnvOutputType( R, SignalToInt( FInputs[ envahd_outputtype]));
        Inc( i);
        Inc( p);
      end;

      aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FData);
    end;

    function    TModEnvAHD.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envahd_trig       : Result := 'trig';
        envahd_mode       : Result := 'mode';
        envahd_ashape     : Result := 'ashape';
        envahd_dshape     : Result := 'dshape';
        envahd_attack     : Result := 'attack';
        envahd_hold       : Result := 'hold';
        envahd_decay      : Result := 'decay';
        envahd_gain       : Result := 'gain';
        envahd_sigin      : Result := 'sigin';
        envahd_mute       : Result := 'mute';
        envahd_arange     : Result := 'arange';
        envahd_hrange     : Result := 'hrange';
        envahd_drange     : Result := 'drange';
        envahd_amod       : Result := 'amod';
        envahd_hmod       : Result := 'hmod';
        envahd_dmod       : Result := 'dmod';
        envahd_amodneg    : Result := 'amodneg';
        envahd_hmodneg    : Result := 'hmodneg';
        envahd_dmodneg    : Result := 'dmodneg';
        envahd_amodlevel  : Result := 'amodlevel';
        envahd_hmodlevel  : Result := 'hmodlevel';
        envahd_dmodlevel  : Result := 'dmodlevel';
        envahd_outputtype : Result := 'outputtype';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModEnvAHD.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envahd_out    : Result := 'out';
        envahd_sigout : Result := 'sigout';
        else            Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModEnvADSR = class( TMod)
  private
    FData      : TSignalArray;
    FState     : TADSRState;
    FTime      : Integer;
    FDuration  : Integer;
    FArmed     : Boolean;
    FLastLevel : TSignal;
    FValue     : TSignal;
    FEnvValue  : TSignal;
    FTrigFlag  : Boolean;
  public
}

const

  envadsr_trig       =  0;
  envadsr_mode       =  1;
  envadsr_ashape     =  2;
  envadsr_dshape     =  3;
  envadsr_rshape     =  4;
  envadsr_attack     =  5;
  envadsr_decay      =  6;
  envadsr_sustain    =  7;
  envadsr_release    =  8;
  envadsr_gain       =  9;
  envadsr_sigin      = 10;
  envadsr_mute       = 11;
  envadsr_arange     = 12;
  envadsr_drange     = 13;
  envadsr_rrange     = 14;
  envadsr_amod       = 15;
  envadsr_dmod       = 16;
  envadsr_smod       = 17;
  envadsr_rmod       = 18;
  envadsr_amodneg    = 19;
  envadsr_dmodneg    = 20;
  envadsr_smodneg    = 21;
  envadsr_rmodneg    = 22;
  envadsr_amodlevel  = 23;
  envadsr_dmodlevel  = 24;
  envadsr_smodlevel  = 25;
  envadsr_rmodlevel  = 26;
  envadsr_outputtype = 27;

  envadsr_out        =  0;
  envadsr_sigout     =  1;

    constructor TModEnvADSR.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      SetLength( FData, 67);       // Should match the viewer width in the designer
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModEnvADSR.SetDefaults; // override;
    begin
      inherited;
      FInputs[ envadsr_gain] := 1; // Set default gain to 1, so it'll work unconnected
      FDezipperMap := [
        envadsr_attack,
        envadsr_decay,
        envadsr_sustain,
        envadsr_release,
        envadsr_amodlevel,
        envadsr_dmodlevel,
        envadsr_smodlevel,
        envadsr_rmodlevel
      ];
    end;

    procedure   TModEnvADSR.DoSlowTick; // override;
    var
      Trig : Boolean;
      AMod : TSignal;
      DMod : TSignal;
      SMod : TSignal;
      RMod : TSignal;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState    := adsrIdle;
        FValue    := 0;
      end;

      Trig      := FTrigFlag;
      FTrigFlag := False;

      if FArmed and Trig and ( SignalToInt( FInputs[ envadsr_mode]) = 1)
      then FState := adsrIdle;

      case FState of

        adsrIdle :

          if Trig
          then begin
            FState     := adsrA;
            FArmed     := False;
            FLastLevel := FValue;
            FTime      := 0;
            AMod       := ( FInputs[ envadsr_amod] - FInputs[ envadsr_amodneg]) * FInputs[ envadsr_amodlevel];
            FDuration  := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( FInputs[ envadsr_attack] * ( 1 + AMod))));
          end;

        adsrA :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then begin
              FLastLevel := FValue;
              FTime      := 0;
              DMod       := ( FInputs[ envadsr_dmod] - FInputs[ envadsr_dmodneg]) * FInputs[ envadsr_dmodlevel];
              FDuration  := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( FInputs[ envadsr_decay] * ( 1 + DMod))));
              FState     := adsrD;
            end
            else begin
              case SignalToEnvShape( FInputs[ envadsr_ashape ]) of
                esLog : FValue := LogAttack( FTime, FDuration, FLastLevel, 1);
                esLin : FValue := LinAttack( FTime, FDuration, FLastLevel, 1);
                else    FValue := ExpAttack( FTime, FDuration, FLastLevel, 1);
              end;
            end;
          end;

        adsrD :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then begin
              FLastLevel := FValue;
              FState     := adsrS;
            end
            else begin
              SMod := ( FInputs[ envadsr_smod] - FInputs[ envadsr_smodneg]) * FInputs[ envadsr_smodlevel];
              case SignalToEnvShape( FInputs[ envadsr_dshape ]) of
                esLog : FValue := LogDecay( FTime, FDuration, FLastLevel, FInputs[ envadsr_sustain] + SMod);
                esLin : FValue := LinDecay( FTime, FDuration, FLastLevel, FInputs[ envadsr_sustain] + SMod);
                else    FValue := ExpDecay( FTime, FDuration, FLastLevel, FInputs[ envadsr_sustain] + SMod);
              end;
            end;
          end;

        adsrS :
          begin
            if not FArmed
            then FArmed := not Trig;

            if FArmed
            then begin
              FTime       := 0;
              RMod        := ( FInputs[ envadsr_rmod] - FInputs[ envadsr_rmodneg]) * FInputs[ envadsr_rmodlevel];
              FDuration   := Max( ShortestPossibleEnvelopeTime, Round( SlowTimeToSampleCount( FInputs[ envadsr_release] * ( 1 + RMod))));
              FState      := adsrR;
            end;
          end;

        adsrR :
          begin
            if not FArmed
            then FArmed := not Trig;

            Inc( FTime);
            if FTime > FDuration
            then FState := adsrWait
            else begin
              case SignalToEnvShape( FInputs[ envadsr_rshape ]) of
                esLog : FValue := LogDecay( FTime, FDuration, FLastLevel, 0);
                esLin : FValue := LinDecay( FTime, FDuration, FLastLevel, 0);
                else    FValue := ExpDecay( FTime, FDuration, FLastLevel, 0);
              end;
            end;
          end;

        adsrWait :
          if not Trig or FArmed
          then FState := adsrIdle;
      end;

      case SignalToInt( FInputs[ envadsr_mute]) of
        1  : FEnvValue := Normalize(  EnvelopeSmoothing * FEnvValue                                    / ( 1 + EnvelopeSmoothing));
        2  : FEnvValue := Normalize(( EnvelopeSmoothing * FEnvValue + 1 * 1)                           / ( 1 + EnvelopeSmoothing));
        else FEnvValue := Normalize(( EnvelopeSmoothing * FEnvValue + FValue * FInputs[ envadsr_gain]) / ( 1 + EnvelopeSmoothing));
      end;
      FOutputs[ envadsr_out] := SignalForEnvOutputType( FEnvValue, SignalToInt( FInputs[ envadsr_outputtype]));
    end;

    procedure   TModEnvADSR.DoTick; // override;
    begin
      if SignalToLogic( FInputs[ envadsr_trig])
      then FTrigFlag := True;
      FOutputs[ envadsr_sigout] := FInputs[ envadsr_sigin] * FOutputs[ envadsr_out]; // sigout := sigin * out
    end;

    procedure   TModEnvADSR.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, LogicToSignal( not ( FState in [ adsrIdle, adsrWait]))], AppLocale));
      aCallback( Self, Format( '%s/%s/trigin/%f', [ aPrefix, Name, FInputs[ envadsr_trig]                                ], AppLocale));
    end;

    procedure   TModEnvADSR.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      ATime  : TSignal;
      DTime  : TSignal;
      STime  : TSignal;
      RTime  : TSignal;
      TTime  : TSignal;
      SLevel : TSignal;
      i      : Integer;
      p      : Integer;
      R      : TSignal;
    begin
      ATime  := FInputs[ envadsr_attack ];
      DTime  := FInputs[ envadsr_decay  ];
      SLevel := FInputs[ envadsr_sustain];
      RTime  := FInputs[ envadsr_release];
      STime  := ( ATime + DTime + RTime) / 3;
      TTime  := ATime + DTime + STime + RTime;

      if TTIme = 0
      then Exit;

      ATime  := ATime * 67 / TTime;
      DTime  := DTime * 67 / TTime;
      RTime  := RTime * 67 / TTime;
      STime  := STime * 67 / TTime;

      i := 0;
      while i < ATime
      do begin
        case SignalToInt( FInputs[ envadsr_ashape]) of
          0  : R := LogAttack( i, ATime, 0, 1);
          1  : R := LinAttack( i, ATime, 0, 1);
          else R := ExpAttack( i, ATime, 0, 1);
        end;
        FData[ i] := SignalForEnvOutputType( R, SignalToInt( FInputs[ envadsr_outputtype]));
        Inc( i);
      end;

      p := 0;
      while i < ATime + DTime
      do begin
        case SignalToEnvShape( FInputs[ envadsr_dshape]) of
          esLog : R := LogDecay( p, DTime, 1, SLevel);
          esLin : R := LinDecay( p, DTime, 1, SLevel);
          else    R := ExpDecay( p, DTime, 1, SLevel);
        end;
        FData[ i] := SignalForEnvOutputType( R, SignalToInt( FInputs[ envadsr_outputtype]));
        Inc( i);
        Inc( p);
      end;

      while i < ATime + DTime + STime
      do begin
        FData[ i] := SignalForEnvOutputType( SLevel, SignalToInt( FInputs[ envadsr_outputtype]));
        Inc( i);
      end;

      p := 0;
      while i < 67
      do begin
        case SignalToEnvShape( FInputs[ envadsr_rshape]) of
          esLog : R := LogDecay( p, RTime, SLevel, 0);
          esLin : R := LinDecay( p, RTime, SLevel, 0);
          else R := ExpDecay( p, RTime, SLevel, 0);
        end;
        FData[ i] := SignalForEnvOutputType( R, SignalToInt( FInputs[ envadsr_outputtype]));
        Inc( i);
        Inc( p);
      end;

      aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FData);
    end;

    function    TModEnvADSR.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envadsr_trig       : Result := 'trig';
        envadsr_mode       : Result := 'mode';
        envadsr_ashape     : Result := 'ashape';
        envadsr_dshape     : Result := 'dshape';
        envadsr_rshape     : Result := 'rshape';
        envadsr_attack     : Result := 'attack';
        envadsr_decay      : Result := 'decay';
        envadsr_sustain    : Result := 'sustain';
        envadsr_release    : Result := 'release';
        envadsr_gain       : Result := 'gain';
        envadsr_sigin      : Result := 'sigin';
        envadsr_mute       : Result := 'mute';
        envadsr_arange     : Result := 'arange';
        envadsr_drange     : Result := 'drange';
        envadsr_rrange     : Result := 'rrange';
        envadsr_amod       : Result := 'amod';
        envadsr_dmod       : Result := 'dmod';
        envadsr_smod       : Result := 'smod';
        envadsr_rmod       : Result := 'rmod';
        envadsr_amodneg    : Result := 'amodneg';
        envadsr_dmodneg    : Result := 'dmodneg';
        envadsr_smodneg    : Result := 'smodneg';
        envadsr_rmodneg    : Result := 'rmodneg';
        envadsr_amodlevel  : Result := 'amodlevel';
        envadsr_dmodlevel  : Result := 'dmodlevel';
        envadsr_smodlevel  : Result := 'smodlevel';
        envadsr_rmodlevel  : Result := 'rmodlevel';
        envadsr_outputtype : Result := 'outputtype';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModEnvADSR.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envadsr_out    : Result := 'out';
        envadsr_sigout : Result := 'sigout';
        else             Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FSteps    : Integer;
    FCounter  : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
  public
}

const

  sequencer_trig   =  0;
  sequencer_reset  =  1;
  sequencer_dir    =  2;
  sequencer_chain  =  3;
  sequencer_rnd    =  4;
  sequencer_active =  5;
  sequencer_step1  =  6;   // Keep steps consecutive
  sequencer_step2  =  7;
  sequencer_step3  =  8;
  sequencer_step4  =  9;
  sequencer_step5  = 10;
  sequencer_step6  = 11;
  sequencer_step7  = 12;
  sequencer_step8  = 13;
  sequencer_step9  = 14;
  sequencer_step10 = 15;
  sequencer_step11 = 16;

  sequencer_out    =  0;   // Keep trigs comsecutive
  sequencer_trig1  =  1;
  sequencer_trig2  =  2;
  sequencer_trig3  =  3;
  sequencer_trig4  =  4;
  sequencer_trig5  =  5;
  sequencer_trig6  =  6;
  sequencer_trig7  =  7;
  sequencer_trig8  =  8;
  sequencer_trig9  =  9;
  sequencer_trig10 = 10;
  sequencer_trig11 = 11;


//  public

    constructor TModSequencer.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModSequencer.SetDefaults; // override;
    begin
      inherited;
      FOutValue                  := 0;    // Neutral output
      FInputs[ sequencer_dir   ] := 1;    // Make seq run forwards by default.
      FInputs[ sequencer_chain ] := 0;    // Chain input to zero
      FInputs[ sequencer_active] := 1;    // Sequencer active when active input not connected
      FCounter                   := 0;
    end;

    procedure   TModSequencer.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 11)
      else inherited;
    end;

    function    TModSequencer.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModSequencer.DoTick; // override;
    begin
      if not FOldRes and SignalToLogic( FInputs[ sequencer_reset])
      then begin
        FInState  := seqClkHigh;
        FCounter  := 0;
        FOutValue := 0;
      end;
      FOldRes := SignalToLogic( FInputs[ sequencer_reset]);
    end;

    procedure   TModSequencer.DoSlowTick; // override;
    var
      i : Integer;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FInState  := seqClkHigh;
        FCounter  := 0;
        FOutValue := 0;
      end;

      case FInState of
        seqClkLow :
          begin
            if SignalToLogic( FInputs[ sequencer_trig])
            then begin
              if FSteps > 0
              then begin
                FInState := seqClkHigh;
                if SignalToLogic( FInputs[ sequencer_dir])
                then FCounter := ( FCounter + 1) mod FSteps
                else begin
                  if FCounter <= 0
                  then FCounter := FSteps - 1
                  else Dec( FCounter);
                end;
              end;
            end;
          end;
        seqClkHigh :
          begin
            if not SignalToLogic( FInputs[ sequencer_trig])
            then FInState := seqClkLow;
          end;
      end; // case
      if SignalToLogic( FInputs[ sequencer_active])
      then FOutValue := FInputs[ FCounter + sequencer_step1] - MiddleNote / 128
      else FOutValue := 0;

      FOutputs[ sequencer_out] := FOutValue + FInputs[ sequencer_chain];

      for i := sequencer_trig1 to sequencer_trig11
      do begin
        if ( i = FCounter + sequencer_trig1) and SignalToLogic( FInputs[ sequencer_active])
        then FOutputs[ i] := LogicToSignal( True )
        else FOutputs[ i] := LogicToSignal( False);
      end;
    end;

    procedure   TModSequencer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, FInputs [ sequencer_trig  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f' , [ aPrefix, Name, FInputs [ sequencer_reset ]], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f', [ aPrefix, Name, FInputs [ sequencer_active]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'   , [ aPrefix, Name, FInputs [ sequencer_dir   ]], AppLocale));
      for i := sequencer_trig1 to sequencer_trig11
      do begin
        if ( i = FCounter + sequencer_trig1) and SignalToLogic( FInputs[ sequencer_active])
        then aCallback( Self, Format( '%s/%s/step%d/1.0', [ aPrefix, Name, i - sequencer_trig1 + 1], AppLocale))
        else aCallback( Self, Format( '%s/%s/step%d/0.0', [ aPrefix, Name, i - sequencer_trig1 + 1], AppLocale));
      end;
    end;

    function    TModSequencer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        sequencer_trig   : Result := 'trig';
        sequencer_reset  : Result := 'reset';
        sequencer_dir    : Result := 'dir';
        sequencer_chain  : Result := 'chain';
        sequencer_rnd    : Result := 'rnd';
        sequencer_active : Result := 'active';
        sequencer_step1  : Result := 'step1';
        sequencer_step2  : Result := 'step2';
        sequencer_step3  : Result := 'step3';
        sequencer_step4  : Result := 'step4';
        sequencer_step5  : Result := 'step5';
        sequencer_step6  : Result := 'step6';
        sequencer_step7  : Result := 'step7';
        sequencer_step8  : Result := 'step8';
        sequencer_step9  : Result := 'step9';
        sequencer_step10 : Result := 'step10';
        sequencer_step11 : Result := 'step11';
        else               Result := NO_INPUT;;
      end;
    end;

    function    TModSequencer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        sequencer_out    : Result := 'out';
        sequencer_trig1  : Result := 'trig1';
        sequencer_trig2  : Result := 'trig2';
        sequencer_trig3  : Result := 'trig3';
        sequencer_trig4  : Result := 'trig4';
        sequencer_trig5  : Result := 'trig5';
        sequencer_trig6  : Result := 'trig6';
        sequencer_trig7  : Result := 'trig7';
        sequencer_trig8  : Result := 'trig8';
        sequencer_trig9  : Result := 'trig9';
        sequencer_trig10 : Result := 'trig10';
        sequencer_trig11 : Result := 'trig11';
        else               Result := NO_OUTPUT;
      end;
    end;

    function    TModSequencer.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'steps';
        else Result := NO_INTERNAL;
      end;
    end;

{ ========
  TModSeqSeq = class( TMod)
  private
    FState           : Word;
    FSteps           : Integer;
    FSequenceCounter : Integer;
    FPeriodCounter   : Integer;
    FOldRes          : Boolean;
    FOldTrig         : Boolean;
    FClkCnt          : Integer;
  public
}

const

  seqseq_res     =  0;
  seqseq_trig    =  1;
  seqseq_dir     =  2;
  seqseq_resmod  =  3;

  seqseq_outres  =  0;
  seqseq_outtrig =  1;
  seqseq_outdir  =  2;
  seqseq_out1    =  3;
  seqseq_out2    =  4;
  seqseq_out3    =  5;
  seqseq_out4    =  6;
  seqseq_out5    =  7;
  seqseq_out6    =  8;
  seqseq_out7    =  9;
  seqseq_out8    = 10;
  seqseq_out9    = 11;
  seqseq_out10   = 12;
  seqseq_out11   = 13;
  seqseq_out12   = 14;
  seqseq_out13   = 15;
  seqseq_out14   = 16;
  seqseq_out15   = 17;
  seqseq_out16   = 18;

  seqseq_clkdel  = 17;

  seqseq_resmod1 =  0;
  seqseq_resmod2 =  1;
  seqseq_resmod3 =  2;

    constructor TModSeqSeq.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModSeqSeq.SetDefaults; // override;
    begin
      FInputs[ seqseq_dir] :=  1;        // Go forwards with unconnected dir input
      FInputs[ seqseq_res] := -1;        // No reset on unconnected input
      FSequenceCounter     := 16;
      FPeriodCounter       := 24;
    end;

    procedure   TModSeqSeq.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 256)
      else inherited;
    end;

    function    TModSeqSeq.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModSeqSeq.DoTick; // override;
    const
      MaxCount = 16;
    var
      Direction : Boolean;
      Trig      : Boolean;
      Res       : Boolean;
      i         : Integer;
      DoChange  : Boolean;
      EndOfSeq  : Boolean;
      MustReset : Boolean;
      ResMod    : Integer;
    begin
      DoChange  := False;
      MustReset := False;

      ResMod    := SignalToInt  ( FInputs[ seqseq_resmod]);
      Direction := SignalToLogic( FInputs[ seqseq_dir   ]);
      Trig      := SignalToLogic( FInputs[ seqseq_trig  ]);
      Res       := SignalToLogic( FInputs[ seqseq_res   ]);
      EndOfSeq  := FPeriodCounter <= 0;

      if (( not FOldRes) and Res) or ResetFlag or EndOfSeq
      then begin
        if Direction
        then begin                                             // Forwards , reset to first state
          FState           :=  0;
          FSequenceCounter := 15;
        end
        else begin                                             // Backwards , reset to last state
          FState           := FSteps div 16;
          FSequenceCounter := FSteps mod 16 - 1;
        end;
        FClkCnt        := seqseq_clkdel;
        FPeriodCounter := FSteps;
        DoChange       := True;
        MustReset      := MustReset or ( ResMod in [ seqseq_resmod1, seqseq_resmod3]);   // Propagate reset on received reset or end of sequence in mode 1 and 3
        ResetFlag      := False;
        FOldTrig       := True;
      end
      else begin
        if Trig and not FOldTrig
        then begin
          if not EndOfSeq
          then Dec( FPeriodCounter);
          if FSequenceCounter > 0
          then Dec( FSequenceCounter)
          else begin
            FSequenceCounter := 15;
            if Direction
            then begin                                         // Forwards, go to nexxt state
              FState := FState + 1;
              if FState > FSteps div 16
              then FState := 0;
                  end
            else begin                                         // Backwards, go to previous state
              if FState > 0
              then Dec( FState)
              else FState := Fsteps div 16;
            end;
            MustReset := MustReset or ( ResMod in [ seqseq_resmod2, seqseq_resmod3]);    // Issue reset after state change in mode 2 and 3
            DoChange  := True;
          end;
          FClkCnt := seqseq_clkdel;
        end
        else if FOldTrig and not Trig
        then begin
          DoChange := True;
          FClkCnt  := seqseq_clkdel;
        end;
        FOldTrig := Trig;
      end;
      FOldRes := Res;

      if FClkCnt > 0
      then begin
        Dec( FClkCnt);
        if FClkCnt = 0
        then FOutputs[ seqseq_outtrig] := LogicToSignal( Trig);
      end;

      if DoChange
      then begin
        for i := 0 to MaxCount - 1
        do begin
          if i = FState
          then FOutputs[ seqseq_out1 + i] := LogicToSignal( True )
          else FOutputs[ seqseq_out1 + i] := LogicToSignal( False);
        end;
        FOutputs[ seqseq_outres] := FInputs[ seqseq_res];
        FOutputs[ seqseq_outdir] := FInputs[ seqseq_dir];
      end;

      if MustReset
      then begin
        FOutputs[ seqseq_outres ] := LogicToSignal( True);
        FOutputs[ seqseq_outtrig] := LogicToSignal( True);
      end;
    end;

    procedure   TModSeqSeq.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := seqseq_out1 to seqseq_out16
      do aCallback( Self, Format( '%s/%s/step%d/%f', [ aPrefix, Name, 1 + i - seqseq_out1, FOutputs[ i]], AppLocale));

      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ seqseq_trig]], AppLocale));
      aCallback( Self, Format( '%s/%s/res/%f' , [ aPrefix, Name, FInputs[ seqseq_res ]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f' , [ aPrefix, Name, FInputs[ seqseq_dir ]], AppLocale));
    end;

    function    TModSeqSeq.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqseq_res    : Result := 'res';
        seqseq_trig   : Result := 'trig';
        seqseq_dir    : Result := 'dir';
        seqseq_resmod : Result := 'resmod';
        else            Result := NO_INPUT;;
      end;
    end;

    function    TModSeqSeq.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqseq_outres  : Result := 'outres';
        seqseq_outtrig : Result := 'outtrig';
        seqseq_outdir  : Result := 'outdir';
        seqseq_out1    : Result := 'out1';
        seqseq_out2    : Result := 'out2';
        seqseq_out3    : Result := 'out3';
        seqseq_out4    : Result := 'out4';
        seqseq_out5    : Result := 'out5';
        seqseq_out6    : Result := 'out6';
        seqseq_out7    : Result := 'out7';
        seqseq_out8    : Result := 'out8';
        seqseq_out9    : Result := 'out9';
        seqseq_out10   : Result := 'out10';
        seqseq_out11   : Result := 'out11';
        seqseq_out12   : Result := 'out12';
        seqseq_out13   : Result := 'out13';
        seqseq_out14   : Result := 'out14';
        seqseq_out15   : Result := 'out15';
        seqseq_out16   : Result := 'out16';
        else             Result := NO_OUTPUT;
      end;
    end;

    function    TModSeqSeq.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'steps';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModSeq16 = class( TMod)
  private
    FOutValue   : TSignal;
    FReset      : Boolean;
    FNextValue  : TSignal;
    FForwards   : Boolean;
    FActive     : Boolean;
    FChainValue : TSignal;
    FSteps      : Integer;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FOldRes     : Boolean;
    FOldActive  : Boolean;
    FTicks      : Integer;
    FPeriodTime : Integer;
  private
}

const

  seq16_trig       =  0;
  seq16_res        =  1;
  seq16_dir        =  2;
  seq16_chain      =  3;
  seq16_rnd        =  4;
  seq16_active     =  5;
  seq16_mode       =  6;
  seq16_curveup    =  7;
  seq16_curvedown  =  8;
  seq16_xfade      =  9;
  seq16_step1      = 10;   // Keep steps consecutive
  seq16_step2      = 11;
  seq16_step3      = 12;
  seq16_step4      = 13;
  seq16_step5      = 14;
  seq16_step6      = 15;
  seq16_step7      = 16;
  seq16_step8      = 17;
  seq16_step9      = 18;
  seq16_step10     = 19;
  seq16_step11     = 20;
  seq16_step12     = 21;
  seq16_step13     = 22;
  seq16_step14     = 23;
  seq16_step15     = 24;
  seq16_step16     = 25;

  seq16_out        =  0;   // Keep trigs comsecutive
  seq16_outres     =  1;
  seq16_outtrig    =  2;
  seq16_outdir     =  3;

  seq16_mode_lin44 = 0;   // linear [-4, 4]
  seq16_mode_db4   = 1;   // exp --140, 12] dB
  seq16_mode_lin11 = 2;   // linear [-1, 1]
  seq16_mode_db0   = 3;   // exp --140, 0] dB
  seq16_mode_lin04 = 4;   // linear [0, 4]
  seq16_mode_lin01 = 5;   // linear [0, 1]
  seq16_mode_note  = 6;   // note values ([0,127] nn - MiddleNote) / 128

  seq16_clkdelay   = 9;

    procedure   TModSeq16.SampleInputs;
    begin
      FInState    := seqClkHigh;
      FActive     := SignalToLogic( FInputs[ seq16_active]) and ( FSteps > 0);
      FChainValue := FInputs[ seq16_chain];
      FForwards   := SignalToLogic( FInputs[ seq16_dir]);
      FReset      := SignalToLogic( FInputs[ seq16_res]);
    end;

    procedure   TModSeq16.MakeNextValue;
    begin
      FOutValue := FNextValue;
      if FActive
      then begin
        if SignalToInt( FInputs[ seq16_mode]) = seq16_mode_note
        then FNextValue := FInputs[ FCounter + seq16_step1] - MiddleNote / 128
        else FNextValue := FInputs[ FCounter + seq16_step1];
      end
      else FNextValue := FChainValue;
    end;

//  public

    constructor TModSeq16.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModSeq16.SetDefaults; // override;
    begin
      inherited;
      FInputs[ seq16_dir   ] := 1;                 // Make seq run forwards by default.
      FInputs[ seq16_chain ] := 0;                 // Chain input to zero wwhen not connected
      FInputs[ seq16_active] := 1;                 // Sequencer active when active input not connected
      FInputs[ seq16_mode  ] := seq16_mode_note;   // Default startup mode is notes
      FNextValue             := 0;                 // Neutral output
      FPeriodTime            := 2;
      FTicks                 := 0;
      SampleInputs;
      if FForwards
      then FCounter := 0
      else FCounter := FSteps - 1;
    end;

    procedure   TModSeq16.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 16)
      else inherited;
    end;

    function    TModSeq16.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModSeq16.DoTick; // override;
    var
      XFade : TSignal;
    begin
      if ResetFlag
      then begin
        SampleInputs;
        FPeriodTime := Max( 2, FTicks);
        FTicks      := 0;
        FNextValue  := 0;
        if FForwards
        then FCounter := 0
        else FCounter := FSteps - 1;
        MakeNextValue;
      end
      else begin
        case FInState of
          seqClkLow :
            begin
              if SignalToLogic( FInputs[ seq16_trig])
              then begin
                SampleInputs;
                if ResetFlag or ( not FOldRes and FReset)
                then begin
                  FPeriodTime := Max( 2, FTicks);
                  FTicks      := 0;
                  if FForwards
                  then FCounter := 0
                  else FCounter := FSteps - 1;
                end
                else begin
                  if FSteps > 0
                  then begin
                    SampleInputs;
                    FPeriodTime := Max( 2, FTicks);
                    FTicks      := 0;
                    if FForwards
                    then FCounter := ( FCounter + 1) mod FSteps
                    else begin
                      if FCounter <= 0
                      then FCounter := FSteps - 1
                      else Dec( FCounter);
                    end;
                  end;
                end;
                MakeNextValue;
              end;
            end;
          seqClkHigh :
            begin
              if not SignalToLogic( FInputs[ seq16_trig])
              then begin
                FInState    := seqClkLow;
                FPeriodTime := FTicks * 2;
              end;
            end;
        end; // case
      end;

      XFade := Clip((( FPeriodTime - FTicks) / FPeriodTime) * FInputs[ seq16_xfade], -1, 1);
//    FOutputs[ seq16_out    ] := FNextValue + XFade * ( FOutValue - FNextValue);

      if FOutValue < FNextValue      // going up, note that as final and current value are reversed this must call decay ...
      then begin
        case SignalToInt( FInputs[ seq16_curveup]) of
          0  : FOutputs[ seq16_out] := LogDecay( XFade, 1, FNextValue, FOutValue);
          1  : FOutputs[ seq16_out] := LinDecay( XFade, 1, FNextValue, FOutValue);
          else FOutputs[ seq16_out] := ExpDecay( XFade, 1, FNextValue, FOutValue);
        end;
      end
      else begin                     // going down
        case SignalToInt( FInputs[ seq16_curvedown]) of
          0  : FOutputs[ seq16_out] := LogAttack( XFade, 1, FNextValue, FOutValue);
          1  : FOutputs[ seq16_out] := LinAttack( XFade, 1, FNextValue, FOutValue);
          else FOutputs[ seq16_out] := ExpAttack( XFade, 1, FNextValue, FOutValue);
        end;
      end;

      FOutputs[ seq16_outres ] := FInputs[ seq16_res ];
      FOutputs[ seq16_outtrig] := FInputs[ seq16_trig];
      FOutputs[ seq16_outdir ] := FInputs[ seq16_dir ];

      if FTicks < MAXINT div 4    // Cap FTicks value in case the module gets not clocked prolly MAXINT div 2 would be OK too
      then Inc( FTicks);

      FOldRes    := FReset;
      ResetFlag  := False;
      FOldActive := FActive;
    end;

    procedure   TModSeq16.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, FInputs [ seq16_trig  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f' , [ aPrefix, Name, FInputs [ seq16_res   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f', [ aPrefix, Name, FInputs [ seq16_active]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'   , [ aPrefix, Name, FInputs [ seq16_dir   ]], AppLocale));
      for i := seq16_step1 to seq16_step16
      do begin
        if ( i = FCounter + seq16_step1) and SignalToLogic( FInputs[ seq16_active])
        then aCallback( Self, Format( '%s/%s/step%d/1.0', [ aPrefix, Name, i - seq16_step1 + 1], AppLocale))
        else aCallback( Self, Format( '%s/%s/step%d/0.0', [ aPrefix, Name, i - seq16_step1 + 1], AppLocale));
      end;
    end;

    function    TModSeq16.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seq16_trig      : Result := 'trig';
        seq16_res       : Result := 'res';
        seq16_dir       : Result := 'dir';
        seq16_chain     : Result := 'chain';
        seq16_rnd       : Result := 'rnd';
        seq16_active    : Result := 'active';
        seq16_mode      : Result := 'mode';
        seq16_xfade     : Result := 'xfade';
        seq16_curveup   : Result := 'curveup';
        seq16_curvedown : Result := 'curvedown';
        seq16_step1     : Result := 'step1';
        seq16_step2     : Result := 'step2';
        seq16_step3     : Result := 'step3';
        seq16_step4     : Result := 'step4';
        seq16_step5     : Result := 'step5';
        seq16_step6     : Result := 'step6';
        seq16_step7     : Result := 'step7';
        seq16_step8     : Result := 'step8';
        seq16_step9     : Result := 'step9';
        seq16_step10    : Result := 'step10';
        seq16_step11    : Result := 'step11';
        seq16_step12    : Result := 'step12';
        seq16_step13    : Result := 'step13';
        seq16_step14    : Result := 'step14';
        seq16_step15    : Result := 'step15';
        seq16_step16    : Result := 'step16';
        else              Result := NO_INPUT;;
      end;
    end;

    function    TModSeq16.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seq16_out     : Result := 'out';
        seq16_outres  : Result := 'outres';
        seq16_outtrig : Result := 'outtrig';
        seq16_outdir  : Result := 'outdir';
        else            Result := NO_OUTPUT;
      end;
    end;

    function    TModSeq16.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'steps';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModTextSequencer = class( TMod)
  private
    FOutValue : TSignal;
    FCounter  : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
  public
}

const

  textsequencer_trig    = 0;
  textsequencer_reset   = 1;
  textsequencer_dir     = 2;
  textsequencer_mode    = 3;
  textsequencer_chain   = 4;
  textsequencer_step1   = 5;

  textsequencer_out     = 0;
  textsequencer_gateout = 1;


    constructor TModTextSequencer.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModTextSequencer.SetDefaults; // override;
    begin
      inherited;
      FOutValue                        := 0;
      FInputs [ textsequencer_dir    ] := 1; // Make seq run forward by default.
      FInputs [ textsequencer_chain  ] := 0; // Clear chain input
      FOutputs[ textsequencer_gateout] := LogicToSignal( False);
    end;

    procedure   TModTextSequencer.DoSlowTick; // override;
    var
      CurrentValue : TSIgnal;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FInState  := seqClkLow;
        FCounter  := 0;
        FOutValue := 0;
      end;

      if not FOldRes and SignalToLogic( FInputs[ textsequencer_reset])
      then begin
        FInState := seqClkLow;
        FCounter := 0;
      end;

      case FInState of

        seqClkLow :
          begin
            if SignalToLogic( FInputs[ textsequencer_trig])
            then begin
              FInState := seqClkHigh;
              if FSteps > 0
              then begin
                if FInputs[ textsequencer_mode] = 0
                then begin
                  if SignalToLogic( FInputs[ textsequencer_dir])
                  then FCounter := ( FCounter + 1) mod FSteps
                  else begin
                    if FCounter <= 0
                    then FCounter := FSteps - 1
                    else Dec( FCounter);
                  end;
                end
                else FCounter := Random( FSteps);
              end;

              CurrentValue := FInputs[ FCounter + textsequencer_step1];
              if not AlmostEqual( CurrentValue, NO_NOTE, DELTA)
              then begin
                FOutValue := CurrentValue - MiddleNote / 128;
                FOutputs[ textsequencer_gateout] := LogicToSignal( True);
              end;
            end;
          end;

        seqClkHigh :
          begin
            if FInputs[ textsequencer_trig] <= 0
            then begin
              FInState := seqClkLow;
              FOutputs[ textsequencer_gateout] := LogicToSignal( False);
            end;
          end;

      end;
      FOutputs[ textsequencer_out] := FOutValue + FInputs[ textsequencer_chain];
      FOldRes := SignalToLogic( FInputs[ textsequencer_reset]);
    end;

    procedure   TModTextSequencer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInputs[ textsequencer_trig    ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInputs[ textsequencer_reset   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'  , [ aPrefix, Name, FInputs[ textsequencer_dir     ]], AppLocale));
      aCallback( Self, Format( '%s/%s/step/%d' , [ aPrefix, Name, FCounter + 1                    ], AppLocale));
      aCallback( Self, Format( '%s/%s/gate/%f' , [ aPrefix, Name, FOutputs[ textsequencer_gateout]], AppLocale));
      aCallback( Self, Format( '%s/%s/steps/%d', [ aPrefix, Name, FSteps                          ], AppLocale));
    end;

    function    TModTextSequencer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        textsequencer_trig  : Result := 'trig';
        textsequencer_reset : Result := 'reset';
        textsequencer_dir   : Result := 'dir';
        textsequencer_mode  : Result := 'mode';
        textsequencer_chain : Result := 'chain';
        else begin
          if anIndex - textsequencer_step1 < FSteps
          then Result := Format( 'step%d', [ anIndex - textsequencer_step1 + 1], AppLocale)
          else Result := NO_INPUT;;
        end;
      end;
    end;

    function    TModTextSequencer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        textsequencer_out     : Result := 'out';
        textsequencer_gateout : Result := 'gateout';
        else                    Result := NO_OUTPUT;
      end;
    end;

    procedure   TModTextSequencer.SetStepValues( const aValues: TSignals);
    begin
      FSteps := aValues.Count;
      SetInputs( textsequencer_step1, aValues);
    end;


{ ========
  TModSeqStep = class( TMod)
  private
    FTime    : TSignal;
    FState   : TSeqStepState;
    FOldRes  : Boolean;
  public
}

const

  seqstep_triginfwd  = 0;
  seqstep_triginbwd  = 1;
  seqstep_chain      = 2;
  seqstep_reset      = 3;
  seqstep_dir        = 4;
  seqstep_note       = 5;
  seqstep_time       = 6;

  seqstep_trigoutfwd = 0;
  seqstep_trigoutbwd = 1;
  seqstep_out        = 2;


    constructor TModSeqStep.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSeqStep.SetDefaults; // override;
    begin
      inherited;
      FInputs[ seqstep_dir] := 1; // Make seq run forward by default.
      FDezipperMap := [
        seqstep_time
      ];
    end;

    procedure   TModSeqStep.DoSlowTick; // override;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        FState := sssIdle;
        FOutputs[ seqstep_trigoutfwd] := LogicToSignal( False);                 // Do not trigger next in forward  direction
        FOutputs[ seqstep_trigoutbwd] := LogicToSignal( False);                 // Do not trigger next in backward direction
      end;

      if not FOldRes and SignalToLogic( FInputs[ seqstep_reset])
      then begin                                                                // Reset from inactive to active
        FState := sssIdle;
        FOutputs[ seqstep_trigoutfwd] := LogicToSignal( False);                 // Do not trigger next in forward  direction
        FOutputs[ seqstep_trigoutbwd] := LogicToSignal( False);                 // Do not trigger next in backward direction
      end;

      case FState of

        sssIdle :
          begin
            if SignalToLogic( FInputs[ seqstep_dir])
            then begin                                                          // running Forwards
              if SignalToLogic( FInputs[ seqstep_triginfwd])                    // Forward trig in
              then begin
                FState := sssActive;
                FTime  := SlowTimeToSampleCount( FInputs[ seqstep_time]);       // Duration
                FOutputs[ seqstep_trigoutfwd] := LogicToSignal( False);         // Do not trigger next in forward  direction
                FOutputs[ seqstep_trigoutbwd] := LogicToSignal( False);         // Do not trigger next in backward direction
              end;
            end
            else begin                                                          // running Backwards
              if SignalToLogic( FInputs[ seqstep_triginbwd])                    // Backward trig in
              then begin
                FState := sssActive;
                FTime  := SlowTimeToSampleCount( FInputs[ seqstep_time]);       // Duration
                FOutputs[ seqstep_trigoutfwd] := LogicToSignal( False);         // Do not trigger next in forward  direction
                FOutputs[ seqstep_trigoutbwd] := LogicToSignal( False);         // Do not trigger next in backward direction
              end;
            end;
          end;

        sssActive :
          begin
            FTime := FTime - 1;
            if FTime <= 0
            then begin
              FState := sssWaiting;
              if SignalToLogic( FInputs[ seqstep_dir])                          // running Forwards
              then FOutputs[ seqstep_trigoutfwd] := LogicToSignal( True)        // Trigger next in forward  direction
              else FOutputs[ seqstep_trigoutbwd] := LogicToSignal( True);       // Trigger next in backward direction
            end;
          end;

        sssWaiting :
          begin
            if SignalToLogic( FInputs[ seqstep_dir])
            then begin                                                          // running Forwards
              if not SignalToLogic( FInputs[ seqstep_triginfwd])                // Forward trig in fell off
              then begin
                FState := sssIdle;
              end;
            end
            else begin                                                          // running Backwards
              if not SignalToLogic( FInputs[ seqstep_triginbwd])                // Backward trig in fell off
              then begin
                FState := sssIdle;
              end;
            end;
          end;

      end;

      if FState = sssActive
      then FOutputs[ seqstep_out] := FInputs[ seqstep_note] - MiddleNote / 128  // Own value to output
      else FOutputs[ seqstep_out] := FInputs[ seqstep_chain];                   // Chained value to output (was compensated there already)

      FOldRes := SignalToLogic( FInputs[ seqstep_reset]);
    end;

    procedure   TModSeqStep.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigfwd/%f', [ aPrefix, Name, FInputs[ seqstep_triginfwd]          ], AppLocale));
      aCallback( Self, Format( '%s/%s/trigbwd/%f', [ aPrefix, Name, FInputs[ seqstep_triginbwd]          ], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'  , [ aPrefix, Name, FInputs[ seqstep_reset    ]          ], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'    , [ aPrefix, Name, FInputs[ seqstep_dir      ]          ], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f' , [ aPrefix, Name, IfThen( FState = sssActive, 1.0, 0.0)], AppLocale));
    end;

    function    TModSeqStep.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqstep_triginfwd : Result := 'triginfwd';
        seqstep_triginbwd : Result := 'triginbwd';
        seqstep_chain     : Result := 'chain';
        seqstep_reset     : Result := 'reset';
        seqstep_dir       : Result := 'dir';
        seqstep_note      : Result := 'note';
        seqstep_time      : Result := 'time';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModSeqStep.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqstep_trigoutfwd : Result := 'trigoutfwd';
        seqstep_trigoutbwd : Result := 'trigoutbwd';
        seqstep_out        : Result := 'out';
        else                 Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSeqClockStep = class( TMod)
  private
    FCount      : Integer;
    FState      : TSeqStepState;
    FOldRes     : Boolean;
    FOldClock   : Boolean;
    FTrigInFwd  : Boolean;
    FTrigInBwd  : Boolean;
    FDir        : Boolean;
    FTrigOutFwd : Boolean;
    FTrigOutBwd : Boolean;
  public
}

const

  seqclockstep_triginfwd  = 0;
  seqclockstep_triginbwd  = 1;
  seqclockstep_chain      = 2;
  seqclockstep_reset      = 3;
  seqclockstep_dir        = 4;
  seqclockstep_clock      = 5;
  seqclockstep_note       = 6;
  seqclockstep_counts     = 7;

  seqclockstep_trigoutfwd = 0;
  seqclockstep_trigoutbwd = 1;
  seqclockstep_out        = 2;


    constructor TModSeqClockStep.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSeqClockStep.SetDefaults; // override;
    begin
      inherited;
      FInputs[ seqclockstep_dir] := 1; // Make seq run forward by default.
    end;

    procedure   TModSeqClockStep.DoSlowTick; // override;
    begin

      if ResetFlag
      then begin
        ResetFlag   := False;
        FState      := sssIdle;
        FTrigOutFwd := False;
        FTrigOutBwd := False;
        FOutputs[ seqclockstep_trigoutfwd] := LogicToSignal( False);                       // Do not trigger next in forward  direction
        FOutputs[ seqclockstep_trigoutbwd] := LogicToSignal( False);                       // Do not trigger next in backward direction
        FOutputs[ seqclockstep_out] := FInputs[ seqclockstep_chain];                       // Chained value to output
      end;

      if not FOldRes and SignalToLogic( FInputs[ seqclockstep_reset])
      then begin                                                                           // Reset from inactive to active - reset is asynchroneous
        FState      := sssIdle;
        FTrigOutFwd := False;
        FTrigOutBwd := False;
        FOutputs[ seqclockstep_trigoutfwd] := LogicToSignal( False);                       // Do not trigger next in forward  direction
        FOutputs[ seqclockstep_trigoutbwd] := LogicToSignal( False);                       // Do not trigger next in backward direction
        FOutputs[ seqclockstep_out] := FInputs[ seqclockstep_chain];                       // Chained value to output
      end;

      FTrigInFwd := SignalToLogic( FInputs[ seqclockstep_triginfwd]);
      FTrigInBwd := SignalToLogic( FInputs[ seqclockstep_triginbwd]);
      FDir       := SignalToLogic( FInputs[ seqclockstep_dir]);

      case FState of

        sssIdle :
          begin
            if ( FDir and FTrigInFwd) or (( not FDir) and FTrigInBwd)
            then begin                                                                     // Trigger seen
              FState      := sssActive;
              FCount      := Ceil( FInputs[ seqclockstep_counts]);                         // Duration
              FTrigOutFwd := False;                                                        // Do not trigger next in forward  direction
              FTrigOutBwd := False;                                                        // Do not trigger next in backward direction
            end;
          end;

        sssWaiting :
          begin
            if ( FDir and ( not FTrigInFwd)) or (( not FDir) and ( not FTrigInBwd))
            then begin
              FState      := sssIdle;                                                      // Trigger fell off - go idle
              FTrigOutFwd := False;                                                        // Do not trigger next in forward  direction
              FTrigOutBwd := False;                                                        // Do not trigger next in backward direction
            end;
          end;
      end;

      if not FOldClock and SignalToLogic( FInputs[ seqclockstep_clock])
      then begin                                                                           // Clock from inactive to active - state changes should only occur here
        case FState of
          sssActive :
            begin
              FCount := FCount - 1;
              if FCount <= 1
              then begin
                FCount := 0;
                FState := sssWaiting;
                if FDir                                                                    // running Forwards
                then begin
                  FTrigOutFwd := True;                                                     // Trigger next in forward  direction
                  FTrigOutBwd := False;                                                    // No Trigger next in backward  direction
                end
                else begin
                  FTrigOutFwd := False;                                                    // No Trigger next in forward direction
                  FTrigOutBwd := True;                                                     // Trigger next in backward direction
                end;
              end;
            end;
        end;
      end;

      FOutputs[ seqclockstep_trigoutfwd] := LogicToSignal( FTrigOutFwd);                   // Clock out internal state to ouptputs
      FOutputs[ seqclockstep_trigoutbwd] := LogicToSignal( FTrigOutBwd);

      if FState = sssActive
      then FOutputs[ seqclockstep_out] := FInputs[ seqclockstep_note] - MiddleNote / 128  // Own     value to output
      else FOutputs[ seqclockstep_out] := FInputs[ seqclockstep_chain];                    // Chained value to output (was compensated there already)

      FOldRes   := SignalToLogic( FInputs[ seqclockstep_triginbwd]);
      FOldClock := SignalToLogic( FInputs[ seqclockstep_clock    ]);
    end;


    procedure   TModSeqClockStep.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigfwd/%f', [ aPrefix, Name, FInputs[ seqclockstep_triginfwd]     ], AppLocale));
      aCallback( Self, Format( '%s/%s/trigbwd/%f', [ aPrefix, Name, FInputs[ seqclockstep_triginbwd]     ], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'  , [ aPrefix, Name, FInputs[ seqclockstep_reset    ]     ], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'    , [ aPrefix, Name, FInputs[ seqclockstep_dir      ]     ], AppLocale));
      aCallback( Self, Format( '%s/%s/clock/%f'  , [ aPrefix, Name, FInputs[ seqclockstep_clock    ]     ], AppLocale));
      aCallback( Self, Format( '%s/%s/count/%d'  , [ aPrefix, Name, FCount                               ], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f' , [ aPrefix, Name, IfThen( FState = sssActive, 1.0, 0.0)], AppLocale));
    end;

    function    TModSeqClockStep.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqclockstep_triginfwd : Result := 'triginfwd';
        seqclockstep_triginbwd : Result := 'triginbwd';
        seqclockstep_chain     : Result := 'chain';
        seqclockstep_reset     : Result := 'reset';
        seqclockstep_dir       : Result := 'dir';
        seqclockstep_clock     : Result := 'clock';
        seqclockstep_note      : Result := 'note';
        seqclockstep_counts    : Result := 'counts';
        else                     Result := NO_INPUT;
      end;
    end;

    function    TModSeqClockStep.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqclockstep_trigoutfwd : Result := 'trigoutfwd';
        seqclockstep_trigoutbwd : Result := 'trigoutbwd';
        seqclockstep_out        : Result := 'out';
        else                      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModCsvData = class( TMod)
  private
    FMemory   : Array[ 0 .. 15] of TSignals;
    FComments : TStringList;
    FFileName : string;
    FChannels : Integer;
    FCounter  : Integer;
    FSteps    : Integer;
    FInState  : TSeqInState;
    FOldRes   : Boolean;
    FMode     : Integer;
    FStopped  : Boolean;
  public
    property    FileName: string read FFileName write SetFileName;
  private
 }

const

  csvdata_reset =  0;
  csvdata_trig  =  1;
  csvdata_dir   =  2;
  csvdata_mode  =  3;

  csvdata_out1  =  0;
  csvdata_out2  =  1;
  csvdata_out3  =  2;
  csvdata_out4  =  3;
  csvdata_out5  =  4;
  csvdata_out6  =  5;
  csvdata_out7  =  6;
  csvdata_out8  =  7;
  csvdata_out9  =  8;
  csvdata_out10 =  9;
  csvdata_out11 = 10;
  csvdata_out12 = 11;
  csvdata_out13 = 12;
  csvdata_out14 = 13;
  csvdata_out15 = 14;
  csvdata_out16 = 15;


    procedure   TModCsvData.LoadFile;
    var
      aData  : TStringList;
      aparts : TStringList;
      aLine  : string;
      i      : Integer;
      p      : Integer;
      S      : string;
      T      : string;
      Cmt    : string;
      First  : Boolean;
    begin
      aData := TStringList.Create;
      try
        if FileExists( FFileName)
        then begin
          aData.LoadFromFile( FFileName);
          First := True;

          for i := Low( FMemory) to High( FMemory)
          do FMemory[ i].Clear;
          FComments.Clear;

          for aLine in aData
          do begin
            // Strip comment, but keep end of line comment for comment label in module
            p := Pos( '#', aLine);
            if p > 0
            then begin
              Cmt := Trim( Copy( aLine, p + 1, Length( aLine)));
              S   := Trim( Copy( aLine, 1, p - 1))
            end
            else begin
              Cmt := '';
              S   := Trim( aLine);
            end;

            // if line not empty after that handle it
            if Length( S) > 0
            then begin
              aParts := Explode( S, ',');
              try
                if First                           // First line determines column count
                then begin
                  FChannels := aParts.Count;
                  First     := False;
                end;
                for i := 0 to FChannels - 1
                do begin
                  if i > aParts.Count - 1
                  then FMemory[ i].AddSignal( 0)   // When not enough data on this line set remaining columns to zero
                  else begin
                    T := Trim( aParts[ i]);
                    FMemory[ i].AddSignal( StrToFloatDef( T, 0));
                  end;
                end;
              finally
                aParts.Free;
              end;
              FComments.Add( Cmt);
            end;
          end;
          FSteps   := FMemory[ 0].Count;
          FCounter := Min( FCounter, FSteps - 1);  // Do not address outside data area
        end;
      finally
        aData.Free;
      end;
    end;

    procedure   TModCsvData.SetFileName( const aValue: string);
    begin
      if  aValue <> FFileName
      then begin
        FFileName := aValue;
        LoadFile;
      end;
    end;

//  public

    constructor TModCsvData.Create( aParent: TSynthPatch; const aName: string); // override;
    var
      i : Integer;
    begin
      inherited;
      for i := Low( FMemory) to High( FMemory)
      do FMemory[ i] := TSignals.Create( 0);
      FComments := TStringList.Create;
      FIsFast := True;
      FIsSlow := True;
    end;

    destructor  TModCsvData.Destroy; // override;
    var
      i : Integer;
    begin
      FreeAndNil( FComments);
      for i := Low( FMemory) to High( FMemory)
      do FreeAndNil( FMemory[ i]);
      inherited;
    end;

    procedure   TModCsvData.SetDefaults; // override;
    begin
      inherited;
      FInputs[ csvdata_dir] := 1; // Make seq run forward by default.
    end;

    procedure   TModCsvData.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModCsvData.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FFileName
      else Result := inherited;
    end;

    procedure   TModCsvData.DoTick; // override;
    // Follow inputs at audio rate
    var
      OldCounter : Integer;
      Forwards   : Boolean;
    begin
      Forwards := SignalToLogic( FInputs[ csvdata_dir]);
      if ResetFlag or ( not FOldRes and SignalToLogic( FInputs[ csvdata_reset]))
      then begin
        ResetFlag := False;
        FInState  := seqClkLow;
        if Forwards
        then FCounter := 0
        else FCounter := FSteps - 1;
        FStopped   := False;
      end;
      FOldRes := SignalToLogic( FInputs[ csvdata_reset]);

      if ( FSteps > 0) and not FStopped
      then begin
        FMode      := SignalToInt( FInputs[ csvdata_mode]);
        OldCounter := FCounter;

        case FInState of
          seqClkLow :
            begin
              if SignalToLogic( FInputs[ csvdata_trig])
              then begin
                FInState := seqClkHigh;
                if Forwards
                then begin
                  FCounter := ( FCounter + 1) mod FSteps;
                  if ( FCounter < OldCounter) and ( FMode = 1)
                  then FStopped := True;
                end
                else begin
                  if FCounter <= 0
                  then begin
                    FCounter := FSteps - 1;
                    if FMode = 1
                    then FStopped := True;
                  end
                  else Dec( FCounter);
                end;
              end;
            end;
          seqClkHigh :
            begin
              if FInputs[ csvdata_trig] <= 0
              then FInState := seqClkLow;
            end;
        end;
      end;
    end;

    procedure   TModCsvData.DoSlowTick; // override;
    // generate output at control rate
    var
      i : Integer;
    begin
      if ( FSteps > 0) and ( FChannels > 0)
      then begin
        for i := csvdata_out1 to Min( FChannels - 1, csvdata_out16)
        do FOutputs[ i] := FMemory[ i - csvdata_out1].FSignals[ FCounter];
      end;
    end;

    procedure   TModCsvData.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'    , [ aPrefix, Name, FInputs[ textsequencer_trig ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'   , [ aPrefix, Name, FInputs[ textsequencer_reset]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'     , [ aPrefix, Name, FInputs[ textsequencer_dir  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/step/%d'    , [ aPrefix, Name, FCounter + 1                 ], AppLocale));
      aCallback( Self, Format( '%s/%s/steps/%d'   , [ aPrefix, Name, FSteps                       ], AppLocale));
      aCallback( Self, Format( '%s/%s/channels/%d', [ aPrefix, Name, FChannels                    ], AppLocale));
    end;

    procedure   TModCsvData.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    var
      Cmt : string;
    begin
      if FCounter < FComments.Count
      then Cmt := FComments[ FCounter]
      else Cmt := '';
      aCallback( Self, Format( '%s/%s/comment/%s' , [ aPrefix, Name, Cmt                          ], AppLocale));
    end;

    function    TModCsvData.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        csvdata_reset : Result := 'reset';
        csvdata_trig  : Result := 'trig';
        csvdata_dir   : Result := 'dir';
        csvdata_mode  : Result := 'mode';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModCsvData.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        csvdata_out1  : Result := 'out1';
        csvdata_out2  : Result := 'out2';
        csvdata_out3  : Result := 'out3';
        csvdata_out4  : Result := 'out4';
        csvdata_out5  : Result := 'out5';
        csvdata_out6  : Result := 'out6';
        csvdata_out7  : Result := 'out7';
        csvdata_out8  : Result := 'out8';
        csvdata_out9  : Result := 'out9';
        csvdata_out10 : Result := 'out10';
        csvdata_out11 : Result := 'out11';
        csvdata_out12 : Result := 'out12';
        csvdata_out13 : Result := 'out13';
        csvdata_out14 : Result := 'out14';
        csvdata_out15 : Result := 'out15';
        csvdata_out16 : Result := 'out16';
        else            Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModKleeBit = class( TMod)
  private
    FPattern    : Byte;
    FOldStep    : Boolean;
    FOldLoad    : Boolean;
    FBus1       : Boolean;
    FBus2       : Boolean;
    FBus3       : Boolean;
    FAnalog     : TSignal;
    FLoopMode   : Boolean;
    FInvertMode : Boolean;
  private
}

const

  kleebit_instep       =  0;
  kleebit_inload       =  1;
  kleebit_loopmode     =  2;
  kleebit_inanalog     =  3;
  kleebit_inbus1       =  4;
  kleebit_inbus2       =  5;
  kleebit_inbus3       =  6;
  kleebit_pre1         =  7;
  kleebit_pre2         =  8;
  kleebit_pre3         =  9;
  kleebit_pre4         = 10;
  kleebit_pre5         = 11;
  kleebit_pre6         = 12;
  kleebit_pre7         = 13;
  kleebit_pre8         = 14;
  kleebit_level1       = 15;
  kleebit_level2       = 16;
  kleebit_level3       = 17;
  kleebit_level4       = 18;
  kleebit_level5       = 19;
  kleebit_level6       = 20;
  kleebit_level7       = 21;
  kleebit_level8       = 22;
  kleebit_bussel1      = 23;
  kleebit_bussel2      = 24;
  kleebit_bussel3      = 25;
  kleebit_bussel4      = 26;
  kleebit_bussel5      = 27;
  kleebit_bussel6      = 28;
  kleebit_bussel7      = 29;
  kleebit_bussel8      = 30;
  kleebit_inbit        = 31;
  kleebit_invertmode   = 32;
  kleebit_mainlevel    = 33;
  kleebit_loopinvert   = 34;
  kleebit_invertinvert = 35;
  kleebit_mainlevelmod = 36;

  kleebit_outanalog    =  0;
  kleebit_outbus1      =  1;
  kleebit_outbus2      =  2;
  kleebit_outbus3      =  3;
  kleebit_outbit       =  4;
  kleebit_outloop      =  5;
  kleebit_outinvert    =  6;


    procedure   TModKleeBit.LoadPreset;
    var
      i : Integer;
    begin
      FPattern := 0;
      for i := kleebit_pre1 to kleebit_pre8
      do begin
        FPattern := FPattern shr 1;
        if SignalToLogic( FInputs[ i])
        then FPattern := FPattern or 128;
      end;
    end;

    procedure   TModKleeBit.Step;
    var
      InBit   : Boolean;
      Carry   : Boolean;
      i       : Integer;
      Bit     : Boolean;
      MainLev : TSignal;
    begin
      FInvertMode := SignalToLogic( FInputs[ kleebit_invertmode]) xor SignalToLogic( FInputs[ kleebit_invertinvert]);
      FLoopMode   := SignalToLogic( FInputs[ kleebit_loopmode  ]) xor SignalToLogic( FInputs[ kleebit_loopinvert  ]);

      Carry := ( FPattern and 128 <> 0) xor FInvertMode;
      if FLoopMode
      then InBit := Carry
      else InBit := SignalToLogic( FInputs[ kleebit_inbit]);

      FPattern := FPattern shl 1;
      if Inbit
      then FPattern := FPattern or 1;

      FBus1   := SignalToLogic( FInputs[ kleebit_inbus1   ]);
      FBus2   := SignalToLogic( FInputs[ kleebit_inbus2   ]);
      FBus3   := SignalToLogic( FInputs[ kleebit_inbus3   ]);
      FAnalog :=                FInputs[ kleebit_inanalog ];
      MainLev :=                FInputs[ kleebit_mainlevel] * FInputs[ kleebit_mainlevelmod];

      for i := 0 to 7
      do begin
        Bit := FPattern and ( 1 shl i) <> 0;
        if Bit
        then FAnalog := FAnalog + FInputs[ kleebit_level1 + i] * MainLev;

        case SignalToInt( FInputs[ kleebit_bussel1 + i]) of
          0 : FBus1 := FBus1 or Bit;
          1 : FBus2 := FBus2 or Bit;
          2 : FBus3 := FBus3 or Bit;
        end;
      end;

      FOutputs[ kleebit_outbit   ] := LogicToSignal( Carry);
      FOutputs[ kleebit_outanalog] := FAnalog;
      FOutputs[ kleebit_outbus1  ] := LogicToSignal( FBus1      );
      FOutputs[ kleebit_outbus2  ] := LogicToSignal( FBus2      );
      FOutputs[ kleebit_outbus3  ] := LogicToSignal( FBus3      );
      FOutputs[ kleebit_outloop  ] := LogicToSignal( FLoopMode  );
      FOutputs[ kleebit_outinvert] := LogicToSignal( FInvertMode);
    end;

//  public

    constructor TModKleeBit.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModKleeBit.SetDefaults; // override;
    begin
      FInputs[ kleebit_inanalog    ] := 0.0;
      FInputs[ kleebit_inbus1      ] := 0.0;
      FInputs[ kleebit_inbus2      ] := 0.0;
      FInputs[ kleebit_inbus3      ] := 0.0;
      FInputs[ kleebit_instep      ] := 0.0;
      FInputs[ kleebit_inload      ] := 0.0;
      FInputs[ kleebit_loopinvert  ] := 0.0;
      FInputs[ kleebit_invertinvert] := 0.0;
      FInputs[ kleebit_mainlevelmod] := 1.0;
      LoadPreset;
    end;

    procedure   TModKleeBit.DoSlowTick; // override;
    var
      aValue : Boolean;
    begin
      if ResetFlag
      then SetDefaults;
      ResetFlag := False;

      aValue := SignalToLogic( FInputs[ kleebit_inload]);
      if not FOldLoad and aValue
      then LoadPreset;
      FOldLoad := aValue;

      aValue := SignalToLogic( FInputs[ kleebit_instep]);
      if not FOldStep and aValue
      then Step;
      FOldStep := aValue;
    end;

    procedure   TModKleeBit.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/step1/%f' , [ aPrefix, Name, LogicToSignal( FPattern and   1 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step2/%f' , [ aPrefix, Name, LogicToSignal( FPattern and   2 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step3/%f' , [ aPrefix, Name, LogicToSignal( FPattern and   4 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step4/%f' , [ aPrefix, Name, LogicToSignal( FPattern and   8 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step5/%f' , [ aPrefix, Name, LogicToSignal( FPattern and  16 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step6/%f' , [ aPrefix, Name, LogicToSignal( FPattern and  32 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step7/%f' , [ aPrefix, Name, LogicToSignal( FPattern and  64 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/step8/%f' , [ aPrefix, Name, LogicToSignal( FPattern and 128 <> 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/loop/%f'  , [ aPrefix, Name, LogicToSignal( FLoopMode            )], AppLocale));
      aCallback( Self, Format( '%s/%s/invert/%f', [ aPrefix, Name, LogicToSignal( FInvertMode          )], AppLocale));
      aCallback( Self, Format( '%s/%s/bus1/%f'  , [ aPrefix, Name, FOutputs[ kleebit_outbus1]           ], AppLocale));
      aCallback( Self, Format( '%s/%s/bus2/%f'  , [ aPrefix, Name, FOutputs[ kleebit_outbus2]           ], AppLocale));
      aCallback( Self, Format( '%s/%s/bus3/%f'  , [ aPrefix, Name, FOutputs[ kleebit_outbus3]           ], AppLocale));
      aCallback( Self, Format( '%s/%s/outbit/%f', [ aPrefix, Name, FOutputs[ kleebit_outbit ]           ], AppLocale));
      aCallback( Self, Format( '%s/%s/step/%f'  , [ aPrefix, Name, FInputs [ kleebit_instep ]           ], AppLocale));
      aCallback( Self, Format( '%s/%s/load/%f'  , [ aPrefix, Name, FInputs [ kleebit_inload ]           ], AppLocale));
      aCallback( Self, Format( '%s/%s/inbit/%f' , [ aPrefix, Name, FInputs [ kleebit_inbit  ]           ], AppLocale));
    end;

    function    TModKleeBit.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        kleebit_instep       : Result := 'instep';
        kleebit_inload       : Result := 'inload';
        kleebit_loopmode     : Result := 'loopmode';
        kleebit_inanalog     : Result := 'inanalog';
        kleebit_inbus1       : Result := 'inbus1';
        kleebit_inbus2       : Result := 'inbus2';
        kleebit_inbus3       : Result := 'inbus3';
        kleebit_pre1         : Result := 'pre1';
        kleebit_pre2         : Result := 'pre2';
        kleebit_pre3         : Result := 'pre3';
        kleebit_pre4         : Result := 'pre4';
        kleebit_pre5         : Result := 'pre5';
        kleebit_pre6         : Result := 'pre6';
        kleebit_pre7         : Result := 'pre7';
        kleebit_pre8         : Result := 'pre8';
        kleebit_level1       : Result := 'level1';
        kleebit_level2       : Result := 'level2';
        kleebit_level3       : Result := 'level3';
        kleebit_level4       : Result := 'level4';
        kleebit_level5       : Result := 'level5';
        kleebit_level6       : Result := 'level6';
        kleebit_level7       : Result := 'level7';
        kleebit_level8       : Result := 'level8';
        kleebit_bussel1      : Result := 'bussel1';
        kleebit_bussel2      : Result := 'bussel2';
        kleebit_bussel3      : Result := 'bussel3';
        kleebit_bussel4      : Result := 'bussel4';
        kleebit_bussel5      : Result := 'bussel5';
        kleebit_bussel6      : Result := 'bussel6';
        kleebit_bussel7      : Result := 'bussel7';
        kleebit_bussel8      : Result := 'bussel8';
        kleebit_inbit        : Result := 'inbit';
        kleebit_invertmode   : Result := 'invertmode';
        kleebit_mainlevel    : Result := 'mainlevel';
        kleebit_loopinvert   : Result := 'loopinvert';
        kleebit_invertinvert : Result := 'invertinvert';
        kleebit_mainlevelmod : Result := 'mainlevelmod';
        else                   Result := NO_INPUT;
      end;
    end;

    function    TModKleeBit.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        kleebit_outanalog : Result := 'outanalog';
        kleebit_outbus1   : Result := 'outbus1';
        kleebit_outbus2   : Result := 'outbus2';
        kleebit_outbus3   : Result := 'outbus3';
        kleebit_outbit    : Result := 'outbit';
        kleebit_outloop   : Result := 'outloop';
        kleebit_outinvert : Result := 'outinvert';
        else                Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSeqRandom = class( TMod)
  private
   FValues       : array[ Byte] of TMidiSeptet;
    FPattern      : array[ Byte] of TMidiSeptet;
    FCount        : Integer;
    FScale        : Integer;
    FBaseNote     : TBaseNote;
    FLowNote      : TMidiSeptet;
    FHighNote     : TMidiSeptet;
    FNoteCount    : Integer;
    FScaleNotes   : array[ TMidiSeptet] of TMidiSeptet;
    FOldTrig      : Boolean;
    FOldOne       : Boolean;
    FPosition     : Byte;
    FDistribution : TSignal;
   private
}


const

  seqrandom_trig       = 0;
  seqrandom_changeone  = 1;
  seqrandom_scale      = 2;
  seqrandom_basenote   = 3;
  seqrandom_lownote    = 4;
  seqrandom_highnote   = 5;
  seqrandom_distmod    = 6;
  seqrandom_distmodamt = 7;
  seqrandom_dist       = 8;
  seqrandom_chain      = 9;

  seqrandom_out        = 0;

  seqrandom_count      = 0;

    function    TModSeqRandom.Transpose( aScale: TScale; anAmount: Integer): TScale;
    var
      i : Integer;
    begin
      Result := [];
      for i := 0 to 11
      do begin
        if i in aScale
        then Result := Result + [ ( i + anAmount) mod 12];
      end;
    end;

    function    TModSeqRandom.Complete( aScale: TScale): TScale;
    var
      i : Integer;
      j : Integer;
    begin
      if ( FLowNote = 0) and ( FHighNote = 0)
      then nop;

      Result := [];
      for i := 0 to 11
      do begin
        if i in aScale
        then begin
          j := i;
          while j <= 127
          do begin
            if ( j >= FLowNote) and ( j <= FHighNote)
            then Result := Result + [ j];
            Inc( j, 12);
          end;
        end;
      end;
    end;

    procedure   TModSeqRandom.MakeScale;
    var
      aScale : TScale;
      i      : TMidiSeptet;
    begin
      aScale := ScaleIndexToData( FScale);
      aScale := Transpose( aScale, FBaseNote);
      aScale := Complete( aScale);
      FNoteCount := 0;
      for i := Low( i) to High( i)
      do begin
        if i in aScale
        then begin
          FScaleNotes[ FNoteCount] := i;
          Inc( FNoteCount);
        end;
      end;
    end;

    function    TModSeqRandom.RandomNote( aPosition: Byte; aDistribution: TSignal): TMidiSeptet;
    var
      aRandomValue : TMidiSeptet;
      anOldValue   : TMidiSeptet;
      aNewValue    : TMidiSeptet;
    begin
      aRandomValue := Random( FNoteCount);
      anOldValue   := FValues[ aPosition];
      aNewValue    := Round( aDistribution * aRandomValue + ( 1 - aDistribution) * anOldValue);
      Result       := FScaleNotes[ aNewValue];
      FValues[ aPosition] := aNewValue;
    end;

    procedure   TModSeqRandom.ChangeScale( aScale: Integer);
    begin
      FScale := aScale;
      MakeScale;
      ChangeAllNotes;
    end;

    procedure   TModSeqRandom.ChangeOneNote( aPosition: Byte);
    begin
      FPattern[ FPosition] := RandomNote( aPosition, FDistribution); // correlated random selection
    end;

    procedure   TModSeqRandom.ChangeAllNotes;
    var
      i : Integer;
    begin
      for i := Low( FPattern) to High( FPattern)
      do FPattern[ i] := RandomNote( i, 1.0);       // totally random selection
    end;

    procedure   TModSeqRandom.Step;
    begin
      if FPosition < FCount - 1
      then Inc( FPosition)
      else FPosition := Low( FPosition);
    end;

//  public

    constructor TModSeqRandom.Create( aParent: TSynthPatch; const aName: string); //override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModSeqRandom.SetDefaults; // override;
    begin
      ChangeScale( 1);
      ChangeScale( 0);
    end;

    procedure   TModSeqRandom.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'count')
      then FCount := Clip( Ceil( aValue), 1, 256)
      else inherited;
    end;

    function    TModSeqRandom.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'count')
      then Result := FCount
      else Result := inherited;
    end;

    procedure   TModSeqRandom.DoSlowTick; // override;
    var
      aBool   : Boolean;
      anInt   : Integer;
      aChange : Boolean;
      aScale  : Integer;
    begin
      if ResetFlag
      then FPosition := 0;
      ResetFlag := False;

      FDistribution := Clip( FInputs[ seqrandom_dist] + FInputs[ seqrandom_distmod] * FInputs[ seqrandom_distmodamt], 0.0, 1.0);

      aBool := False;
      anInt := SignalToInt( 12 * FInputs[ seqrandom_basenote]);
      if anInt <> FBaseNote
      then begin
        aBool     := True;
        FBaseNote := anInt;
      end;
      anInt := SignalToInt( UnitsToNoteNumber( FInputs[ seqrandom_lownote]));
      if anInt <> FLowNote
      then begin
        aBool    := True;
        FLowNote := anInt;
      end;
      anInt := SignalToInt( UnitsToNoteNumber( FInputs[ seqrandom_highnote]));
      if anInt <> FHighNote
      then begin
        aBool     := True;
        FHighNote := anInt;
      end;

      aScale := SignalToInt( FInputs[ seqrandom_scale]);
      if aBool or ( aScale <> FScale)
      then ChangeScale( aSCale);


      aBool := SignalToLogic( FInputs[ seqrandom_trig]);
      if not FOldTrig and aBool
      then begin
        aChange := SignalToLogic( FInputs[ seqrandom_changeone]);
        if not FOldOne and aChange
        then ChangeOneNote( FPosition);
        FOldOne := aChange;
        Step;
      end;
      FOldTrig := aBool;

      FOutputs[ seqrandom_out] := NoteNumberToUnits( FPattern[ FPosition] - MiddleNote) + FInputs[ seqrandom_chain];
    end;

    procedure   TModSeqRandom.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'     , [ aPrefix, Name, FInputs[ seqrandom_trig     ]], AppLocale));
      aCallback( Self, Format( '%s/%s/changeone/%f', [ aPrefix, Name, FInputs[ seqrandom_changeone]], AppLocale));
    end;

    function    TModSeqRandom.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqrandom_trig       : Result := 'trig';
        seqrandom_changeone  : Result := 'changeone';
        seqrandom_scale      : Result := 'scale';
        seqrandom_basenote   : Result := 'basenote';
        seqrandom_lownote    : Result := 'lownote';
        seqrandom_highnote   : Result := 'highnote';
        seqrandom_distmod    : Result := 'distmod';
        seqrandom_distmodamt : Result := 'distmodamt';
        seqrandom_dist       : Result := 'dist';
        seqrandom_chain      : Result := 'chain';
        else                   Result := NO_INPUT;
      end;
    end;

    function    TModSeqRandom.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqrandom_out : Result := 'out';
        else            Result := NO_OUTPUT;
      end;
    end;

    function    TModSeqRandom.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        seqrandom_count : Result := 'count';
        else              Result := NO_INTERNAL;
      end;
    end;


{ ========
  TMorseSymbol = (
    msDot,
    msDash,
    msInterSymbol,
    msShortGap,
    msMediumGap
  );

  TMorseSymbols = array of TMorseSymbol;

  TModMorse = class( TMod)
  private
    FMessage     : string;
    FSymbols     : string;
    FSymbolIndex : Integer;
    FOldReset    : Boolean;
    FOldClock    : Boolean;
  public
    property    Message: string read FMessage write SetMessage;
  public
}

const

  morse_reset = 0;
  morse_clock = 1;

  morse_out   = 0;


    function    TModMorse.CharToSymbols( aChar: Char): string;
    // Using ITU encoding
    begin
      case UpCase( aChar) of
        'A' : Result := '_ ___   ';                   'B' : Result := '___ _ _ _   ';
        'C' : Result := '___ _ ___ _   ';             'D' : Result := '___ _ _   ';
        'E' : Result := '_   ';                       'F' : Result := '_ _ ___ _   ';
        'G' : Result := '___ ___ _   ';               'H' : Result := '_ _ _ _   ';
        'I' : Result := '_ _   ';                     'J' : Result := '_ ___ ___ ___   ';
        'K' : Result := '___ _ ___   ';               'L' : Result := '_ ___ _ _   ';
        'M' : Result := '___ ___   ';                 'N' : Result := '___ _   ';
        'O' : Result := '___ ___ ___   ';             'P' : Result := '_ ___ ____   ';
        'Q' : Result := '___ ___ _ ___   ';           'R' : Result := '_ ___ _   ';
        'S' : Result := '_ _ _   ';                   'T' : Result := '___   ';
        'U' : Result := '_ _ ___   ';                 'V' : Result := '_ _ _ ___   ';
        'W' : Result := '_ ___ ___   ';               'X' : Result := '___ _ _ ___   ';
        'Y' : Result := '___ _ ___ ___   ';           'Z' : Result := '___ ___ _ _   ';
        '0' : Result := '___ ___ ___ ___ ___   ';     '1' : Result := '_ ___ ___ ___ ___   ';
        '2' : Result := '_ _ ___ ___ ___   ';         '3' : Result := '_ _ _ ___ ___   ';
        '4' : Result := '_ _ _ ____   ';              '5' : Result := '_ _ _ _ _   ';
        '6' : Result := '____ _ _ _   ';              '7' : Result := '___ ___ _ _ _   ';
        '8' : Result := '___ ___ ___ _ _   ';         '9' : Result := '___ ___ ___ ___ _   ';
        '*' : Result := '_ _ _ ___ ___ ___ _ _ _   '; // SOS symbol
        else  Result := '       ';                    // Inter word delay, for all unknown chars including a space
      end;
    end;

    procedure   TModMorse.MessageToSymbols;
    var
      i : Integer;
    begin
      FSymbols := '';
      for i := 1 to Length( FMessage)
      do FSymbols := FSymbols + CharToSymbols( FMessage[ i]);
    end;

    procedure   TModMorse.SetMessage( const aValue: string);
    begin
      if aValue <> FMessage
      then begin
        FMessage := aValue;
        MessageToSymbols;
      end;
    end;

//  public

    constructor TModMorse.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
    end;

    procedure   TModMorse.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'text')
      then Message := aValue
      else inherited
    end;

    function    TModMorse.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'text')
      then result := Message
      else Result := inherited;
    end;

    procedure   TModMorse.DoSlowTick; // override;
    var
      NewReset : Boolean;
      NewClock : Boolean;
    begin
      NewReset := SignalToLogic( FInputs[ morse_reset]);
      NewClock := SignalToLogic( FInputs[ morse_clock]);

      if ( not FOldReset and NewReset) or ResetFlag or (  FSymbolIndex > Length( FSymbols))
      then begin
        ResetFlag            := False;
        FSymbolIndex         := 1;
        FOutputs[ morse_out] := 0;
      end;

      if not FOldClock and NewClock
      then begin
        if ( FSymbols = '') or ( FSymbols[ FSymbolIndex] = ' ')
        then FOutputs[ morse_out] := 0
        else FOutputs[ morse_out] := 1;
        Inc( FSymbolIndex);
      end;

      FOldReset := NewReset;
      FOldClock := NewClock;
    end;

    procedure   TModMorse.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, LogicToSignal( FOldReset)], AppLocale));
      aCallback( Self, Format( '%s/%s/clock/%f', [ aPrefix, Name, LogicToSignal( FOldClock)], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f'  , [ aPrefix, Name, FOutputs[ morse_out]     ], AppLocale));
    end;

    function    TModMorse.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        morse_reset : Result := 'reset';
        morse_clock : Result := 'clock';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModMorse.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        morse_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModGateSeq = class( TMod)
  private
    FOutValue   : TSignal;
    FSteps      : Integer;
    FOldRes     : Boolean;
    FCounter    : Integer;
    FInState    : TSeqInState;
    FForwards   : Boolean;
    FReset      : Boolean;
  private
}

const

  gateseq_trig    =  0;
  gateseq_step1   =  1;
  gateseq_step2   =  2;
  gateseq_step3   =  3;
  gateseq_step4   =  4;
  gateseq_step5   =  5;
  gateseq_step6   =  6;
  gateseq_step7   =  7;
  gateseq_step8   =  8;
  gateseq_step9   =  9;
  gateseq_step10  = 10;
  gateseq_step11  = 11;
  gateseq_step12  = 12;
  gateseq_step13  = 13;
  gateseq_step14  = 14;
  gateseq_step15  = 15;
  gateseq_step16  = 16;
  gateseq_chain   = 17;
  gateseq_res     = 18;
  gateseq_active  = 19;
  gateseq_dir     = 20;

  gateseq_out     =  0;
  gateseq_outres  =  1;
  gateseq_outtrig =  2;
  gateseq_outdir  =  3;

  gateseq_steps  =  0;


    procedure   TModGateSeq.SampleInputs;
    begin
      FInState  := seqClkHigh;
      FForwards := SignalToLogic( FInputs[ gateseq_dir]);
      FReset    := SignalToLogic( FInputs[ gateseq_res]);
    end;

//  public

    constructor TModGateSeq.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModGateSeq.SetDefaults; // override;
    begin
      inherited;
      FOutValue                 := 0;
      FInputs [ gateseq_chain ] := 0; // Chain input to zero
      FInputs [ gateseq_active] := 1; // Sequencer active by default
      FInputs [ gateseq_dir   ] := 1; // Forwards by default
      SampleInputs;
      if FForwards
      then FCounter := 0
      else FCounter := FSteps - 1;
    end;

    procedure   TModGateSeq.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'steps')
      then FSteps := Clip( Ceil( aValue), 1, 16)
      else inherited;
    end;

    function    TModGateSeq.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := FSteps
      else Result := inherited;
    end;

    procedure   TModGateSeq.DoTick; // override;
    var
      Mode : Integer;
    begin
      if ResetFlag
      then begin
        SampleInputs;
        if FForwards
        then FCounter := 0
        else FCounter := FSteps - 1;
      end
      else begin
        case FInState of
          seqClkLow :
            begin
              if SignalToLogic( FInputs[ gateseq_trig])
              then begin
                SampleInputs;
                if ResetFlag or ( not FOldRes and FReset)
                then begin
                  if FForwards
                  then FCounter := 0
                  else FCounter := FSteps - 1;
                end
                else begin
                  if FSteps > 0
                  then begin
                    SampleInputs;
                    if FForwards
                    then FCounter := ( FCounter + 1) mod FSteps
                    else begin
                      if FCounter <= 0
                      then FCounter := FSteps - 1
                      else Dec( FCounter);
                    end;
                  end;
                end;
                Mode := SignalToInt( FInputs[ gateseq_step1 + FCounter]);
                if ( Mode in [ 1, 2]) and SignalToLogic( FInputs[ gateseq_active])
                then FOutValue := 1.0
                else FOutValue := FInputs[ gateseq_chain];
              end;
            end;
          seqClkHigh :
            begin
              if not SignalToLogic( FInputs[ gateseq_trig])
              then begin
                FInState := seqClkLow;
                Mode := SignalToInt( FInputs[ gateseq_step1 + FCounter]);
                if ( Mode = 2) and SignalToLogic( FInputs[ gateseq_active])
                then FOutValue := 1.0
                else FOutValue := FInputs[ gateseq_chain];
              end;
            end;
        end; // case
      end;
      ResetFlag := False;
      FOldRes   := FReset;
      FOutputs[ gateseq_out    ] := FOutValue;
      FOutputs[ gateseq_outres ] := FInputs[ gateseq_res ];
      FOutputs[ gateseq_outtrig] := FInputs[ gateseq_trig];
      FOutputs[ gateseq_outdir ] := FInputs[ gateseq_dir ];
    end;

    procedure   TModGateSeq.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f'  , [ aPrefix, Name, FInputs [ gateseq_trig  ]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f' , [ aPrefix, Name, LogicToSignal( FOldRes  )], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f', [ aPrefix, Name, FInputs [ gateseq_active]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'   , [ aPrefix, Name, FInputs [ gateseq_dir   ]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f'   , [ aPrefix, Name, FOutputs[ gateseq_out   ]], AppLocale));
      for i := gateseq_step1 to gateseq_step16
      do begin
        if ( i = FCounter + gateseq_step1) and SignalToLogic( FInputs[ gateseq_active])
        then aCallback( Self, Format( '%s/%s/step%d/1.0', [ aPrefix, Name, i - gateseq_step1 + 1], AppLocale))
        else aCallback( Self, Format( '%s/%s/step%d/0.0', [ aPrefix, Name, i - gateseq_step1 + 1], AppLocale));
      end;
    end;

    function    TModGateSeq.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        gateseq_trig   : Result := 'trig';
        gateseq_step1  : Result := 'step1';
        gateseq_step2  : Result := 'step2';
        gateseq_step3  : Result := 'step3';
        gateseq_step4  : Result := 'step4';
        gateseq_step5  : Result := 'step5';
        gateseq_step6  : Result := 'step6';
        gateseq_step7  : Result := 'step7';
        gateseq_step8  : Result := 'step8';
        gateseq_step9  : Result := 'step9';
        gateseq_step10 : Result := 'step10';
        gateseq_step11 : Result := 'step11';
        gateseq_step12 : Result := 'step12';
        gateseq_step13 : Result := 'step13';
        gateseq_step14 : Result := 'step14';
        gateseq_step15 : Result := 'step15';
        gateseq_step16 : Result := 'step16';
        gateseq_chain  : Result := 'chain';
        gateseq_res    : Result := 'res';
        gateseq_active : Result := 'active';
        gateseq_dir    : Result := 'dir';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModGateSeq.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        gateseq_out     : Result := 'out';
        gateseq_outres  : Result := 'outres';
        gateseq_outtrig : Result := 'outtrig';
        gateseq_outdir  : Result := 'outdir';
        else              Result := NO_OUTPUT;
      end;
    end;

    function    TModGateSeq.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        gateseq_steps : Result := 'steps';
        else            Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModMemory = class( TMod)
  private
    FMemoryL : TSignals;
    FMemoryR : TSignals;
  public
}

    constructor TModMemory.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FMemoryL := TSignals.Create( 0);
      FMemoryR := TSignals.Create( 0);
      FIsFast := True;
      FIsSlow := False;
    end;

    destructor  TModMemory.Destroy; // override;
    begin
      FreeAndNil( FMemoryL);
      FreeAndNil( FMemoryR);
      inherited;
    end;

    function    TModMemory.Convolve( aSample: TSignal): TSignalPair;
    begin
      Result.Left  := FMemoryL.Convolve( aSample);
      Result.Right := FMemoryR.Convolve( aSample);
    end;


{ ========
  TModDelay = class( TModMemory)
  private
    FLength    : TSignal;
    FInPointer : Integer;
  public
}

const

  delay_in       = 0;
  delay_delay    = 1;

  delay_out      = 0;
  delay_outfixed = 1;

  delay_length   = 0;


    procedure   TModDelay.SetDefaults; // override;
    begin
      inherited;
      FInputs[ delay_delay] := 1.0; // Set maximum delay time if nothing connected
    end;

    procedure   TModDelay.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      NewMem: TSignals;
      OldMem: TSignals;
    begin
      if SameText( aName, 'length')
      then begin
        if aValue < 1e-6
        then aValue := 1e-6;
        if aValue <> FMemLength
        then begin
          NewMem     := TSignals.Create( Ceil( TimeToSampleCount( aValue)));
          OldMem     := FMemoryL;
          FMemLength := 0;
          FMemoryL   := NewMem;
          FMemLength := aValue;

          FreeAndNil( OldMem);
        end;
      end
      else inherited;
    end;

    function    TModDelay.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'length')
      then Result := FMemLength
      else Result := inherited;
    end;

    procedure   TModDelay.DoTick; // override;
    var
      OutPointerVar : TSignal;
      SampCount     : Integer;
    begin
      if Assigned( FMemoryL) and ( FMemoryL.Count > 0)
      then begin
        SampCount                    := FMemoryL.Count;
        FInPointer                   := FInPointer mod SampCount;
        FOutputs[ delay_outfixed]    := FMemoryL.Signal[ FInPointer];
        FMemoryL.Signal[ FInPointer] := Clip( FInputs[ delay_in], -16, 16);
        OutPointerVar                := FMod( SampCount + FInPointer - Clip( FInputs[ delay_delay], 0.0, 1.0) * ( TimeToSampleCount( FMemLength) - 1), SampCount);
        FOutputs[ delay_out]         := FMemoryL.Interpolate( OutPointerVar);
        FInPointer                   := ( FInPointer + 1) mod SampCount;
      end;
    end;

    function    TModDelay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        delay_in    : Result := 'in';
        delay_delay : Result := 'delay';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModDelay.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        delay_out      : Result := 'out';
        delay_outfixed : Result := 'outfixed';
        else             Result := NO_OUTPUT;
      end;
    end;

    function    TModDelay.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        delay_length :  Result := 'length';
        else            Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModClockedDelay = class( TModMemory)
  private
    FLength    : TSignal;
    FInPointer : Integer;
  public
}

const

  clockeddelay_in       = 0;
  clockeddelay_clock    = 1;

  clockeddelay_out      = 0;

  clockeddelay_length   = 0;


    procedure   TModClockedDelay.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      NewMem: TSignals;
      OldMem: TSignals;
    begin
      if SameText( aName, 'length')
      then begin
        if aValue < 1
        then aValue := 1;
        if aValue <> FMemLength
        then begin
          NewMem     := TSignals.Create( SignalToInt( aValue));
          OldMem     := FMemoryL;
          FMemLength := 0;
          FMemoryL   := NewMem;
          FMemLength := aValue;

          FreeAndNil( OldMem);
        end;
      end
      else inherited;
    end;

    function    TModClockedDelay.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'length')
      then Result := FMemLength
      else Result := inherited;
    end;

    procedure   TModClockedDelay.DoTick; // override;
    var
      SampCount : Integer;
      Clock     : Boolean;
    begin
      if ResetFlag
      then begin
        if Assigned( FMemoryL)
        then FMemoryL.Erase;
        ResetFlag := False;
      end;
      if Assigned( FMemoryL) and ( FMemoryL.Count > 0)
      then begin
        Clock := SignalToLogic( FInputs[ clockeddelay_clock]);
        if not FOldClock and Clock
        then begin
          SampCount                    := FMemoryL.Count;
          FInPointer                   := FInPointer mod SampCount;
          FOutputs[ clockeddelay_out]  := FMemoryL.Signal[ FInPointer];
          FMemoryL.Signal[ FInPointer] := FInputs[ clockeddelay_in];
          FInPointer                   := ( FInPointer + 1) mod SampCount;
        end;
        FOldClock := Clock;
      end;
    end;

    procedure   TModClockedDelay.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/clock/%f', [ aPrefix, Name, LogicToSignal( FOldClock)], AppLocale));
    end;

    function    TModClockedDelay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        clockeddelay_in    : Result := 'in';
        clockeddelay_clock : Result := 'clock';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModClockedDelay.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        clockeddelay_out : Result := 'out';
        else               Result := NO_OUTPUT;
      end;
    end;

    function    TModClockedDelay.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        clockeddelay_length :  Result := 'length';
        else                   Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModWave = class( TModMemory)
  private
    FFileName    : string;
    FDuration    : string;
    FIsStereo    : Boolean;
  public
    property    FileName: string read FFileName write SetFileName;
  private
}

const

  wave_outl = 0;
  wave_outr = 1;


    procedure   TModWave.SetDuration( aValue: Cardinal);
    begin
      if aValue < 1000
      then FDuration := Format( '%.3f ms', [ aValue / 1.0], AppLocale)
      else if aValue >= 10000
      then FDuration := Format( '%.2d:%.2d', [ Round( aValue / 1000) div 60, Round( aValue / 1000) mod 60])
      else FDuration := Format( '%.3f s', [ aValue / 1000.0], AppLocale);
    end;

    procedure   TModWave.SwapMemoryFor( aMemory: TSignalArray);
    var
      aNewMemoryL  : TSignals;
      anOldMemoryL : TSignals;
      aNewMemoryR  : TSignals;
      anOldMemoryR : TSignals;
      i            : Integer;
    begin
      if FIsStereo
      then begin
        aNewMemoryL  := TSignals.Create( Length( aMemory) div 2);
        aNewMemoryR  := TSignals.Create( Length( aMemory) div 2);
        for i := 0 to aNewMemoryL.Count - 1
        do begin
          aNewMemoryL.Signal[ i] := aMemory[ 2 * i    ];
          aNewMemoryR.Signal[ i] := aMemory[ 2 * i + 1];
        end;

        anOldMemoryL := FMemoryL;
        anOldMemoryR := FMemoryR;
        FMemoryL     := aNewMemoryL;
        FMemoryR     := aNewMemoryR;

        FreeAndNil( anOldMemoryL);
        FreeAndNil( anOldMemoryR);
      end
      else begin
        aNewMemoryL  := TSignals.Create( Length( aMemory));
        for i := 0 to Length( aMemory) - 1
        do aNewMemoryL.Signal[ i] := aMemory[ i];

        anOldMemoryL := FMemoryL;
        FMemoryL     := aNewMemoryL;

        FreeAndNil( anOldMemoryL);
      end;
    end;

    procedure   TModWave.LoadFile;
    const
      BufSize = 8192;
    type
      TSample = SmallInt;
      TBuffer = array[ 0 .. BufSize - 1] of TSample;

      procedure SetSamples16bit44100Hz( var aMemory: TSignalArray; aBuf: TBuffer; aSampleCount: Cardinal);
      const
        MaxVal = High( TSample) / 2.0;
      var
        i        : Cardinal;
        Position : Cardinal;
      begin
        Position := Length( aMemory);
        SetLength( aMemory, Position + aSampleCount);
        for i := 0 to aSampleCount - 1
        do begin
          aMemory[ Position] := aBuf[ i] / MaxVal;
          Inc( Position);
        end;
      end;

    var
      aMemory    : TSignalArray;
      aWaveFile  : TWaveFile;
      aByteCount : Cardinal;
      aBuf       : TBuffer;
      aDuration  : Cardinal;
      aFormat    : TPCMFormat;
    begin
      SetLength( aMemory, 0);
      aDuration := 0;
      if FileExists( FileName) // todo: and FileSize < 44100 * 2 * 2 * 301 // 44k1, stereo, 16 bit and duration < 301 s.
      then begin
        aWaveFile := TWaveFile.Create( FileName, fmOpenRead);
        try
          if aWaveFile.Valid
          then begin
            aFormat := aWaveFile.PCMFormat;
            if ( aFormat = Stereo16bit44100Hz) or ( aFormat = Mono16bit44100Hz)
            then begin
              aDuration := aWaveFile.Length;
              aWaveFile.Position := 0;
              if aWaveFile.BeginRead
              then begin
                try
                  while Cardinal( aWaveFile.Position) < aWaveFile.DataSize
                  do begin
                    aByteCount := aWaveFile.Read( aBuf, BufSize * SizeOf( TSample));
                    case aFormat of
                      Stereo16bit44100Hz : begin FIsStereo := True ; SetSamples16bit44100Hz( aMemory, aBuf, aByteCount div SizeOf( TSample)); end;
                      Mono16bit44100Hz   : begin FIsStereo := False; SetSamples16bit44100Hz( aMemory, aBuf, aByteCount div SizeOf( TSample)); end;
                    end;
                  end;
                finally
                  aWaveFile.EndRead;
                end;
              end;
            end;
          end;
        finally
          aWaveFile.Free;
        end;
      end;
      SwapMemoryFor( aMemory);
      SetDuration( aDuration);
    end;

    procedure   TModWave.SetFileName( const aValue: string);
    begin
      if  aValue <> FFileName
      then begin
        FFileName := aValue;
        LoadFile;
      end;
    end;

//  public

    procedure   TModWave.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModWave.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FFileName
      else Result := inherited;
    end;

    procedure   TModWave.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    const
      MonoText : array[ Boolean] of string = ( 'M', 'S');
    begin
      aCallback( Self, Format( '%s/%s/duration/%s', [ aPrefix, Name, FDuration]           , AppLocale));
      aCallback( Self, Format( '%s/%s/mono/%s'    , [ aPrefix, Name, MonoText[ FIsStereo]], AppLocale));
    end;

    function    TModWave.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        wave_outl :  Result := 'out';        // Left  output
        wave_outr :  Result := 'outr';       // Right output
        else         Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModWavePlayer = class( TModWave)
  private
    FPlayPointer : TSignal;
    FOldRes      : Boolean;
  public
}

const

  waveplayer_speed     = 0;
  waveplayer_position  = 1;
  waveplayer_res       = 2;
  waveplayer_mode      = 3;
  waveplayer_fm        = 4;
{ waveplayer_convolve  = 5; }


    procedure   TModWavePlayer.SetDefaults; // override;
    begin
      inherited;
      FInputs[ waveplayer_speed]    := 0.0; // Set zero speed
      FInputs[ waveplayer_position] := 0.0; // Set start position
      FInputs[ waveplayer_res]      := 0.0; // Reset to zero
      FInputs[ waveplayer_fm]       := 0.0; // FrequencyToUnits( ReferenceA); // Set FM amount to neutral to make speed work ...
      SetDuration( 0);
    end;

    procedure   TModWavePlayer.DoTick; // override;
    var
      SampCount : Integer;
      Increment : TSignal;
      Position  : TSignal;
      MakeMono  : Boolean;
    begin
      if ResetFlag
      then begin
        FPlayPointer := 0.0;
        ResetFlag    := False;
      end;

      if ( not FOldRes) and SignalToLogic( FInputs[ waveplayer_res])
      then FPlayPointer := 0.0;
      FOldRes := SignalToLogic( FInputs[ waveplayer_res]);

      if FIsStereo
      then begin
        MakeMono := SignalToInt( FInputs[ waveplayer_mode]) = 0;
        if Assigned( FMemoryL) and Assigned( FMemoryR)
        then begin
          SampCount := FMemoryL.Count;
          if SampCount > 0
          then begin
            Increment    := Clip( FInputs[ waveplayer_speed] * LookupFrequency( FInputs[ waveplayer_fm]) / LookupFrequency( 0), - 16, 16);
            Position     := ( 32 + Clip( FInputs[ waveplayer_position],  -16, 116)) * SampCount;
            FPlayPointer := FMod( FPlayPointer + Increment + SampCount, SampCount);
            Position     := FMod( Position + FPlayPointer, SampCount);
            if MakeMono
            then begin
              FOutputs[ wave_outl] := ( FMemoryL.Interpolate( Position) + FMemoryR.Interpolate( Position)) / 2;
              FOutputs[ wave_outr] := FOutputs[ wave_outl];
            end
            else begin
              FOutputs[ wave_outl] := FMemoryL.Interpolate( Position);
              FOutputs[ wave_outr] := FMemoryR.Interpolate( Position);
            end;
          end
          else begin
            FOutputs[ wave_outl] := 0;
            FOutputs[ wave_outr] := 0;
          end;
        end
        else begin
          FOutputs[ wave_outl] := 0;
          FOutputs[ wave_outr] := 0;
        end;
      end
      else begin
        if Assigned( FMemoryL)
        then begin
          SampCount := FMemoryL.Count;
          if SampCount > 0
          then begin
            Increment     := Clip( FInputs[ waveplayer_speed] * LookupFrequency( FInputs[ waveplayer_fm]) / LookupFrequency( 0), - 16, 16);
            Position      := ( 32 + Clip( FInputs[ waveplayer_position],  -16, 16)) * SampCount;
            FPlayPointer  := FMod( FPlayPointer + Increment + SampCount, SampCount);
            Position      := FMod( Position + FPlayPointer, SampCount);
            FOutputs[ wave_outl] := FMemoryL.Interpolate( Position);
          end
          else FOutputs[ wave_outl] := 0;
        end
        else FOutputs[ wave_outl] := 0;
        FOutputs[ wave_outr] := FOutputs[ wave_outl];
      end;
    end;

    procedure   TModWavePlayer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/res/%f', [ aPrefix, Name, FInputs[ waveplayer_res]], AppLocale));
    end;

    function    TModWavePlayer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        waveplayer_speed    : Result := 'speed';
        waveplayer_position : Result := 'position';
        waveplayer_res      : Result := 'res';
        waveplayer_mode     : Result := 'mode';            // 0 = mono, 1 = stereo
        waveplayer_fm       : Result := 'fm';
     // waveplayer_convolve : Result := 'convolve';
        else                  Result := NO_INPUT;
      end;
    end;


{ ========
  TModSampleAndHold = class( TMod)
  private
    FLastTrig: Boolean;
  public
}

const

  sampleandhold_trig = 0;
  sampleandhold_in   = 1;

  sampleandhold_out  = 0;


    constructor TModSampleAndHold.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModSampleAndHold.DoTick; // override;
    var
      Trig : Boolean;
    begin
      Trig := SignalToLogic( FInputs[ sampleandhold_trig]);
      if not FLastTrig and Trig
      then FOutputs[ sampleandhold_out] := FInputs[ sampleandhold_in];
      FLastTrig := Trig;
    end;

    procedure   TModSampleAndHold.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ sampleandhold_trig]], AppLocale));
    end;

    function    TModSampleAndHold.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        sampleandhold_trig : Result := 'trig';
        sampleandhold_in   : Result := 'in';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModSampleAndHold.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        sampleandhold_out : Result := 'out';
        else                Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMultiSAndH = class( TMod)
  private
    FLastTrig: Boolean;
  public
}

const

  multisandh_trig  =  0;
  multisandh_in1   =  1;
  multisandh_in2   =  2;
  multisandh_in3   =  3;
  multisandh_in4   =  4;
  multisandh_in5   =  5;
  multisandh_in6   =  6;
  multisandh_in7   =  7;
  multisandh_in8   =  8;
  multisandh_in9   =  9;
  multisandh_in10  = 10;
  multisandh_in11  = 11;
  multisandh_in12  = 12;

  multisandh_out1  =  0;
  multisandh_out2  =  1;
  multisandh_out3  =  2;
  multisandh_out4  =  3;
  multisandh_out5  =  4;
  multisandh_out6  =  5;
  multisandh_out7  =  6;
  multisandh_out8  =  7;
  multisandh_out9  =  8;
  multisandh_out10 =  9;
  multisandh_out11 = 10;
  multisandh_out12 = 11;


    constructor TModMultiSAndH.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMultiSAndH.DoTick; // override;
    var
      Trig : Boolean;
      i    : Integer;
    begin
      Trig := SignalToLogic( FInputs[ multisandh_trig]);
      if not FLastTrig and Trig
      then begin
        for i := multisandh_out1 to multisandh_out12
        do FOutputs[ i] := FInputs[ i + multisandh_in1 - multisandh_out1];
      end;
      FLastTrig := Trig;
    end;

    procedure   TModMultiSAndH.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ multisandh_trig]], AppLocale));
    end;

    function    TModMultiSAndH.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        multisandh_trig : Result := 'trig';
        multisandh_in1  : Result := 'in1';
        multisandh_in2  : Result := 'in2';
        multisandh_in3  : Result := 'in3';
        multisandh_in4  : Result := 'in4';
        multisandh_in5  : Result := 'in5';
        multisandh_in6  : Result := 'in6';
        multisandh_in7  : Result := 'in7';
        multisandh_in8  : Result := 'in8';
        multisandh_in9  : Result := 'in9';
        multisandh_in10 : Result := 'in10';
        multisandh_in11 : Result := 'in11';
        multisandh_in12 : Result := 'in12';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModMultiSAndH.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        multisandh_out1  : Result := 'out1';
        multisandh_out2  : Result := 'out2';
        multisandh_out3  : Result := 'out3';
        multisandh_out4  : Result := 'out4';
        multisandh_out5  : Result := 'out5';
        multisandh_out6  : Result := 'out6';
        multisandh_out7  : Result := 'out7';
        multisandh_out8  : Result := 'out8';
        multisandh_out9  : Result := 'out9';
        multisandh_out10 : Result := 'out10';
        multisandh_out11 : Result := 'out11';
        multisandh_out12 : Result := 'out12';
        else               Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMux = class( TMod)
  private
    FState : Integer;
  public
}

const

  mux_in0     =  0;
  mux_in1     =  1;
  mux_in2     =  2;
  mux_in3     =  3;
  mux_in4     =  4;
  mux_in5     =  5;
  mux_in6     =  6;
  mux_in7     =  7;
  mux_in8     =  8;
  mux_in9     =  9;
  mux_in10    = 10;
  mux_in11    = 11;
  mux_control = 12;
  mux_xfade   = 13;

  mux_out     =  0;

    constructor TModMux.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMux.DoTick; // override;
    var
      XFade : TSignal;
      State : TSignal;
      Delta : TSignal;
      Next  : Integer;
      Curr  : TSignal;
    begin
      XFade  := FInputs[ mux_xfade];
      State  := FMod( 12 * Clip( FInputs[ mux_control], 0, 1), 12);
      FState := Trunc( State);
      Next   := ( FState + 1) mod 12;
      Delta  := XFade * ( State - FState);
      Curr   := FInputs[ FState];
      FOutputs[ mux_out] := Curr + Delta * ( FInputs[ Next] - Curr);
    end;

    procedure   TModMux.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := 0 to 11
      do aCallback( Self, Format( '%s/%s/%d/%f', [ aPrefix, Name, i, LogicToSignal( FState = i)], AppLocale));
    end;

    function    TModMux.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mux_in0     : Result := 'in0';
        mux_in1     : Result := 'in1';
        mux_in2     : Result := 'in2';
        mux_in3     : Result := 'in3';
        mux_in4     : Result := 'in4';
        mux_in5     : Result := 'in5';
        mux_in6     : Result := 'in6';
        mux_in7     : Result := 'in7';
        mux_in8     : Result := 'in8';
        mux_in9     : Result := 'in9';
        mux_in10    : Result := 'in10';
        mux_in11    : Result := 'in11';
        mux_control : Result := 'control';
        mux_xfade   : Result := 'xfade';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModMux.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mux_out : Result := 'out';
        else      Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDemux = class( TMod)
  private
    FState : Integer;
  public
}

const

  demux_in     =  0;

  demux_out0    =  0;
  demux_out1    =  1;
  demux_out2    =  2;
  demux_out3    =  3;
  demux_out4    =  4;
  demux_out5    =  5;
  demux_out6    =  6;
  demux_out7    =  7;
  demux_out8    =  8;
  demux_out9    =  9;
  demux_out10   = 10;
  demux_out11   = 11;
  demux_control = 12;
  demux_xfade   = 13;

    constructor TModDemux.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModDemux.DoTick; // override;
    var
      XFade : TSignal;
      State : TSignal;
      Delta : TSignal;
      Next  : Integer;
      Input : TSignal;
    begin
      XFade  := FInputs[ demux_xfade];
      State  := FMod( 12 * Clip( FInputs[ demux_control], 0, 1), 12);
      FState := Trunc( State);
      Next   := ( FState + 1) mod 12;
      Delta  := XFade * ( State - FState);
      Input  := FInputs[ demux_in];

      FOutputs[ FState] :=       Delta  * Input;
      FOutputs[ Next  ] := ( 1 - Delta) * Input;
    end;

    procedure   TModDemux.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := 0 to 11
      do aCallback( Self, Format( '%s/%s/%d/%f', [ aPrefix, Name, i, LogicToSignal( FState = i)], AppLocale));
    end;

    function    TModDemux.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        demux_in      : Result := 'in';
        demux_control : Result := 'control';
        demux_xfade   : Result := 'xfade';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModDemux.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        demux_out0  : Result := 'out0';
        demux_out1  : Result := 'out1';
        demux_out2  : Result := 'out2';
        demux_out3  : Result := 'out3';
        demux_out4  : Result := 'out4';
        demux_out5  : Result := 'out5';
        demux_out6  : Result := 'out6';
        demux_out7  : Result := 'out7';
        demux_out8  : Result := 'out8';
        demux_out9  : Result := 'out9';
        demux_out10 : Result := 'out10';
        demux_out11 : Result := 'out11';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTrackAndHold = class( TMod)
  public
}

const

  trackandhold_trig = 0;
  trackandhold_in   = 1;

  trackandhold_out  = 0;


    constructor TModTrackAndHold.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModTrackAndHold.DoTick; // override;
    begin
      if SignalToLogic( FInputs[ trackandhold_trig])
      then FOutputs[ trackandhold_out] := FInputs[ trackandhold_in];
    end;

    procedure   TModTrackAndHold.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInputs[ trackandhold_trig]], AppLocale));
    end;

    function    TModTrackAndHold.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        trackandhold_trig : Result := 'trig';
        trackandhold_in   : Result := 'in';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModTrackAndHold.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        trackandhold_out : Result := 'out';
        else               Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSwitch2to1 = class( TMod)
  public
}

const

    switch2to1_control = 0;
    switch2to1_in1     = 1;
    switch2to1_in2     = 2;

    switch2to1_out     = 0;


    constructor TModSwitch2to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModSwitch2to1.DoTick; // override;
    begin
      if SignalToLogic( FInputs[ switch2to1_control])
      then FOutputs[ switch2to1_out] := FInputs[ switch2to1_in2]
      else FOutputs[ switch2to1_out] := FInputs[ switch2to1_in1];
    end;

    procedure   TModSwitch2to1.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/control/%f', [ aPrefix, Name, FInputs[ switch2to1_control]], AppLocale));
    end;

    function    TModSwitch2to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        switch2to1_control : Result := 'control';
        switch2to1_in1     : Result := 'in1';
        switch2to1_in2     : Result := 'in2';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModSwitch2to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        switch2to1_out : Result := 'out';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModXFade = class( TMod)
  public
}

const

  xfade_control  = 0;
  xfade_in1      = 1;
  xfade_in2      = 2;
  xfade_position = 3;
  xfade_posmod   = 4;

  xfade_out      = 0;


    constructor TModXFade.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModXFade.SetDefaults; // override;
    begin
      FInputs[ xfade_control ] := 0; // Equal amounts of in1 and in2
      FInputs[ xfade_position] := 0;
      FDezipperMap := [
        xfade_position,
        xfade_posmod
      ];
    end;

    procedure   TModXFade.DoTick; // override;
    var
      Control: TSignal;
    begin
      Control := FInputs[ xfade_position] + FInputs[ xfade_control] * FInputs[ xfade_posmod];
      FOutputs[ xfade_out] := (( 1 - Control) * FInputs[ xfade_in1] + ( 1 + Control) * FInputs[ xfade_in2]) / 2;
    end;

    function    TModXFade.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        xfade_control  : Result := 'control';
        xfade_in1      : Result := 'in1';
        xfade_in2      : Result := 'in2';
        xfade_position : Result := 'position';
        xfade_posmod   : Result := 'posmod';
        else             Result := NO_INPUT;
      end;
    end;

    function    TModXFade.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        xfade_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPan = class( TMod)
  public
}

const

  pan_control  = 0;
  pan_in       = 1;
  pan_position = 2;
  pan_posmod   = 3;

  pan_out1     = 0;
  pan_out2     = 1;


    constructor TModPan.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModPan.SetDefaults; // override;
    begin
      FInputs[ pan_control ] := 0; // Mid position
      FDezipperMap := [
        pan_position,
        pan_posmod
      ];
    end;

    procedure   TModPan.DoTick; // override;
    var
      Control : TSignal;
      Inp     : TSignal;
    begin
      Control    := FInputs[ pan_position] + FInputs[ pan_control] * FInputs[ pan_posmod];
      Inp        := FInputs[ pan_in];
      FOutputs[ pan_out1] := (( 1 - Control) * Inp) / 2;
      FOutputs[ pan_out2] := (( 1 + Control) * Inp) / 2;
    end;

    function    TModPan.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pan_control  : Result := 'control';
        pan_in       : Result := 'in';
        pan_position : Result := 'position';
        pan_posmod   : Result := 'posmod';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModPan.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pan_out1 : Result := 'out1';
        pan_out2 : Result := 'out2';
        else       Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixM4to1 = class( TMod)
  public
}

const

  mixm4to1_in      =  0;
  mixm4to1_in1     =  1;
  mixm4to1_in2     =  2;
  mixm4to1_in3     =  3;
  mixm4to1_in4     =  4;
  mixm4to1_mute1   =  5;
  mixm4to1_mute2   =  6;
  mixm4to1_mute3   =  7;
  mixm4to1_mute4   =  8;
  mixm4to1_level1  =  9;
  mixm4to1_level2  = 10;
  mixm4to1_level3  = 11;
  mixm4to1_level4  = 12;
  mixm4to1_levmod1 = 13;
  mixm4to1_levmod2 = 14;
  mixm4to1_levmod3 = 15;
  mixm4to1_levmod4 = 16;

  mixm4to1_out     =  0;


    constructor TModMixM4to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMixM4to1.SetDefaults; // override;
    begin
      FInputs[ mixm4to1_in ]     := 0; // Unused inputs should be zero
      FInputs[ mixm4to1_in1]     := 0;
      FInputs[ mixm4to1_in2]     := 0;
      FInputs[ mixm4to1_in3]     := 0;
      FInputs[ mixm4to1_in4]     := 0;
      FInputs[ mixm4to1_levmod1] := 1;
      FInputs[ mixm4to1_levmod2] := 1;
      FInputs[ mixm4to1_levmod3] := 1;
      FInputs[ mixm4to1_levmod4] := 1;
      FDezipperMap := [
        mixm4to1_mute1,
        mixm4to1_mute2,
        mixm4to1_mute3,
        mixm4to1_mute4,
        mixm4to1_level1,
        mixm4to1_level2,
        mixm4to1_level3,
        mixm4to1_level4
      ];
    end;

    procedure   TModMixM4to1.DoTick; // override;
    begin
      FOutputs[ mixm4to1_out] :=
        FInputs[ mixm4to1_in ] +
        FInputs[ mixm4to1_in1] * FInputs[ mixm4to1_level1] * FInputs[ mixm4to1_levmod1] * SignalToMute( FInputs[ mixm4to1_mute1]) +
        FInputs[ mixm4to1_in2] * FInputs[ mixm4to1_level2] * FInputs[ mixm4to1_levmod2] * SignalToMute( FInputs[ mixm4to1_mute2]) +
        FInputs[ mixm4to1_in3] * FInputs[ mixm4to1_level3] * FInputs[ mixm4to1_levmod3] * SignalToMute( FInputs[ mixm4to1_mute3]) +
        FInputs[ mixm4to1_in4] * FInputs[ mixm4to1_level4] * FInputs[ mixm4to1_levmod4] * SignalToMute( FInputs[ mixm4to1_mute4]) ;
    end;

    function    TModMixM4to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixm4to1_in      : Result := 'in';
        mixm4to1_in1     : Result := 'in1';
        mixm4to1_in2     : Result := 'in2';
        mixm4to1_in3     : Result := 'in3';
        mixm4to1_in4     : Result := 'in4';
        mixm4to1_mute1   : Result := 'mute1';
        mixm4to1_mute2   : Result := 'mute2';
        mixm4to1_mute3   : Result := 'mute3';
        mixm4to1_mute4   : Result := 'mute4';
        mixm4to1_level1  : Result := 'level1';
        mixm4to1_level2  : Result := 'level2';
        mixm4to1_level3  : Result := 'level3';
        mixm4to1_level4  : Result := 'level4';
        mixm4to1_levmod1 : Result := 'levmod1';
        mixm4to1_levmod2 : Result := 'levmod2';
        mixm4to1_levmod3 : Result := 'levmod3';
        mixm4to1_levmod4 : Result := 'levmod4';
        else               Result := NO_INPUT;
      end;
    end;

    function    TModMixM4to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixm4to1_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixS2to1 = class( TMod)
  public
}

const

  mixs2to1_inl     =  0;
  mixs2to1_inr     =  1;
  mixs2to1_inl1    =  2;
  mixs2to1_inr1    =  3;
  mixs2to1_inl2    =  4;
  mixs2to1_inr2    =  5;
  mixs2to1_mute1   =  6;
  mixs2to1_mute2   =  7;
  mixs2to1_level1  =  8;
  mixs2to1_level2  =  9;
  mixs2to1_levmod1 = 10;
  mixs2to1_levmod2 = 11;

  mixs2to1_outl    =  0;
  mixs2to1_outr    =  1;


    constructor TModMixS2to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMixS2to1.SetDefaults; // override;
    begin
      FInputs[ mixs2to1_inl    ] := 0; // Unused inputs should be zero
      FInputs[ mixs2to1_inr    ] := 0;
      FInputs[ mixs2to1_inl1   ] := 0;
      FInputs[ mixs2to1_inr1   ] := 0;
      FInputs[ mixs2to1_inl2   ] := 0;
      FInputs[ mixs2to1_inr2   ] := 0;
      FInputs[ mixs2to1_levmod1] := 1;
      FInputs[ mixs2to1_levmod2] := 1;
      FDezipperMap := [
        mixs2to1_mute1,
        mixs2to1_mute2,
        mixs2to1_level1,
        mixs2to1_level2
      ];
    end;

    procedure   TModMixS2to1.DoTick; // override;
    var
      lev1 : TSignal;
      lev2 : TSignal;
    begin
      lev1 := FInputs[ mixs2to1_levmod1] * FInputs[ mixs2to1_level1] * SignalToMute( FInputs[ mixs2to1_mute1]);
      lev2 := FInputs[ mixs2to1_levmod2] * FInputs[ mixs2to1_level2] * SignalToMute( FInputs[ mixs2to1_mute2]);

      FOutputs[ mixs2to1_outl] :=
        FInputs[ mixs2to1_inl]         +
        FInputs[ mixs2to1_inl1] * lev1 +
        FInputs[ mixs2to1_inl2] * lev2 ;
      FOutputs[ mixs2to1_outr] :=
        FInputs[ mixs2to1_inr]         +
        FInputs[ mixs2to1_inr1] * lev1 +
        FInputs[ mixs2to1_inr2] * lev2 ;
    end;

    function    TModMixS2to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixs2to1_inl     : Result := 'inl';
        mixs2to1_inr     : Result := 'inr';
        mixs2to1_inl1    : Result := 'inl1';
        mixs2to1_inr1    : Result := 'inr1';
        mixs2to1_inl2    : Result := 'inl2';
        mixs2to1_inr2    : Result := 'inr2';
        mixs2to1_mute1   : Result := 'mute1';
        mixs2to1_mute2   : Result := 'mute2';
        mixs2to1_level1  : Result := 'level1';
        mixs2to1_level2  : Result := 'level2';
        mixs2to1_levmod1 : Result := 'levmod1';
        mixs2to1_levmod2 : Result := 'levmod2';
        else               Result := NO_INPUT;
      end;
    end;

    function    TModMixS2to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixs2to1_outl : Result := 'outl';
        mixs2to1_outr : Result := 'outr';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixT1to1 = class( TMod)
  public
}

const

  mixt1to1_inl     = 0;
  mixt1to1_inm     = 1;
  mixt1to1_inr     = 2;
  mixt1to1_inl1    = 3;
  mixt1to1_inm1    = 4;
  mixt1to1_inr1    = 5;
  mixt1to1_mute1   = 6;
  mixt1to1_level1  = 7;
  mixt1to1_levmod1 = 8;

  mixt1to1_outl    = 0;
  mixt1to1_outm    = 1;
  mixt1to1_outr    = 2;

    constructor TModMixT1to1.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMixT1to1.SetDefaults; // override;
    begin
      FInputs[ mixt1to1_inl    ] := 0; // Unused inputs should be zero
      FInputs[ mixt1to1_inm    ] := 0;
      FInputs[ mixt1to1_inr    ] := 0;
      FInputs[ mixt1to1_inl1   ] := 0;
      FInputs[ mixt1to1_inm1   ] := 0;
      FInputs[ mixt1to1_inr1   ] := 0;
      FInputs[ mixt1to1_levmod1] := 1;
      FDezipperMap := [
        mixt1to1_mute1,
        mixt1to1_level1
      ];
    end;

    procedure   TModMixT1to1.DoTick; // override;
    var
      lev1 : TSignal;
    begin
      lev1 := FInputs[ mixt1to1_levmod1] * FInputs[ mixt1to1_level1] * SignalToMute( FInputs[ mixt1to1_mute1]);

      FOutputs[ mixt1to1_outl] := FInputs[ mixt1to1_inl] + FInputs[ mixt1to1_inl1] * lev1;
      FOutputs[ mixt1to1_outm] := FInputs[ mixt1to1_inm] + FInputs[ mixt1to1_inm1] * lev1;
      FOutputs[ mixt1to1_outr] := FInputs[ mixt1to1_inr] + FInputs[ mixt1to1_inr1] * lev1;
    end;

    function    TModMixT1to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixt1to1_inl     : Result := 'inl';
        mixt1to1_inm     : Result := 'inm';
        mixt1to1_inr     : Result := 'inr';
        mixt1to1_inl1    : Result := 'inl1';
        mixt1to1_inm1    : Result := 'inm1';
        mixt1to1_inr1    : Result := 'inr1';
        mixt1to1_mute1   : Result := 'mute1';
        mixt1to1_level1  : Result := 'level1';
        mixt1to1_levmod1 : Result := 'levmod1';
        else               Result := NO_INPUT;
      end;
    end;

    function    TModMixT1to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixt1to1_outl : Result := 'outl';
        mixt1to1_outm : Result := 'outm';
        mixt1to1_outr : Result := 'outr';
        else            Result := NO_OUTPUT;
      end;
    end;



{ ========
 TModMixM6to6 = class( TMod)
  public
}

const

  mixm6to6_in1       =  0;
  mixm6to6_in2       =  1;
  mixm6to6_in3       =  2;
  mixm6to6_in4       =  3;
  mixm6to6_in5       =  4;
  mixm6to6_in6       =  5;
  mixm6to6_mute1     =  6;
  mixm6to6_mute2     =  7;
  mixm6to6_mute3     =  8;
  mixm6to6_mute4     =  9;
  mixm6to6_mute5     = 10;
  mixm6to6_mute6     = 11;
  mixm6to6_inchcol1  = 12;
  mixm6to6_inchcol2  = 13;
  mixm6to6_inchcol3  = 14;
  mixm6to6_inchcol4  = 15;
  mixm6to6_inchcol5  = 16;
  mixm6to6_inchcol6  = 17;
  mixm6to6_levelr1c1 = 18;
  mixm6to6_levelr1c2 = 19;
  mixm6to6_levelr1c3 = 20;
  mixm6to6_levelr1c4 = 21;
  mixm6to6_levelr1c5 = 22;
  mixm6to6_levelr1c6 = 23;
  mixm6to6_levelr2c1 = 24;
  mixm6to6_levelr2c2 = 25;
  mixm6to6_levelr2c3 = 26;
  mixm6to6_levelr2c4 = 27;
  mixm6to6_levelr2c5 = 28;
  mixm6to6_levelr2c6 = 29;
  mixm6to6_levelr3c1 = 30;
  mixm6to6_levelr3c2 = 31;
  mixm6to6_levelr3c3 = 32;
  mixm6to6_levelr3c4 = 33;
  mixm6to6_levelr3c5 = 34;
  mixm6to6_levelr3c6 = 35;
  mixm6to6_levelr4c1 = 36;
  mixm6to6_levelr4c2 = 37;
  mixm6to6_levelr4c3 = 38;
  mixm6to6_levelr4c4 = 39;
  mixm6to6_levelr4c5 = 40;
  mixm6to6_levelr4c6 = 41;
  mixm6to6_levelr5c1 = 42;
  mixm6to6_levelr5c2 = 43;
  mixm6to6_levelr5c3 = 44;
  mixm6to6_levelr5c4 = 45;
  mixm6to6_levelr5c5 = 46;
  mixm6to6_levelr5c6 = 47;
  mixm6to6_levelr6c1 = 48;
  mixm6to6_levelr6c2 = 49;
  mixm6to6_levelr6c3 = 50;
  mixm6to6_levelr6c4 = 51;
  mixm6to6_levelr6c5 = 52;
  mixm6to6_levelr6c6 = 53;
  mixm6to6_inr1c1    = 54;
  mixm6to6_inr1c2    = 55;
  mixm6to6_inr1c3    = 56;
  mixm6to6_inr1c4    = 57;
  mixm6to6_inr1c5    = 58;
  mixm6to6_inr1c6    = 59;
  mixm6to6_inr2c1    = 60;
  mixm6to6_inr2c2    = 61;
  mixm6to6_inr2c3    = 62;
  mixm6to6_inr2c4    = 63;
  mixm6to6_inr2c5    = 64;
  mixm6to6_inr2c6    = 65;
  mixm6to6_inr3c1    = 66;
  mixm6to6_inr3c2    = 67;
  mixm6to6_inr3c3    = 68;
  mixm6to6_inr3c4    = 69;
  mixm6to6_inr3c5    = 70;
  mixm6to6_inr3c6    = 71;
  mixm6to6_inr4c1    = 72;
  mixm6to6_inr4c2    = 73;
  mixm6to6_inr4c3    = 74;
  mixm6to6_inr4c4    = 75;
  mixm6to6_inr4c5    = 76;
  mixm6to6_inr4c6    = 77;
  mixm6to6_inr5c1    = 78;
  mixm6to6_inr5c2    = 79;
  mixm6to6_inr5c3    = 80;
  mixm6to6_inr5c4    = 81;
  mixm6to6_inr5c5    = 82;
  mixm6to6_inr5c6    = 83;
  mixm6to6_inr6c1    = 84;
  mixm6to6_inr6c2    = 85;
  mixm6to6_inr6c3    = 86;
  mixm6to6_inr6c4    = 87;
  mixm6to6_inr6c5    = 88;
  mixm6to6_inr6c6    = 89;

  mixm6to6_out1      =  0;
  mixm6to6_out2      =  1;
  mixm6to6_out3      =  2;
  mixm6to6_out4      =  3;
  mixm6to6_out5      =  4;
  mixm6to6_out6      =  5;
  mixm6to6_outchrow1 =  6;
  mixm6to6_outchrow2 =  7;
  mixm6to6_outchrow3 =  8;
  mixm6to6_outchrow4 =  9;
  mixm6to6_outchrow5 = 10;
  mixm6to6_outchrow6 = 11;


    constructor TModMixM6to6.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMixM6to6.SetDefaults; // override;
    begin
      FInputs[ mixm6to6_inchcol1] := 0.0; // Zero chain col inputs
      FInputs[ mixm6to6_inchcol2] := 0.0;
      FInputs[ mixm6to6_inchcol3] := 0.0;
      FInputs[ mixm6to6_inchcol4] := 0.0;
      FInputs[ mixm6to6_inchcol5] := 0.0;
      FInputs[ mixm6to6_inchcol6] := 0.0;
      FDezipperMap := [
        mixm6to6_mute1     .. mixm6to6_mute6,
        mixm6to6_levelr1c1 .. mixm6to6_levelr6c6
      ];
    end;

    procedure   TModMixM6to6.DoTick; // override;
    var
      In1 : TSignal;
      In2 : TSignal;
      In3 : TSignal;
      In4 : TSignal;
      In5 : TSignal;
      In6 : TSignal;
    begin
      In1 := FInputs[ mixm6to6_in1] * SignalToMute( FInputs[ mixm6to6_mute1]);
      In2 := FInputs[ mixm6to6_in2] * SignalToMute( FInputs[ mixm6to6_mute2]);
      In3 := FInputs[ mixm6to6_in3] * SignalToMute( FInputs[ mixm6to6_mute3]);
      In4 := FInputs[ mixm6to6_in4] * SignalToMute( FInputs[ mixm6to6_mute4]);
      In5 := FInputs[ mixm6to6_in5] * SignalToMute( FInputs[ mixm6to6_mute5]);
      In6 := FInputs[ mixm6to6_in6] * SignalToMute( FInputs[ mixm6to6_mute6]);

      FOutputs[ mixm6to6_out1] := In1 * ( FInputs[ mixm6to6_levelr1c1] + FInputs[ mixm6to6_inr1c1]) + In2 * ( FInputs[ mixm6to6_levelr2c1] + FInputs[ mixm6to6_inr2c1]) + In3 * ( FInputs[ mixm6to6_levelr3c1] + FInputs[ mixm6to6_inr3c1]) + In4 * ( FInputs[ mixm6to6_levelr4c1] + FInputs[ mixm6to6_inr4c1]) + In5 * ( FInputs[ mixm6to6_levelr5c1] + FInputs[ mixm6to6_inr5c1]) + In6 * ( FInputs[ mixm6to6_levelr6c1] + FInputs[ mixm6to6_inr6c1]) + FInputs[ mixm6to6_inchcol1];
      FOutputs[ mixm6to6_out2] := In1 * ( FInputs[ mixm6to6_levelr1c2] + FInputs[ mixm6to6_inr1c2]) + In2 * ( FInputs[ mixm6to6_levelr2c2] + FInputs[ mixm6to6_inr2c2]) + In3 * ( FInputs[ mixm6to6_levelr3c2] + FInputs[ mixm6to6_inr3c2]) + In4 * ( FInputs[ mixm6to6_levelr4c2] + FInputs[ mixm6to6_inr4c2]) + In5 * ( FInputs[ mixm6to6_levelr5c2] + FInputs[ mixm6to6_inr5c2]) + In6 * ( FInputs[ mixm6to6_levelr6c2] + FInputs[ mixm6to6_inr6c2]) + FInputs[ mixm6to6_inchcol2];
      FOutputs[ mixm6to6_out3] := In1 * ( FInputs[ mixm6to6_levelr1c3] + FInputs[ mixm6to6_inr1c3]) + In2 * ( FInputs[ mixm6to6_levelr2c3] + FInputs[ mixm6to6_inr2c3]) + In3 * ( FInputs[ mixm6to6_levelr3c3] + FInputs[ mixm6to6_inr3c3]) + In4 * ( FInputs[ mixm6to6_levelr4c3] + FInputs[ mixm6to6_inr4c3]) + In5 * ( FInputs[ mixm6to6_levelr5c3] + FInputs[ mixm6to6_inr5c3]) + In6 * ( FInputs[ mixm6to6_levelr6c3] + FInputs[ mixm6to6_inr6c3]) + FInputs[ mixm6to6_inchcol3];
      FOutputs[ mixm6to6_out4] := In1 * ( FInputs[ mixm6to6_levelr1c4] + FInputs[ mixm6to6_inr1c4]) + In2 * ( FInputs[ mixm6to6_levelr2c4] + FInputs[ mixm6to6_inr2c4]) + In3 * ( FInputs[ mixm6to6_levelr3c4] + FInputs[ mixm6to6_inr3c4]) + In4 * ( FInputs[ mixm6to6_levelr4c4] + FInputs[ mixm6to6_inr4c4]) + In5 * ( FInputs[ mixm6to6_levelr5c4] + FInputs[ mixm6to6_inr5c4]) + In6 * ( FInputs[ mixm6to6_levelr6c4] + FInputs[ mixm6to6_inr6c4]) + FInputs[ mixm6to6_inchcol4];
      FOutputs[ mixm6to6_out5] := In1 * ( FInputs[ mixm6to6_levelr1c5] + FInputs[ mixm6to6_inr1c5]) + In2 * ( FInputs[ mixm6to6_levelr2c5] + FInputs[ mixm6to6_inr2c5]) + In3 * ( FInputs[ mixm6to6_levelr3c5] + FInputs[ mixm6to6_inr3c5]) + In4 * ( FInputs[ mixm6to6_levelr4c5] + FInputs[ mixm6to6_inr4c5]) + In5 * ( FInputs[ mixm6to6_levelr5c5] + FInputs[ mixm6to6_inr5c5]) + In6 * ( FInputs[ mixm6to6_levelr6c5] + FInputs[ mixm6to6_inr6c5]) + FInputs[ mixm6to6_inchcol5];
      FOutputs[ mixm6to6_out6] := In1 * ( FInputs[ mixm6to6_levelr1c6] + FInputs[ mixm6to6_inr1c6]) + In2 * ( FInputs[ mixm6to6_levelr2c6] + FInputs[ mixm6to6_inr2c6]) + In3 * ( FInputs[ mixm6to6_levelr3c6] + FInputs[ mixm6to6_inr3c6]) + In4 * ( FInputs[ mixm6to6_levelr4c6] + FInputs[ mixm6to6_inr4c6]) + In5 * ( FInputs[ mixm6to6_levelr5c6] + FInputs[ mixm6to6_inr5c6]) + In6 * ( FInputs[ mixm6to6_levelr6c6] + FInputs[ mixm6to6_inr6c6]) + FInputs[ mixm6to6_inchcol6];

      FOutputs[ mixm6to6_outchrow1] := In1;
      FOutputs[ mixm6to6_outchrow2] := In2;
      FOutputs[ mixm6to6_outchrow3] := In3;
      FOutputs[ mixm6to6_outchrow4] := In4;
      FOutputs[ mixm6to6_outchrow5] := In5;
      FOutputs[ mixm6to6_outchrow6] := In6;
    end;

    function    TModMixM6to6.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixm6to6_in1       : Result := 'in1';
        mixm6to6_in2       : Result := 'in2';
        mixm6to6_in3       : Result := 'in3';
        mixm6to6_in4       : Result := 'in4';
        mixm6to6_in5       : Result := 'in5';
        mixm6to6_in6       : Result := 'in6';
        mixm6to6_mute1     : Result := 'mute1';
        mixm6to6_mute2     : Result := 'mute2';
        mixm6to6_mute3     : Result := 'mute3';
        mixm6to6_mute4     : Result := 'mute4';
        mixm6to6_mute5     : Result := 'mute5';
        mixm6to6_mute6     : Result := 'mute6';
        mixm6to6_inchcol1  : Result := 'inchcol1';
        mixm6to6_inchcol2  : Result := 'inchcol2';
        mixm6to6_inchcol3  : Result := 'inchcol3';
        mixm6to6_inchcol4  : Result := 'inchcol4';
        mixm6to6_inchcol5  : Result := 'inchcol5';
        mixm6to6_inchcol6  : Result := 'inchcol6';
        mixm6to6_levelr1c1 : Result := 'levelr1c1';
        mixm6to6_levelr1c2 : Result := 'levelr1c2';
        mixm6to6_levelr1c3 : Result := 'levelr1c3';
        mixm6to6_levelr1c4 : Result := 'levelr1c4';
        mixm6to6_levelr1c5 : Result := 'levelr1c5';
        mixm6to6_levelr1c6 : Result := 'levelr1c6';
        mixm6to6_levelr2c1 : Result := 'levelr2c1';
        mixm6to6_levelr2c2 : Result := 'levelr2c2';
        mixm6to6_levelr2c3 : Result := 'levelr2c3';
        mixm6to6_levelr2c4 : Result := 'levelr2c4';
        mixm6to6_levelr2c5 : Result := 'levelr2c5';
        mixm6to6_levelr2c6 : Result := 'levelr2c6';
        mixm6to6_levelr3c1 : Result := 'levelr3c1';
        mixm6to6_levelr3c2 : Result := 'levelr3c2';
        mixm6to6_levelr3c3 : Result := 'levelr3c3';
        mixm6to6_levelr3c4 : Result := 'levelr3c4';
        mixm6to6_levelr3c5 : Result := 'levelr3c5';
        mixm6to6_levelr3c6 : Result := 'levelr3c6';
        mixm6to6_levelr4c1 : Result := 'levelr4c1';
        mixm6to6_levelr4c2 : Result := 'levelr4c2';
        mixm6to6_levelr4c3 : Result := 'levelr4c3';
        mixm6to6_levelr4c4 : Result := 'levelr4c4';
        mixm6to6_levelr4c5 : Result := 'levelr4c5';
        mixm6to6_levelr4c6 : Result := 'levelr4c6';
        mixm6to6_levelr5c1 : Result := 'levelr5c1';
        mixm6to6_levelr5c2 : Result := 'levelr5c2';
        mixm6to6_levelr5c3 : Result := 'levelr5c3';
        mixm6to6_levelr5c4 : Result := 'levelr5c4';
        mixm6to6_levelr5c5 : Result := 'levelr5c5';
        mixm6to6_levelr5c6 : Result := 'levelr5c6';
        mixm6to6_levelr6c1 : Result := 'levelr6c1';
        mixm6to6_levelr6c2 : Result := 'levelr6c2';
        mixm6to6_levelr6c3 : Result := 'levelr6c3';
        mixm6to6_levelr6c4 : Result := 'levelr6c4';
        mixm6to6_levelr6c5 : Result := 'levelr6c5';
        mixm6to6_levelr6c6 : Result := 'levelr6c6';
        mixm6to6_inr1c1    : Result := 'inr1c1';
        mixm6to6_inr1c2    : Result := 'inr1c2';
        mixm6to6_inr1c3    : Result := 'inr1c3';
        mixm6to6_inr1c4    : Result := 'inr1c4';
        mixm6to6_inr1c5    : Result := 'inr1c5';
        mixm6to6_inr1c6    : Result := 'inr1c6';
        mixm6to6_inr2c1    : Result := 'inr2c1';
        mixm6to6_inr2c2    : Result := 'inr2c2';
        mixm6to6_inr2c3    : Result := 'inr2c3';
        mixm6to6_inr2c4    : Result := 'inr2c4';
        mixm6to6_inr2c5    : Result := 'inr2c5';
        mixm6to6_inr2c6    : Result := 'inr2c6';
        mixm6to6_inr3c1    : Result := 'inr3c1';
        mixm6to6_inr3c2    : Result := 'inr3c2';
        mixm6to6_inr3c3    : Result := 'inr3c3';
        mixm6to6_inr3c4    : Result := 'inr3c4';
        mixm6to6_inr3c5    : Result := 'inr3c5';
        mixm6to6_inr3c6    : Result := 'inr3c6';
        mixm6to6_inr4c1    : Result := 'inr4c1';
        mixm6to6_inr4c2    : Result := 'inr4c2';
        mixm6to6_inr4c3    : Result := 'inr4c3';
        mixm6to6_inr4c4    : Result := 'inr4c4';
        mixm6to6_inr4c5    : Result := 'inr4c5';
        mixm6to6_inr4c6    : Result := 'inr4c6';
        mixm6to6_inr5c1    : Result := 'inr5c1';
        mixm6to6_inr5c2    : Result := 'inr5c2';
        mixm6to6_inr5c3    : Result := 'inr5c3';
        mixm6to6_inr5c4    : Result := 'inr5c4';
        mixm6to6_inr5c5    : Result := 'inr5c5';
        mixm6to6_inr5c6    : Result := 'inr5c6';
        mixm6to6_inr6c1    : Result := 'inr6c1';
        mixm6to6_inr6c2    : Result := 'inr6c2';
        mixm6to6_inr6c3    : Result := 'inr6c3';
        mixm6to6_inr6c4    : Result := 'inr6c4';
        mixm6to6_inr6c5    : Result := 'inr6c5';
        mixm6to6_inr6c6    : Result := 'inr6c6';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModMixM6to6.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixm6to6_out1      : Result := 'out1';
        mixm6to6_out2      : Result := 'out2';
        mixm6to6_out3      : Result := 'out3';
        mixm6to6_out4      : Result := 'out4';
        mixm6to6_out5      : Result := 'out5';
        mixm6to6_out6      : Result := 'out6';
        mixm6to6_outchrow1 : Result := 'outchrow1';
        mixm6to6_outchrow2 : Result := 'outchrow2';
        mixm6to6_outchrow3 : Result := 'outchrow3';
        mixm6to6_outchrow4 : Result := 'outchrow4';
        mixm6to6_outchrow5 : Result := 'outchrow5';
        mixm6to6_outchrow6 : Result := 'outchrow6';
        else                 Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMixAudio = class( TMod)
  public
}

const

  mixaudio_insolo    =  0;
  mixaudio_inchain1l =  1;
  mixaudio_inchain1r =  2;
  mixaudio_inchain2l =  3;
  mixaudio_inchain2r =  4;
  mixaudio_in1       =  5;
  mixaudio_in2       =  6;
  mixaudio_in3       =  7;
  mixaudio_in4       =  8;
  mixaudio_in5       =  9;
  mixaudio_in6       = 10;
  mixaudio_in7       = 11;
  mixaudio_in8       = 12;
  mixaudio_pan1      = 13;
  mixaudio_pan2      = 14;
  mixaudio_pan3      = 15;
  mixaudio_pan4      = 16;
  mixaudio_pan5      = 17;
  mixaudio_pan6      = 18;
  mixaudio_pan7      = 19;
  mixaudio_pan8      = 20;
  mixaudio_vol11     = 21;
  mixaudio_vol12     = 22;
  mixaudio_vol13     = 23;
  mixaudio_vol14     = 24;
  mixaudio_vol15     = 25;
  mixaudio_vol16     = 26;
  mixaudio_vol17     = 27;
  mixaudio_vol18     = 28;
  mixaudio_vol21     = 29;
  mixaudio_vol22     = 30;
  mixaudio_vol23     = 31;
  mixaudio_vol24     = 32;
  mixaudio_vol25     = 33;
  mixaudio_vol26     = 34;
  mixaudio_vol27     = 35;
  mixaudio_vol28     = 36;
  mixaudio_mute1     = 37;
  mixaudio_mute2     = 38;
  mixaudio_mute3     = 39;
  mixaudio_mute4     = 40;
  mixaudio_mute5     = 41;
  mixaudio_mute6     = 42;
  mixaudio_mute7     = 43;
  mixaudio_mute8     = 44;
  mixaudio_solo1     = 45;
  mixaudio_solo2     = 46;
  mixaudio_solo3     = 47;
  mixaudio_solo4     = 48;
  mixaudio_solo5     = 49;
  mixaudio_solo6     = 50;
  mixaudio_solo7     = 51;
  mixaudio_solo8     = 52;

  mixaudio_outsolo   =  0;
  mixaudio_out1l     =  1;
  mixaudio_out1r     =  2;
  mixaudio_out2l     =  3;
  mixaudio_out2r     =  4;

    constructor TModMixAudio.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModMixAudio.SetDefaults; // override;
    begin
      FDezipperMap := [
        mixaudio_pan1 .. mixaudio_mute8
      ];
    end;

    procedure   TModMixAudio.DoTick; // override;
    var
      Inp      : TSignal;
      Control  : TSignal;
      Signal1L : TSignal;
      Signal1R : TSignal;
      Signal2L : TSignal;
      Signal2R : TSignal;
    begin
      if SignalToLogic( FInputs[ mixaudio_insolo])
      then begin
        Signal1L := FInputs[ mixaudio_inchain1l];
        Signal1R := FInputs[ mixaudio_inchain1r];
        Signal2L := FInputs[ mixaudio_inchain2l];
        Signal2R := FInputs[ mixaudio_inchain2r];
        FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
      end
      else begin
        if SignalToLogic( FInputs[ mixaudio_solo1])
        then begin
          Inp      := FInputs[ mixaudio_in1] * SignalToMute( FInputs[ mixaudio_mute1]);
          Control  := FInputs[ mixaudio_pan1];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol11];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol11];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol21];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol21];
          FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ mixaudio_solo2])
        then begin
          Inp      := FInputs[ mixaudio_in2] * SignalToMute( FInputs[ mixaudio_mute2]);
          Control  := FInputs[ mixaudio_pan2];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol12];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol12];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol22];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol22];
          FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ mixaudio_solo3])
        then begin
          Inp      := FInputs[ mixaudio_in3] * SignalToMute( FInputs[ mixaudio_mute3]);
          Control  := FInputs[ mixaudio_pan3];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol13];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol13];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol23];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol23];
          FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ mixaudio_solo4])
        then begin
          Inp      := FInputs[ mixaudio_in4] * SignalToMute( FInputs[ mixaudio_mute4]);
          Control  := FInputs[ mixaudio_pan4];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol14];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol14];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol24];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol24];
          FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ mixaudio_solo5])
        then begin
          Inp      := FInputs[ mixaudio_in5] * SignalToMute( FInputs[ mixaudio_mute5]);
          Control  := FInputs[ mixaudio_pan5];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol15];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol15];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol25];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol25];
          FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ mixaudio_solo6])
        then begin
          Inp      := FInputs[ mixaudio_in6] * SignalToMute( FInputs[ mixaudio_mute6]);
          Control  := FInputs[ mixaudio_pan6];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol16];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol16];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol26];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol26];
          FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ mixaudio_solo7])
        then begin
          Inp      := FInputs[ mixaudio_in7] * SignalToMute( FInputs[ mixaudio_mute7]);
          Control  := FInputs[ mixaudio_pan7];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol17];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol17];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol27];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol27];
          FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
        end
        else if SignalToLogic( FInputs[ mixaudio_solo8])
        then begin
          Inp      := FInputs[ mixaudio_in8] * SignalToMute( FInputs[ mixaudio_mute8]);
          Control  := FInputs[ mixaudio_pan8];
          Signal1L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol18];
          Signal1R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol18];
          Signal2L := ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol28];
          Signal2R := ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol28];
          FOutputs[ mixaudio_outsolo] := LogicToSignal( True);
        end
        else begin
          Signal1L := FInputs[ mixaudio_inchain1l];
          Signal1R := FInputs[ mixaudio_inchain1r];
          Signal2L := FInputs[ mixaudio_inchain2l];
          Signal2R := FInputs[ mixaudio_inchain2r];

          Inp      := FInputs[ mixaudio_in1] * SignalToMute( FInputs[ mixaudio_mute1]);
          Control  := FInputs[ mixaudio_pan1];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol11];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol11];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol21];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol21];

          Inp      := FInputs[ mixaudio_in2] * SignalToMute( FInputs[ mixaudio_mute2]);
          Control  := FInputs[ mixaudio_pan2];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol12];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol12];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol22];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol22];

          Inp      := FInputs[ mixaudio_in3] * SignalToMute( FInputs[ mixaudio_mute3]);
          Control  := FInputs[ mixaudio_pan3];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol13];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol13];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol23];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol23];

          Inp      := FInputs[ mixaudio_in4] * SignalToMute( FInputs[ mixaudio_mute4]);
          Control  := FInputs[ mixaudio_pan4];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol14];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol14];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol24];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol24];

          Inp      := FInputs[ mixaudio_in5] * SignalToMute( FInputs[ mixaudio_mute5]);
          Control  := FInputs[ mixaudio_pan5];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol15];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol15];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol25];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol25];

          Inp      := FInputs[ mixaudio_in6] * SignalToMute( FInputs[ mixaudio_mute6]);
          Control  := FInputs[ mixaudio_pan6];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol16];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol16];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol26];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol26];

          Inp      := FInputs[ mixaudio_in7] * SignalToMute( FInputs[ mixaudio_mute7]);
          Control  := FInputs[ mixaudio_pan7];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol17];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol17];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol27];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol27];

          Inp      := FInputs[ mixaudio_in8] * SignalToMute( FInputs[ mixaudio_mute8]);
          Control  := FInputs[ mixaudio_pan8];
          Signal1L := Signal1L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol18];
          Signal1R := Signal1R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol18];
          Signal2L := Signal2L + ((( 1 - Control) * Inp) * 0.5) * FInputs[ mixaudio_vol28];
          Signal2R := Signal2R + ((( 1 + Control) * Inp) * 0.5) * FInputs[ mixaudio_vol28];

          FOutputs[ mixaudio_outsolo] := LogicToSignal( False);
        end;
      end;
      FOutputs[ mixaudio_out1l] := Signal1L;
      FOutputs[ mixaudio_out1r] := Signal1R;
      FOutputs[ mixaudio_out2l] := Signal2L;
      FOutputs[ mixaudio_out2r] := Signal2R;
    end;

    procedure   TModMixAudio.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/solo/%f', [ aPrefix, Name, FOutputs[ mixaudio_outsolo]], AppLocale));
    end;

    function    TModMixAudio.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixaudio_insolo    : Result := 'insolo';
        mixaudio_inchain1l : Result := 'inchain1l';
        mixaudio_inchain1r : Result := 'inchain1r';
        mixaudio_inchain2l : Result := 'inchain2l';
        mixaudio_inchain2r : Result := 'inchain2r';
        mixaudio_in1       : Result := 'in1';
        mixaudio_in2       : Result := 'in2';
        mixaudio_in3       : Result := 'in3';
        mixaudio_in4       : Result := 'in4';
        mixaudio_in5       : Result := 'in5';
        mixaudio_in6       : Result := 'in6';
        mixaudio_in7       : Result := 'in7';
        mixaudio_in8       : Result := 'in8';
        mixaudio_pan1      : Result := 'pan1';
        mixaudio_pan2      : Result := 'pan2';
        mixaudio_pan3      : Result := 'pan3';
        mixaudio_pan4      : Result := 'pan4';
        mixaudio_pan5      : Result := 'pan5';
        mixaudio_pan6      : Result := 'pan6';
        mixaudio_pan7      : Result := 'pan7';
        mixaudio_pan8      : Result := 'pan8';
        mixaudio_vol11     : Result := 'vol11';
        mixaudio_vol12     : Result := 'vol12';
        mixaudio_vol13     : Result := 'vol13';
        mixaudio_vol14     : Result := 'vol14';
        mixaudio_vol15     : Result := 'vol15';
        mixaudio_vol16     : Result := 'vol16';
        mixaudio_vol17     : Result := 'vol17';
        mixaudio_vol18     : Result := 'vol18';
        mixaudio_vol21     : Result := 'vol21';
        mixaudio_vol22     : Result := 'vol22';
        mixaudio_vol23     : Result := 'vol23';
        mixaudio_vol24     : Result := 'vol24';
        mixaudio_vol25     : Result := 'vol25';
        mixaudio_vol26     : Result := 'vol26';
        mixaudio_vol27     : Result := 'vol27';
        mixaudio_vol28     : Result := 'vol28';
        mixaudio_mute1     : Result := 'mute1';
        mixaudio_mute2     : Result := 'mute2';
        mixaudio_mute3     : Result := 'mute3';
        mixaudio_mute4     : Result := 'mute4';
        mixaudio_mute5     : Result := 'mute5';
        mixaudio_mute6     : Result := 'mute6';
        mixaudio_mute7     : Result := 'mute7';
        mixaudio_mute8     : Result := 'mute8';
        mixaudio_solo1     : Result := 'solo1';
        mixaudio_solo2     : Result := 'solo2';
        mixaudio_solo3     : Result := 'solo3';
        mixaudio_solo4     : Result := 'solo4';
        mixaudio_solo5     : Result := 'solo5';
        mixaudio_solo6     : Result := 'solo6';
        mixaudio_solo7     : Result := 'solo7';
        mixaudio_solo8     : Result := 'solo8';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModMixAudio.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        mixaudio_outsolo : Result := 'outsolo';
        mixaudio_out1l   : Result := 'out1l';
        mixaudio_out1r   : Result := 'out1r';
        mixaudio_out2l   : Result := 'out2l';
        mixaudio_out2r   : Result := 'out2r';
        else               Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModQuantize = class( TMod)
  public
}

const

  quantize_qsteps = 0;
  quantize_in     = 1;
  quantize_mute   = 2;
  quantize_level  = 3;

  quantize_out    = 0;


    constructor TModQuantize.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModQuantize.DoTick; // override;
    var
      Steps: Integer;
    begin
      Steps := Max( 1, Round( FInputs[ quantize_qsteps]));
      FOutputs[ quantize_out] := Round( Steps * FInputs[ quantize_level] * FInputs[ quantize_in] * SignalToMute( FInputs[ quantize_mute] )) / Steps;
    end;

    function    TModQuantize.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        quantize_qsteps : Result := 'qsteps';
        quantize_in     : Result := 'in';
        quantize_mute   : Result := 'mute';
        quantize_level  : Result := 'level';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModQuantize.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        quantize_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModHardClip = class( TMod)
  public
}

const

  hardclip_in        = 0;
  hardclip_mute      = 1;
  hardclip_lowlevel  = 2;
  hardclip_highlevel = 3;
  hardclip_outamp    = 4;

  hardclip_out       = 0;

    constructor TModHardClip.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModHardClip.SetDefaults; // override;
    begin
      FDezipperMap := [
        hardclip_mute,
        hardclip_lowlevel,
        hardclip_highlevel,
        hardclip_outamp
      ];
    end;

    procedure   TModHardClip.DoTick; // override;
    begin
      FOutputs[ hardclip_out] := FInputs[ hardclip_outamp] * SignalToMute( FInputs[ hardclip_mute]) * Clip( FInputs[ hardclip_in], FInputs[ hardclip_lowlevel], FInputs[ hardclip_highlevel]);
    end;

    function    TModHardClip.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        hardclip_in        : Result := 'in';
        hardclip_mute      : Result := 'mute';
        hardclip_lowlevel  : Result := 'lowlevel';
        hardclip_highlevel : Result := 'highlevel';
        hardclip_outamp    : Result := 'outamp';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModHardClip.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        hardclip_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTransInv = class( TMod)
  public
}

const

  transinv_in1            = 0;
  transinv_in2            = 1;
  transinv_in3            = 2;
  transinv_in4            = 3;
  transinv_amtpivot       = 4;
  transinv_modulo         = 5;
  transinv_mode           = 6;

  transinv_out1           = 0;
  transinv_out2           = 1;
  transinv_out3           = 2;
  transinv_out4           = 3;

  transinv_mode_transpose = 0;
  transinv_mode_invert    = 1;


    constructor TModTransInv.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModTransInv.DoTick; // override;
    // transpose :: T( amt, x) -> ( amt + x) mod modulo // amt and x being integer
    // invert    := I( piv, x) -> ( piv - x) mod modulo // piv and x being integer
    var
      Mode      : Integer;
      Modulo    : Integer;
      AmtPivot  : TSignal;
      InValue1  : TSignal;
      InValue2  : TSignal;
      InValue3  : TSignal;
      InValue4  : TSignal;
    begin
      Mode     := SignalToInt( FInputs[ transinv_mode  ]);
      Modulo   := Clip( SignalToInt( FInputs[ transinv_modulo]), 1, 128);
      AmtPivot := UnitsToNoteNumber( FInputs[ transinv_amtpivot]) - MiddleNote;
      InValue1 := UnitsToNoteNumber( FInputs[ transinv_in1]);
      InValue2 := UnitsToNoteNumber( FInputs[ transinv_in2]);
      InValue3 := UnitsToNoteNumber( FInputs[ transinv_in3]);
      InValue4 := UnitsToNoteNumber( FInputs[ transinv_in4]);

      case Mode of
        transinv_mode_transpose : begin
          FOutputs[ transinv_out1] := NoteNumberToUnits( FMod( AmtPivot + InValue1, Modulo));
          FOutputs[ transinv_out2] := NoteNumberToUnits( FMod( AmtPivot + InValue2, Modulo));
          FOutputs[ transinv_out3] := NoteNumberToUnits( FMod( AmtPivot + InValue3, Modulo));
          FOutputs[ transinv_out4] := NoteNumberToUnits( FMod( AmtPivot + InValue4, Modulo));
        end;
        else begin
          FOutputs[ transinv_out1] := NoteNumberToUnits( FMod( AmtPivot - InValue1, Modulo));
          FOutputs[ transinv_out2] := NoteNumberToUnits( FMod( AmtPivot - InValue2, Modulo));
          FOutputs[ transinv_out3] := NoteNumberToUnits( FMod( AmtPivot - InValue3, Modulo));
          FOutputs[ transinv_out4] := NoteNumberToUnits( FMod( AmtPivot - InValue4, Modulo));
        end;
      end;
    end;

    function    TModTransInv.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        transinv_in1      : Result := 'in1';
        transinv_in2      : Result := 'in2';
        transinv_in3      : Result := 'in3';
        transinv_in4      : Result := 'in4';
        transinv_amtpivot : Result := 'amtpivot';
        transinv_modulo   : Result := 'modulo';
        transinv_mode     : Result := 'mode';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModTransInv.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        transinv_out1 : Result := 'out1';
        transinv_out2 : Result := 'out2';
        transinv_out3 : Result := 'out3';
        transinv_out4 : Result := 'out4';
        else            Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModScala = class( TMod)
  private
    FMappingFileName : string;
    FScaleFileName   : string;
    FScale           : TScalaScale;
    FLastNote        : Integer;
    FlastOut         : TSignal;
  public
    property    MappingFileName : string read FMappingFileName write SetMappingFileName;
    property    ScaleFileName   : string read FScaleFileName   write SetScaleFileName;
  private
}

const

  scala_in  = 0;

  scala_out = 0;

    procedure   TModScala.SetMappingFileName( const aValue: string);
    begin
      if aValue <> FMappingFileName
      then begin
        FMappingFileName      := aValue;
        FScale.KeyMapFileName := FMappingFileName;
      end;
    end;

    procedure   TModScala.SetScaleFileName  ( const aValue: string);
    begin
      if aValue <> FScaleFileName
      then begin
        FScaleFileName  := aValue;
        FScale.FileName := FScaleFileName;
      end;
    end;

    function    TModScala.Lookup( aValue: TSignal): TSignal;
    var
      aNote : Integer;
    begin
      aNote  := Trunc( UnitsToNoteNumber( aValue));
      if aNote = FLastNote
      then Result := FLastOut
      else begin
        Result := FScale.KeyToNote( aNote);

        FLastNote := aNote;
        FLastOut  := Result;
      end;
    end;

//  public

    constructor TModScala.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsSlow := True;
      FScale  := TScalaScale.Create;
    end;

    destructor  TModScala.Destroy; // override;
    begin
      FreeAndNil( FScale);
      inherited;
    end;

    procedure   TModScala.SetDefaults; // override;
    begin
      SlowTick;
    end;

    procedure   TModScala.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'mapping')
      then MappingFileName := aValue
      else if SameText( aName, 'scale')
      then ScaleFileName := aValue
      else inherited;
    end;

    function    TModScala.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'mapping')
      then Result := FMappingFileName
      else if SameText( aName, 'scale')
      then Result := FScaleFileName
      else Result := inherited;
    end;

    procedure   TModScala.DoSlowTick; // override;
    begin
      FOutputs[ scala_out] := Lookup( FInputs[ scala_in]);
    end;

    function    TModScala.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        scala_in     : Result := 'in';
        else           Result := NO_INPUT;
      end;
    end;

    function    TModScala.OutputName( anIndex: Integer): string; // override;
    begin
     case anIndex of
        scala_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModTerraGen = class( TMod)
  public
}

const

  terragen_x            =  0;
  terragen_y            =  1;
  terragen_mute         =  2;
  terragen_algo         =  3;

  terragen_out          =  0;

  terragen_off          =  0; // Output is zero
  terragen_inx          =  1; // Just x input
  terragen_iny          =  2; // Just y input
  terragen_xy           =  3; // Just y input
  terragen_curtis_roads =  4; // Curtis roads standdard example
  terragen_bass         =  5; // all others from http://mymbs.mobeard.org/~pfisher/FOV2-0010016C/FOV2-0010016E/FOV2-001001A3/chapters/26mikelson/index%20.html
  terragen_lead1        =  6;
  terragen_hollow_lead  =  7;
  terragen_pulsar       =  8;
  terragen_flat_growl   =  9;
  terragen_atan         = 10;

    constructor TModTerragen.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModTerraGen.SetDefaults; // override;
    begin
      inherited;
      FDezipperMap := [
        terragen_mute
      ];
      FInputs[ terragen_x] := 1.0;        // Set unused input to 1
      FInputs[ terragen_y] := 1.0;        // Set unused input to 1
    end;

    procedure   TModTerraGen.DoTick; // override;
    var
      x, y : TSignal;
      algo : Integer;
    begin
      x    := Clip( FInputs[ terragen_x], -1, 1);
      y    := Clip( FInputs[ terragen_y], -1, 1);
      algo := SignalToInt( FInputs[ terragen_algo]);
      case algo of
        terragen_off          : FOutputs[ terragen_out] := 0;
        terragen_inx          : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * x;
        terragen_iny          : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * y;
        terragen_xy           : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * x * y;
        terragen_curtis_roads : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * ( x - y) * ( x - 1) * ( x + 1) * ( y - 1) * ( y + 1);
        terragen_bass         : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * Sqrt( Abs( x * x * x * x + y * y * y * y * y - x * x - y * y - x * x * y * y));
        terragen_lead1        : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * ( x - x * x * x / 12 - y * y / 4 + 0.5);
        terragen_hollow_lead  : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * ( -5 * x / ( x * x + y * y + 1));
        terragen_pulsar       : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * Sqr( Sin( x * x + y * y));
        terragen_flat_growl   : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * Sqrt( Abs( ( 1 - x * x) * (( 1 - y) + y * cos( 1 / ( x * x + 0.01))))) - 1;
        terragen_atan         : FOutputs[ terragen_out] := SignalToMute( FInputs[ terragen_mute]) * ArcTan( 15 * x * y) / HALF_PI;
        else                    FOutputs[ terragen_out] := 0;
      end;
    end;

    function    TModTerraGen.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        terragen_x    : Result := 'x';
        terragen_y    : Result := 'y';
        terragen_mute : Result := 'mute';
        terragen_algo : Result := 'algo';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModTerraGen.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        terragen_out : Result := 'out';
        else           Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPluck = class( TMod)
  private
    FMemory      : TSignals;
    FMemLength   : TSignal;
    FInPointer   : Integer;
    FOutValue    : TSignal;
    FOldFeedback : TSignal;
  public
 }

 const

   pluck_pluck       =  0;
   pluck_plucklevel  =  1;
   pluck_feedback    =  2;
   pluck_brightness  =  3;
   pluck_brighctrl   =  4;
   pluck_bclevel     =  5;
   pluck_fm          =  6;
   pluck_fmlevel     =  7;
   pluck_frequency   =  8;
   pluck_cents       =  9;
   pluck_mute        = 10;
   pluck_noise       = 11;
   pluck_noisemod    = 12;
   pluck_noisedepth  = 13;

   pluck_out         =  0;

    constructor TModPluck.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FMemLength   := 1 / LookupFrequency( 0);
      FMemory      := TSignals.Create( Ceil( TimeToSampleCount( FMemLength))); // lowest note about 8.176 Hz - for default tuning
      FOldFeedback := 0.0;
      FIsFast := True;
      FIsSlow := False;
    end;

    destructor  TModPluck.Destroy; // override;
    begin
      FreeAndNil( FMemory);
    end;

    procedure   TModPluck.SetDefaults; // override;
    begin
      FDezipperMap := [
        pluck_feedback,
        pluck_brightness,
        pluck_fmlevel,
        pluck_frequency,
        pluck_cents,
        pluck_mute,
        pluck_noisemod,
        pluck_noisedepth
      ];
    end;

    procedure   TModPluck.DoTick; // override;
    var
      OutPointer : TSignal;
      SampCount  : Integer;
      Alfa       : TSignal;
      Delay      : TSignal;
      Feedback   : TSignal;
      Signal     : TSignal;
      Frequency  : TSignal;
      Noise      : TSignal;
    begin
      SampCount                   := FMemory.Count;
      FInPointer                  := FInPointer mod SampCount;

      Noise                       := Clip( FInputs[ pluck_noise] + FInputs[ pluck_noisemod] * FInputs[ pluck_noisedepth], 0, 1);
      Frequency                   := LookupFrequency( FInputs[ pluck_frequency] + FInputs[ pluck_fm] * FInputs[ pluck_fmlevel] + FInputs[ pluck_cents]);
      Delay                       := 4 / Frequency; // but why 4 / ...
      Alfa                        := Clip( FInputs[ pluck_brightness] + FInputs[ pluck_brighctrl] * FInputs[ pluck_bclevel], 0, 1);
      Feedback                    := FOldFeedback + Alfa * ( FOutValue * FInputs[ pluck_feedback] - FOldFeedback);
      if Random > Noise
      then Feedback := - Feedback;
      Signal                      := FInputs[ pluck_pluck] * FInputs[ pluck_plucklevel] - Feedback;
      FMemory.Signal[ FInPointer] := Clip( Signal, -16, 16);
      OutPointer                  := FMod( SampCount + FInPointer - Delay * ( TimeToSampleCount( FMemLength) - 1), SampCount);
      FOutValue                   := Normalize( FMemory.Interpolate( OutPointer));

      FInPointer                  := ( FInPointer + 1) mod SampCount;
      FOldFeedback                := Feedback;
      FOutputs[ pluck_out]        := Normalize( FOutValue * SignalToMute( FInputs[ pluck_mute]));
    end;

    function    TModPluck.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         pluck_pluck       : Result := 'pluck';
         pluck_plucklevel  : Result := 'plucklevel';
         pluck_feedback    : Result := 'feedback';
         pluck_brightness  : Result := 'brightness';
         pluck_brighctrl   : Result := 'brighctrl';
         pluck_bclevel     : Result := 'bclevel';
         pluck_fm          : Result := 'fm';
         pluck_fmlevel     : Result := 'fmlevel';
         pluck_frequency   : Result := 'frequency';
         pluck_cents       : Result := 'cents';
         pluck_mute        : Result := 'mute';
         pluck_noise       : Result := 'noise';
         pluck_noisemod    : Result := 'noisemod';
         pluck_noisedepth  : Result := 'noisedepth';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModPluck.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pluck_out : Result := 'out';
        else        Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModDisplay = class( TMod)
  private
    FMax1: TSignal;
    FMax2: TSignal;
  public
}

const

  display_in1   = 0;
  display_in2   = 1;
  display_mode  = 2;
  display_speed = 3;


    constructor TModDisplay.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModDisplay.DoTick; // override;
    begin
      if SignalToInt( FInputs[ display_mode]) = 0
      then begin
        FMax1 := Max( FMax1, Abs( FInputs[ display_in1]));
        FMax2 := Max( FMax2, Abs( FInputs[ display_in2]));
      end
      else begin
        FMax1 := Max( FMax1, FInputs[ display_in1]);
        FMax2 := Max( FMax2, FInputs[ display_in2]);
      end;
    end;

    procedure   TModDisplay.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    const
      LOWVAL  = -45.0;
      HIGHVAL =  18.0;
      LOWLED  =   0.0;
      HIGHLED =   1.0;
    var
      adB1  : TSignal;
      adB2  : TSignal;
      Speed : TSignal;
    begin
      Speed := 0.707 * ( 0.2 + 0.8 * FInputs[ display_speed]);
      adB1  := RangeMap( Clip( SignalTodB( FMax1), LOWVAL, HIGHVAL), LOWVAL, HIGHVAL, LOWLED, HIGHLED);
      adB2  := RangeMap( Clip( SignalTodB( FMax2), LOWVAL, HIGHVAL), LOWVAL, HIGHVAL, LOWLED, HIGHLED);
      aCallback( Self, Format( '%s/%s/level1/%f', [ aPrefix, Name, adB1], AppLocale));
      aCallback( Self, Format( '%s/%s/level2/%f', [ aPrefix, Name, adB2], AppLocale));
      FMax1 := FMax1 * Speed;
      FMax2 := FMax2 * Speed;
    end;

    function    TModDisplay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        display_in1   : Result := 'in1';
        display_in2   : Result := 'in2';
        display_mode  : Result := 'mode';
        display_speed : Result := 'speed';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModDisplay.OutputName( anIndex: Integer): string; // override;
    begin
      Result := NO_OUTPUT;
    end;


{ ========
  TModStateSpace = class( TMod)
  private
    FA        : TSignalMatrix;
    FB        : TSignalMatrix;
    FC        : TSignalMatrix;
    FD        : TSignalMatrix;
    FX        : TSignalMatrix;
    FNewX     : TSignalMatrix;
    FOut      : TSignalMatrix;
    FIn       : TSignalMatrix;
    FRunnable : Boolean;
  private
}

    function    TModStateSpace.Consistent: Boolean;
    begin
      Result :=
        FA.IsSquare                     and
        ( FA   .RowCount = FX.RowCount) and
        ( FX   .ColCount = 1          ) and
        ( FNewX.RowCount = FX.RowCount) and
        ( FNewX.ColCount = FX.ColCount) and
        ( FB   .RowCount = FX.RowCount) and
        ( FB   .ColCount = InputCount ) and
        ( FC   .RowCount = OutputCount) and
        ( FC   .ColCount = FX.ColCount) and
        ( FD   .RowCount = OutputCount) and
        ( FD   .ColCount = InputCount );
    end;

    procedure   TModStateSpace.ClearState;
    begin
      FX   .RowCount := FA.RowCount;
      FX   .ColCount := 1;
      FNewX.RowCount := FA.RowCount;
      FNewX.ColCount := 1;
      FX   .Clear;
      FNewX.Clear;
    end;

//  protected

    procedure   TModStateSpace.Run;
    var
      Tmp : TSignalMatrix;
    begin
      Tmp := TSignalMatrix.Create( FInputs);
      try
        FIn.Assign( Tmp);
        FIn.Normalize;

        Tmp.Assign( FB);                          // FNewX( z) = FA.FX( z) + FB.FIn( z)
        Tmp.Normalize;
        Tmp.Mult( FIn);
        Tmp.Normalize;
        FNewX.Assign( FA);
        FNewX.Normalize;
        FNewX.Mult( FX);
        FNewX.Normalize;
        FNewX.Add( Tmp);
        FNewX.Normalize;

        Tmp.Assign( FD);                          // FOut( z) = FC.FX( z) + FD.FIn( z)
        Tmp.Normalize;
        Tmp.Mult( FIn);
        Tmp.Normalize;
        FOut.Assign( FC);
        FOut.Normalize;
        FOut.Mult( FX);
        FOut.Normalize;
        FOut.Add( Tmp);
        FOut.Normalize;

        FOut.PopulateVector( FOutputs);

        FX.Assign( FNewX);
      finally
        Tmp.Free;
      end;
    end;

//  public

    constructor TModStateSpace.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIn   := TSignalMatrix.Create;
      FOut  := TSignalMatrix.Create;
      FA    := TSignalMatrix.Create;
      FB    := TSignalMatrix.Create;
      FC    := TSignalMatrix.Create;
      FD    := TSignalMatrix.Create;
      FX    := TSignalMatrix.Create;
      FNewX := TSignalMatrix.Create;
      Setup;
      ClearState;
      FIsFast := True;
      FIsSlow := False;
      FRunnable := Consistent;
    end;

    destructor  TModStateSpace.Destroy; // override;
    begin
      FRunnable := False;
      FreeAndNil( FNewX);
      FreeAndNil( FX   );
      FreeAndNil( FD   );
      FreeAndNil( FC   );
      FreeAndNil( FB   );
      FreeAndNil( FA   );
      FreeAndNil( FOut );
      FreeAndNil( FIn  );
      inherited;
    end;

    procedure   TModStateSpace.DoTick; // override;
    begin
      if FRunnable
      then Run;
    end;

    function    TModStateSpace.InputName( anIndex: Integer): string; // override;
    begin
      Result := NO_INPUT;
    end;

    function    TModStateSpace.OutputName( anIndex: Integer): string; // override;
    begin
      Result := NO_OUTPUT;
    end;


{ ========
  TModMidiIn = class( TMod)
  public
}

    constructor TModMidiIn.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsMidi := True;
    end;


{ ========
  TModMidiNoteIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FLastNote    : Byte;
    FLastChannel : Byte;
    FOldChannel  : Byte;
    FGate        : Boolean;
    FNote        : TSignal;
    FVelocityOn  : TSignal;
    FVelocityOff : TSignal;
  public
}

const

  midinotein_ch     = 0;

  midinotein_gate   = 0;
  midinotein_note   = 1;
  midinotein_velon  = 2;
  midinotein_veloff = 3;
{ midinotein_bend   = 4; }


    procedure   TModMidiNoteIn.NotesOff;
    begin
      FGate        := False;
      FVelocityOn  := 0;
      FVelocityOff := 0;
    end;

    procedure   TModMidiNoteIn.AcceptMidi( const aMsg: TMidiMessage); // override;
    begin
      if ( FChannel = 255) or ( aMsg.Channel = FChannel)            // 255 being omni
      then begin
        if Length( aMsg.Data) = 2
        then begin
          if aMSg.Command = CNoteOn
          then begin
            FGate        := True;
            FLastNote    := aMsg.Data[ 0];
            FLastChannel := aMSg.Channel;
            FNote        := NoteNumberToUnits( FLastNote - MiddleNote);
            FVelocityOn  := aMsg.Data[ 1] / 128;
          end
          else if ( aMsg.Command = CNoteOff) and ( aMsg.Data[ 0] = FLastNote) and ( aMsg.Channel = FLastChannel)
          then begin
            FGate        := False;
            FVelocityOff := aMsg.Data[ 1] / 128;
          end
          else if ( aMsg.Command = CControlChange) and ( aMsg.Data[ 0] = CCAllNotesOff) and ( aMsg.Data[ 1] = 0) // All notes off
          then NotesOff;
        end;
      end;
    end;

    procedure   TModMidiNoteIn.DoSlowTick; // override;
    begin
      if ResetFlag
      then begin
        ResetFlag := False;
        NotesOff;
      end;

      FChannel := SignalToInt( Clip( FInputs[ midinotein_ch], 0, 16));
      if FChannel = 16
      then FChannel := 255;
      if FOldChannel <> FChannel
      then NotesOff;
      FOutputs[ midinotein_gate  ] := LogicToSignal( FGate);
      FOutputs[ midinotein_note  ] := FNote;
      FOutputs[ midinotein_velon ] := FVelocityOn;
      FOutputs[ midinotein_veloff] := FVelocityOff;
      FOldChannel                  := FChannel;
    end;

    procedure   TModMidiNoteIn.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiNoteIn.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midinotein_ch : Result := 'ch';
        else            Result := NO_INPUT;
      end;
    end;

    function    TModMidiNoteIn.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midinotein_gate   : Result := 'gate';
        midinotein_note   : Result := 'note';
        midinotein_velon  : Result := 'velon';
        midinotein_veloff : Result := 'veloff';
     // midinotein_bend   : Result := 'bend';
        else                Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMidiCCIn = class( TModMidiIn)
  private
    FChannel     : Byte;
    FCC          : Byte;
    FGate        : Boolean;
    FValue       : TSignal;
    FCounter     : Integer;
    FLedCounter  : Integer;
  public
}

const

  midiccin_ch    = 0;
  midiccin_cc    = 1;

  midiccin_gate  = 0;
  midiccin_value = 1;


    procedure   TModMidiCCIn.AcceptMidi( const aMsg: TMidiMessage); // override;
    begin
      if ( FChannel = 255) or ( aMsg.Channel = FChannel)            // 255 being omni
      then begin
        if Length( aMsg.Data) = 2
        then begin
          if ( aMsg.Command = CControlChange) and ( aMsg.Data[ 0] = FCC)
          then begin
            FValue      := aMsg.Data[ 1];
            FCounter    := 44;                 //   1 ms gate pulse
            FLedCounter := 11025;              // 250 ms LED  pulse
            FGate       := True;
          end;
        end;
      end;
    end;

    procedure   TModMidiCCIn.DoSlowTick; // override;
    begin
      if FLedCounter > 0
      then Dec( FLedCounter);
      FChannel := SignalToInt( FInputs[ midiccin_ch]);
      FCC      := SignalToInt( FInputs[ midiccin_cc]);
      if FChannel = 16
      then FChannel := 255;
      if FCounter > 0
      then Dec( FCounter)
      else FGate := False;
      FOutputs[ midiccin_gate ] := LogicToSignal( FGate);
      FOutputs[ midiccin_value] := FValue / 128;
    end;

    procedure   TModMidiCCIn.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FLedCounter <> 0)], AppLocale));
    end;

    function    TModMidiCCIn.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midiccin_ch : Result := 'ch';
        midiccin_cc : Result := 'cc';
        else          Result := NO_INPUT;
      end;
    end;

    function    TModMidiCCIn.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midiccin_gate  : Result := 'gate';
        midiccin_value : Result := 'value';
        else             Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModMidiOut = class( TMod)
  public
}

    constructor TModMidiOut.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := False;
      FIsMidi := False;
    end;


{ ========
  TModMidiNoteOut = class( TModMidiOut)
  private
    FLastNote    : Byte;
    FLastChannel : Byte;
    FGate        : Boolean;
    FDidSend     : Boolean;
  public
}

const

  midinoteout_ch     = 0;
  midinoteout_gate   = 1;
  midinoteout_note   = 2;
  midinoteout_velon  = 3;
  midinoteout_veloff = 4;


    procedure   TModMidiNoteOut.SetDefaults; // override;
    begin
      FInputs[ midinoteout_gate  ] := 0;    // Gate off
      FInputs[ midinoteout_note  ] := 0;    // Note 0
      FInputs[ midinoteout_velon ] := 1;    // Max velon
      FInputs[ midinoteout_veloff] := 1;    // Max veloff
    end;

    procedure   TModMidiNoteOut.DoSlowTick; // override;
    // FNote := ( FLastNote - 69) / 128;
    var
      NewGate : Boolean;
      aMsg    : TMidiMessage;
    begin
      NewGate := SignalToLogic( FInputs[ midinoteout_gate]);
      if NewGate <> FGate
      then begin
        if NewGate
        then begin
          FLastChannel := SignalToInt( FInputs[ midinoteout_ch]);
          FLastNote    := SignalToInt( Clip( FInputs[ midinoteout_note] * 128 + MiddleNote, 0, 127));
          SetLength( aMsg.Data, 2);
          aMsg.Command  := CNoteOn;
          aMsg.Channel  := FLastChannel;
          aMsg.Data[ 0] := FlastNote;
          aMsg.Data[ 1] := SignalToInt( Clip( FInputs[ midinoteout_velon], 0, 1) * 127);
          SendMidi( aMsg);
          FDidSend := True;
        end
        else begin
          if FDidSend
          then begin
            FDidSend := False;
            SetLength( aMsg.Data, 2);
            aMsg.Command  := CNoteOff;
            aMsg.Channel  := FLastChannel;
            aMsg.Data[ 0] := FlastNote;
            aMsg.Data[ 1] := SignalToInt( Clip( FInputs[ midinoteout_veloff], 0, 1) * 127);
            SendMidi( aMsg);
          end;
        end;
        FGate := NewGate;
      end;
    end;

    procedure   TModMidiNoteOut.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiNoteOut.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midinoteout_ch     : Result := 'ch';
        midinoteout_gate   : Result := 'gate';
        midinoteout_note   : Result := 'note';
        midinoteout_velon  : Result := 'velon';
        midinoteout_veloff : Result := 'veloff';
        else                 Result := NO_INPUT;
      end;
    end;


{ ========
  TModMidiCCOut = class( TModMidiOut)
  private
    FGate  : Boolean;
  public
}

const

  midiccout_ch    = 0;
  midiccout_cc    = 1;
  midiccout_gate  = 2;
  midiccout_value = 3;


    procedure   TModMidiCCOut.DoSlowTick; // override;
    var
      NewGate : Boolean;
      aMsg    : TMidiMessage;
    begin
      NewGate := SignalToLogic( FInputs[ midiccout_gate]);
      if NewGate <> FGate
      then begin
        if NewGate
        then begin
          SetLength( aMsg.Data, 2);
          aMsg.Command  := CControlChange;
          aMsg.Channel  := SignalToInt( FInputs[ midiccout_ch]);
          aMsg.Data[ 0] := SignalToInt( FInputs[ midiccout_cc]);
          aMsg.Data[ 1] := SignalToInt( Clip( FInputs[ midiccout_value], 0, 1) * 127);
          SendMidi( aMsg);
        end;
        FGate := NewGate;
      end;
    end;

    procedure   TModMidiCCOut.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/gate/%f', [ aPrefix, Name, LogicToSignal( FGate)], AppLocale));
    end;

    function    TModMidiCCOut.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midiccout_ch    : Result := 'ch';
        midiccout_cc    : Result := 'cc';
        midiccout_gate  : Result := 'gate';
        midiccout_value : Result := 'value';
        else              Result := NO_INPUT;
      end;
    end;


{ ========
  TModMidiPlayer = class( TMod)
  private
    FState           : TMidiPlayerState;
    FReader          : TWrenMidiReader;
    FEvent           : TWrenMidiTimedEvent;
    FFileName        : string;
    FReading         : Boolean;
    FCounter         : Integer;
    FChannel         : Byte;
    FOldTrig         : Boolean;
    FOldReset        : Boolean;
    FGateCount       : Integer;
    FNote            : TSignal;
    FVelocityOn      : TSignal;
    FVelocityOff     : TSignal;
    FDoneCounter     : Integer;
    FDoneLedCounter  : Integer;
    FResetLedCounter : Integer;
    FChOffset        : Integer;
  public
    property    FileName: string read FFileName write SetFileName;
  private
}

const

  midiplayer_trig      = 0;
  midiplayer_ch        = 1;
  midiplayer_reset     = 2;
  midiplayer_mode      = 3;

  midiplayer_note      = 0;
  midiplayer_velon     = 1;
  midiplayer_veloff    = 2;
  midiplayer_gate      = 3;
  midiplayer_done      = 4;

  midiplayer_mode_once = 0;
  midiplayer_mode_loop = 1;

    procedure   TModMidiPlayer.SetFileName( const aValue: string);
    begin
      try
        if aValue <> FFileName
        then begin
          if Assigned( FReader)
          then begin
            if FileExists( aValue)
            then begin
              FReading := True;
              try
                NotesOff;
                FFileName := aValue;
                FReader.ReadFile( FFileName);
                if FReader.Format = 1
                then FChOffset :=  0
                else FChOffset := -1;
                FEvent := GetNextEvent;
              finally
                FReading := False;
              end;
            end
            else begin
              NotesOff;
              FEvent    := nil;
              FFileName := aValue;
            end;
          end;
        end;
      except
        on E: Exception
        do KilledException( E, True);
      end;
    end;

    function    TModMidiPlayer.GetNextEvent: TWrenMidiTimedEvent;
    begin
      Result := FReader.NextTimedEvent( FChannel + FChOffset);
    end;

//  public

    procedure   TModMidiPlayer.NotesOff;
    begin
      FGateCount   := 0;
      FVelocityOn  := 0;
      FVelocityOff := 0;
    end;

    constructor TModMidiPlayer.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FReader   := TWrenMidiReader.Create;
      FIsFast   := False;
      FState    := msActive;
      FChOffset := -1;
      FIsFast   := False;
      FIsSlow   := True;
    end;

    destructor  TModMidiPlayer.Destroy; // override;
    begin
      FreeAndNil( FReader);
    end;

    procedure   TModMidiPlayer.DoSlowTick; // override;
    var
      Trig      : Boolean;
      Reset     : Boolean;
      Channel   : Integer;
      Mode      : Integer;
      MidiEvent : TWrenMidiEvent;
    begin
      Reset := SignalToLogic( FInputs[ midiplayer_reset]);
      if ResetFlag or ( Reset and not FOldReset)
      then begin
        ResetFlag := False;
        FOldReset := Reset;
        NotesOff;
        FCounter := 0;
        FOldTrig := False;
        if Assigned( FReader)
        then FReader.Reset;
        FEvent := nil;
        FResetLedCounter := 1000;
        FState := msActive;
      end;

      Channel  := SignalToInt  ( FInputs[ midiplayer_ch   ]);
      if Channel <> FChannel
      then NotesOff;
      FChannel := Channel;

      if FReading
      then NotesOff
      else begin
        Trig := SignalToLogic( FInputs[ midiplayer_trig]);
        if Trig and not FOldTrig
        then begin

          if FCounter = 0
          then begin
            if Assigned( FEvent)
            then begin

              MidiEvent := FEvent.Event;
              if FEvent.IsAllNotesOffEvent
              then NotesOff
              else if FEvent.IsEndOfTrackEvent
              then begin
                Mode := SignalToInt( FInputs[ midiplayer_mode]);
                if Mode = midiplayer_mode_once
                then FState := msDone;
                FReader.Reset;
                FEvent := nil;
                NotesOff;
                FDoneCounter    :=  100;
                FDoneLedCounter := 1000;
              end
              else if FEvent.IsNoteOffEvent
              then begin
                FVelocityOff := TWrenMidiNoteOffEvent( MidiEvent).Velocity / 128;
                if FGateCount > 0
                then Dec( FGateCount);
              end
              else if FEvent.IsNoteOnEvent
              then begin
                FNote       := NoteNumberToUnits( TWrenMidiNoteOffEvent( MidiEvent).Key - MiddleNote);
                FVelocityOn := TWrenMidiNoteOffEvent( MidiEvent).Velocity / 128;
                if FVelocityOn = 0
                then Dec( FGateCount)
                else Inc( FGateCount);
              end;
            end;

            if FState = msActive
            then begin
              FEvent := GetNextEvent;
              if not Assigned( FEvent)
              then begin
                FReader.Reset;
                FEvent:= GetNextEvent;
              end;
              if Assigned( FEvent)
              then FCounter := FEvent.DeltaTime.DeltaTime;
            end
            else begin
              FEvent := nil;
              NotesOff;
            end;
          end
          else Dec( FCounter)

        end;
        FOldTrig := Trig;
      end;

      if FDoneCounter     > 0 then Dec( FDoneCounter    );
      if FDoneLedCounter  > 0 then Dec( FDoneLedCounter );
      if FResetLedCounter > 0 then Dec( FResetLedCounter);

      FOutputs[ midiplayer_note  ] := FNote;
      FOutputs[ midiplayer_velon ] := FVelocityOn;
      FOutputs[ midiplayer_veloff] := FVelocityOff;
      FOutputs[ midiplayer_gate  ] := LogicToSignal( FGateCount   > 0);
      FOutputs[ midiplayer_done  ] := LogicToSignal( FDoneCounter > 0);
    end;

    procedure   TModMidiPlayer.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModMidiPlayer.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FileName
      else Result := inherited;
    end;

    procedure   TModMidiPlayer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, LogicToSignal( FOldTrig            )], AppLocale));
      aCallback( Self, Format( '%s/%s/gate/%f' , [ aPrefix, Name, LogicToSignal( FGateCount       > 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/done/%f' , [ aPrefix, Name, LogicToSignal( FDoneLedCounter  > 0)], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, LogicToSignal( FResetLedCounter > 0)], AppLocale));
    end;

    function    TModMidiPlayer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midiplayer_trig  : Result := 'trig';
        midiplayer_ch    : Result := 'ch';
        midiplayer_reset : Result := 'reset';
        midiplayer_mode  : Result := 'mode';
        else               Result := NO_INPUT;
      end;
    end;

    function    TModMidiPlayer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        midiplayer_note   : Result := 'note';
        midiplayer_velon  : Result := 'velon';
        midiplayer_veloff : Result := 'veloff';
        midiplayer_gate   : Result := 'gate';
        midiplayer_done   : Result := 'done';
       else                Result := NO_OUTPUT;
      end;
    end;


{$ifdef win32}
{ ========
  TModReverb = class( TMod)
  private
    FReverb : TReverb;
  public
}

const

  reverb_inl      = 0;
  reverb_inr      = 1;
  reverb_roomsize = 2;
  reverb_damp     = 3;
  reverb_wet      = 4;
  reverb_dry      = 5;
  reverb_width    = 6;
  reverb_mode     = 7;
  reverb_mute     = 8;
  reverb_freezein = 9;

  reverb_outl     = 0;
  reverb_outr     = 1;

    constructor TModReverb.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FReverb := TReverb.Create( System_Rate);
      FIsFast := True;
      FIsSlow := True;
    end;

    destructor  TModReverb.Destroy; // override;
    begin
      FreeAndNil( FReverb);
      inherited;
    end;

    procedure   TModReverb.DoSlowTick; // override;
    begin
      with FReverb
      do begin
        setroomsize( FInputs[ reverb_roomsize]);
        setdamp    ( FInputs[ reverb_damp    ]);
        setwet     ( FInputs[ reverb_wet     ]);
        setdry     ( FInputs[ reverb_dry     ]);
        setwidth   ( FInputs[ reverb_width   ]);
        setmode    ( LogicToSignal( SignalToLogic( FInputs[ reverb_mode]) xor SignalToLogic( FInputs[ reverb_freezein])));
      end;
    end;

    procedure   TModReverb.SetDefaults; // override;
    begin
      FDezipperMap := [
        reverb_roomsize,
        reverb_damp    ,
        reverb_wet     ,
        reverb_dry     ,
        reverb_width   ,
        reverb_mute
      ];
    end;

    procedure   TModReverb.DoTick; // override;
    var
      OutL : Single;
      OutR : Single;
      Mute : TSignal;
    begin
      FReverb.processreplace( Normalize( FInputs[ reverb_inl]), Normalize( FInputs[ reverb_inr]), OutL, OutR);
      Mute := SignalToMute( FInputs[ reverb_mute]);
      FOutputs[ reverb_outl] := Normalize( OutL) * Mute;
      FOutputs[ reverb_outr] := Normalize( OutR) * Mute;
    end;

    procedure   TModReverb.SampleRateChanged; // override;
    var
      OldReverb : TReverb;
      NewReverb : TReverb;
    begin
      OldReverb := FReverb;
      NewReverb := TReverb.Create( System_Rate);
      FReverb   := NewReverb;
      FreeAndNil( OldReverb);
    end;

    function    TModReverb.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        reverb_inl      : Result := 'inl';
        reverb_inr      : Result := 'inr';
        reverb_roomsize : Result := 'roomsize';
        reverb_damp     : Result := 'damp';
        reverb_wet      : Result := 'wet';
        reverb_dry      : Result := 'dry';
        reverb_width    : Result := 'width';
        reverb_mode     : Result := 'mode';
        reverb_mute     : Result := 'mute';
        reverb_freezein : Result := 'freezein';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModReverb.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        reverb_outl : Result := 'outl';
        reverb_outr : Result := 'outr';
        else          Result := NO_OUTPUT;
      end;
    end;
{$endif}

{ ========
  TModEnvelope = class( TMod)
  private
    FValue     : TSignal;
    FAttack    : TSignal;
    FRelease   : TSignal;
    FPeakValue : TSignal;
    FPeakDecay : TSignal;
  public
}


const

  envelope_in          = 0;
  envelope_attack      = 1;
  envelope_release     = 2;
  envelope_peakdecay   = 3;
  envelope_peakdecayon = 4;

  envelope_out         = 0;
  envelope_invout      = 1;

    constructor TModEnvelope.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModEnvelope.SetDefaults; // override;
    begin
      FInputs[ envelope_attack ] := 1;
      FInputs[ envelope_release] := 1;
      DoSlowTick;
    end;

    procedure   TModEnvelope.DoSlowTick; // override;
    begin
      FAttack    := Exp( - 1 / Max( 1, SlowTimeToSampleCount( FInputs[ envelope_attack   ])));
      FRelease   := Exp( - 1 / Max( 1, SlowTimeToSampleCount( FInputs[ envelope_release  ])));
      FPeakDecay := Exp( - 1 / Max( 1, SlowTimeToSampleCount( FInputs[ envelope_peakdecay])));
    end;

    procedure   TModEnvelope.DoTick; // override;
    var
      aValue : TSignal;
      aPeak  : TSignal;
    begin
      aValue := Abs( FInputs[ envelope_in]);
      if FValue < aValue
      then begin
        FValue := FValue * FAttack;
        FValue := FValue + ( 1 - FAttack) * aValue;
        if FValue > FPeakValue
        then FPeakValue := FValue;
      end
      else begin
        FValue := FValue * FRelease;
        FValue := FValue + ( 1 - FRelease) * aValue;
        if SignalToLogic( FInputs[ envelope_peakdecayon])
        then begin
          aPeak      := FPeakValue;
          FPeakValue := FPeakValue * FPeakDecay;
          FPeakValue := FPeakValue + ( 1 - FPeakDecay) * aPeak;
        end;
      end;
      FOutputs[ envelope_out   ] := FValue;
      FOutputs[ envelope_invout] := FPeakValue - FValue;
    end;

    function    TModEnvelope.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envelope_in          : Result := 'in';
        envelope_attack      : Result := 'attack';
        envelope_release     : Result := 'release';
        envelope_peakdecay   : Result := 'peakdecay';
        envelope_peakdecayon : Result := 'peakdecayon';
        else                   Result := NO_INPUT;
      end;
    end;

    function    TModEnvelope.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        envelope_out    : Result := 'out';
        envelope_invout : Result := 'invout';
        else              Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModCompressor = class( TMod)
  private
    const
      COMP_MAX_GAIN =  64.0;
      COMP_MIN_GAIN = -32.0;
  private
    FValue       : TSignal;
    FAttack      : TSignal;
    FRelease     : TSignal;
    FRefLevel    : TSignal;
    FPreGain     : TSignal;
    FGain        : TSignal;
    FBypass      : Integer;
    FOutL        : TSignal;
    FOutR        : TSignal;
    FLoopAmp     : TSignal;
    FMaxLoopAmp  : TSignal;
    FOldAttack   : TSignal;
    FOldRelease  : TSignal;
    FMinDbVal    : TSignal;
    FMaxDbVal    : TSignal;
  public
}

const

  compressor_inl            = 0;
  compressor_inr            = 1;
  compressor_reflevel       = 2;
  compressor_gain           = 3;
  compressor_attack         = 4;
  compressor_release        = 5;
  compressor_bypass         = 6;
  compressor_preamp         = 7;

  compressor_outl           = 0;
  compressor_outr           = 1;

  compressor_bypass_off     = 0; // normal working mode
  compressor_bypass_byp     = 1; // bypass, outputs are same as inputs

    constructor TModCompressor.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FMaxDbVal := Power( 10, COMP_MAX_GAIN / 20);
      FMinDbVal := Power( 10, COMP_MIN_GAIN / 20);
      FIsFast   := True;
      FIsSlow   := True;
    end;

    procedure   TModCompressor.SetDefaults; // override;
    begin
      FDezipperMap := [
        compressor_reflevel,
        compressor_gain    ,
        compressor_attack  ,
        compressor_release
      ];
      FInputs[ compressor_attack ] := 1;
      FInputs[ compressor_release] := 1;
      DoSlowTick;
    end;

    procedure   TModCompressor.DoSlowTick; // override;
    var
      aValue: TSignal;
    begin
      aValue := FInputs[ compressor_attack];
      if aValue <> FOldAttack
      then begin
        FOldAttack := aValue;
        FAttack    := Exp( - 1 / Max( 1, SlowTimeToSampleCount( aValue)));
      end;

      aValue := FInputs[ compressor_release];
      if aValue <> FOldRelease
      then begin
        FOldRelease := aValue;
        FRelease    := Exp( - 1 / Max( 1, SlowTimeToSampleCount( aValue)));
      end;

      FRefLevel   := FInputs[ compressor_reflevel];
      FPreGain    := FInputs[ compressor_preamp  ];
      FGain       := FInputs[ compressor_gain    ];
      FBypass     := SignalToInt( FInputs[ compressor_bypass]);
      FMaxLoopAmp := Max( FMaxLoopAmp, FLoopAmp);
    end;

    procedure   TModCompressor.DoTick; // override;
    var
      anInL      : TSignal;
      anInR      : TSignal;
      anOutLevel : TSignal;
      anEnvValue : TSignal;
    begin
      anInL      := FPreGain * FInputs [ compressor_inl];
      anInR      := FPreGain * FInputs [ compressor_inr];
      anOutlevel := Max( Abs( FOutL), Abs( FOutR));

      if FValue < anOutLevel
      then anEnvValue := FAttack  * ( FValue - anOutLevel) + anOutLevel
      else anEnvValue := FRelease * ( FValue - anOutLevel) + anOutLevel;

      FValue   := CompressorSmoothing * ( anEnvValue - FValue) + FValue;
      FLoopAmp := Clip( FGain * ( FRefLevel -  FValue), FMinDbVal, FMaxDbVal);
      FOutL    := FLoopAmp * anInL;
      FOutR    := FLoopAmp * anInR;

      case FBypass of
        compressor_bypass_byp : // outputs are same as inputs
          begin
            FOutputs[ compressor_outl] := FInputs[ compressor_inl]; // Do not use anInL - that one has pre-gain applied
            FOutputs[ compressor_outr] := FInputs[ compressor_inr]; // Likewise for anInR
          end;
        else begin              // normal operation
          FOutputs[ compressor_outl] := FOutL;
          FOutputs[ compressor_outr] := FOutR;
        end;
      end;
    end;

    procedure   TModCompressor.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    const
      LOWLED  =   0.000;
      HIGHLED =   1.000;
      DECAY   =   0.707;
    var
      adB : TSignal;
    begin
      adB := RangeMap( Clip( SignalTodB( FMaxLoopAmp), COMP_MIN_GAIN, COMP_MAX_GAIN), COMP_MIN_GAIN, COMP_MAX_GAIN, LOWLED, HIGHLED);
      aCallback( Self, Format( '%s/%s/level/%f', [ aPrefix, Name, adB], AppLocale));
      FMaxLoopAmp := FMaxLoopAmp * DECAY;
    end;

    function    TModCompressor.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        compressor_inl       : Result := 'inl';
        compressor_inr       : Result := 'inr';
        compressor_reflevel  : Result := 'reflevel';
        compressor_gain      : Result := 'gain';
        compressor_attack    : Result := 'attack';
        compressor_release   : Result := 'release';
        compressor_bypass    : Result := 'bypass';
        compressor_preamp    : Result := 'preamp';
        else                   Result := NO_INPUT;
      end;
    end;

    function    TModCompressor.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        compressor_outl : Result := 'outl';
        compressor_outr : Result := 'outr';
        else              Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModFormant = class( TMod)
  private
    FCoeffStore    : array[ 0 .. 10] of TSignalArray;
    FCoeffiecients : array[ 0 .. 10] of TSignal;
    FMemory        : TSignalArray;
    FVowel         : TSignal;
  public
}

const

  formant_in         = 0;
  formant_vowel      = 1;
  formant_vowelin    = 2;
  formant_controlmod = 3;

  formant_out        = 0;

    constructor TModFormant.Create( aParent: TSynthPatch; const aName: string); // override;
    var
      i : Integer;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
      SetLength( FMemory, 10);
      for i := 0 to 10
      do FCoeffStore[ i] := TSignals.Create( 5);

      // A

      FCoeffStore[  0].FSignals[ 0] :=    8.11044e-06;
      FCoeffStore[  1].FSignals[ 0] :=    8.943665402;
      FCoeffStore[  2].FSignals[ 0] :=  -36.83889529 ;
      FCoeffStore[  3].FSignals[ 0] :=   92.01697887 ;
      FCoeffStore[  4].FSignals[ 0] := -154.337906   ;
      FCoeffStore[  5].FSignals[ 0] :=  181.6233289  ;
      FCoeffStore[  6].FSignals[ 0] := -151.8651235  ;
      FCoeffStore[  7].FSignals[ 0] :=   89.09614114 ;
      FCoeffStore[  8].FSignals[ 0] :=  -35.10298511 ;
      FCoeffStore[  9].FSignals[ 0] :=    8.388101016;
      FCoeffStore[ 10].FSignals[ 0] :=   -0.923313471;

      // E

      FCoeffStore[  0].FSignals[ 1] :=    4.36215e-06;
      FCoeffStore[  1].FSignals[ 1] :=    8.90438318 ;
      FCoeffStore[  2].FSignals[ 1] :=  -36.55179099 ;
      FCoeffStore[  3].FSignals[ 1] :=   91.05750846 ;
      FCoeffStore[  4].FSignals[ 1] := -152.422234   ;
      FCoeffStore[  5].FSignals[ 1] :=  179.1170248  ;
      FCoeffStore[  6].FSignals[ 1] := -149.6496211  ;
      FCoeffStore[  7].FSignals[ 1] :=   87.78352223 ;
      FCoeffStore[  8].FSignals[ 1] :=  -34.60687431 ;
      FCoeffStore[  9].FSignals[ 1] :=    8.282228154;
      FCoeffStore[ 10].FSignals[ 1] :=   -0.914150747;

      // I

      FCoeffStore[  0].FSignals[ 2] :=    3.33819e-06;
      FCoeffStore[  1].FSignals[ 2] :=    8.893102966;
      FCoeffStore[  2].FSignals[ 2] :=  -36.49532826 ;
      FCoeffStore[  3].FSignals[ 2] :=   90.96543286 ;
      FCoeffStore[  4].FSignals[ 2] := -152.4545478  ;
      FCoeffStore[  5].FSignals[ 2] :=  179.4835618  ;
      FCoeffStore[  6].FSignals[ 2] := -150.315433   ;
      FCoeffStore[  7].FSignals[ 2] :=   88.43409371 ;
      FCoeffStore[  8].FSignals[ 2] :=  -34.98612086 ;
      FCoeffStore[  9].FSignals[ 2] :=    8.407803364;
      FCoeffStore[ 10].FSignals[ 2] :=   -0.932568035;

       // O

      FCoeffStore[  0].FSignals[ 3] :=    1.13572e-06;
      FCoeffStore[  1].FSignals[ 3] :=    8.994734087;
      FCoeffStore[  2].FSignals[ 3] :=  -37.2084849  ;
      FCoeffStore[  3].FSignals[ 3] :=   93.22900521 ;
      FCoeffStore[  4].FSignals[ 3] := -156.6929844  ;
      FCoeffStore[  5].FSignals[ 3] :=  184.596544   ;
      FCoeffStore[  6].FSignals[ 3] := -154.3755513  ;
      FCoeffStore[  7].FSignals[ 3] :=   90.49663749 ;
      FCoeffStore[  8].FSignals[ 3] :=  -35.58964535 ;
      FCoeffStore[  9].FSignals[ 3] :=    8.478996281;
      FCoeffStore[ 10].FSignals[ 3] :=   -0.929252233;

      // U

      FCoeffStore[  0].FSignals[ 4] :=    4.09431e-07;
      FCoeffStore[  1].FSignals[ 4] :=    8.997322763;
      FCoeffStore[  2].FSignals[ 4] :=  -37.20218544 ;
      FCoeffStore[  3].FSignals[ 4] :=   93.11385476 ;
      FCoeffStore[  4].FSignals[ 4] := -156.2530937  ;
      FCoeffStore[  5].FSignals[ 4] :=  183.7080141  ;
      FCoeffStore[  6].FSignals[ 4] := -153.2631681  ;
      FCoeffStore[  7].FSignals[ 4] :=   89.59539726 ;
      FCoeffStore[  8].FSignals[ 4] :=  -35.12454591 ;
      FCoeffStore[  9].FSignals[ 4] :=    8.338655623;
      FCoeffStore[ 10].FSignals[ 4] :=   -0.910251753;
    end;

    destructor  TModFormant.Destroy; // override;
    var
      i : Integer;
    begin
      for i := 0 to 10
      do FCoeffStore[ i].Free;
      inherited;
    end;

    procedure   TModFormant.SetDefaults; // override;
    begin
      FDezipperMap := [
        formant_vowel,
        formant_controlmod
      ];
    end;

    procedure   TModFormant.DoSlowTick; // override;
    var
      i : Integer;
    begin
      FVowel := Clip( 4 * ( FInputs[ formant_vowel] + FInputs[ formant_controlmod] * FInputs[ formant_vowelin]), 0.01, 3.99);
      for i := 0 to 10
      do FCoeffiecients[ i] := Normalize( FCoeffStore[ i].Interpolate( FVowel));
    end;

    procedure   TModFormant.DoTick; // override;
    begin
      FOutputs[ formant_out] :=
        Normalize(
          FCoeffiecients[  0] * FInputs[ formant_in] +
          FCoeffiecients[  1] * FMemory[ 0]          +
          FCoeffiecients[  2] * FMemory[ 1]          +
          FCoeffiecients[  3] * FMemory[ 2]          +
          FCoeffiecients[  4] * FMemory[ 3]          +
          FCoeffiecients[  5] * FMemory[ 4]          +
          FCoeffiecients[  6] * FMemory[ 5]          +
          FCoeffiecients[  7] * FMemory[ 6]          +
          FCoeffiecients[  8] * FMemory[ 7]          +
          FCoeffiecients[  9] * FMemory[ 8]          +
          FCoeffiecients[ 10] * FMemory[ 9]
        );

      FMemory[ 9] := FMemory[ 8];
      FMemory[ 8] := FMemory[ 7];
      FMemory[ 7] := FMemory[ 6];
      FMemory[ 6] := FMemory[ 5];
      FMemory[ 5] := FMemory[ 4];
      FMemory[ 4] := FMemory[ 3];
      FMemory[ 3] := FMemory[ 2];
      FMemory[ 2] := FMemory[ 1];
      FMemory[ 1] := FMemory[ 0];
      FMemory[ 0] := FOutputs[ formant_out];
    end;

    function    TModFormant.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        formant_in         : Result := 'in';
        formant_vowel      : Result := 'vowel';
        formant_vowelin    : Result := 'vowelin';
        formant_controlmod : Result := 'controlmod';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModFormant.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        formant_out : Result := 'out';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModFlangeChorus = class( TMod)
  // This is a copy from some archaic older FX code I had (originally Turbo Pascal era)
  private
    FDryMix   : TSignal;
    FWetMix   : TSignal;
    FFeedBack : TSignal;
    FRate     : TSignal;
    FSweep    : TSignal;
    FDelay    : TSignal;
    FFrozen   : Boolean;
  private
    fp        : Integer;
    ep1       : Integer;
    ep2       : Integer;
    step      : Integer;
    depth     : Integer;
    idelay    : Integer;
    min_sweep : Integer;
    max_sweep : Integer;
    wlSweep   : TLongWord;
    Buf       : array[ 0 .. DELAY_BUF_SIZE] of TSignal;
  private
}

const

  flangechorus_in       = 0;
  flangechorus_mute     = 1;
  flangechorus_freeze   = 2;
  flangechorus_dry      = 3;
  flangechorus_wet      = 4;
  flangechorus_feedback = 5;
  flangechorus_rate     = 6;
  flangechorus_sweep    = 7;
  flangechorus_delay    = 8;
  flangechorus_freezein = 9;

  flangechorus_out      = 0;

    procedure   TModFlangeChorus.Setup;
      //  flange_chorus - original comments
      //
      //  Does flanging/chorusing family of effects based on a single
      //  varying delay.
      //
      //  dry_mix     mix of unaffected signal ( -0.999 to 0.999)
      //  wet_mix     mix of affected signal ( -0.999 - 0.999)
      //  feedback    amount of recirculation ( -0.9 - 0.9)
      //  rate        rate of delay change in millisecs per sec
      //  sweep       sweep range in millisecs
      //  delay       fixed additional delay in millisecs
    begin
      // fetch params
      step   := Round( FRate  * 65.536);                 // done: So Rate goes  0 - 1 now
      depth  := Round( FSweep * System_Rate / 1000);     // done: so Sweep goes 0 - 1 now
      idelay := Round( FDelay * System_Rate / 1000);     // done: so Delay goes 0 - 1 now

      // init/calc some stuff
      max_sweep := DELAY_BUF_SIZE - 2 - idelay;
      min_sweep := max_sweep - depth;

      if min_sweep < 0
      then min_sweep := 0;

      wlSweep.Hi := ( min_sweep + max_sweep) div 2;
      wlSweep.Lo := 0;

      // init store and read ptrs to known value
      fp  := 0;
      ep1 := 0;
      ep2 := 0;
    end;

//  public

    constructor TModFlangeChorus.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
      Setup;
    end;

    procedure   TModFlangeChorus.SetDefaults; // override;
    begin
      FDezipperMap := [
        flangechorus_mute    ,
        flangechorus_dry     ,
        flangechorus_wet     ,
        flangechorus_feedback
      ];
    end;

    procedure   TModFlangeChorus.DoSlowTick; // override;
    var
      OldRate  : TSignal;
      OldSweep : TSignal;
      OldDelay : TSignal;
    begin
      FFrozen   := SignalToLogic( FInputs[ flangechorus_freeze  ]) xor SignalToLogic( FInputs[ flangechorus_freezein]);
      FDryMix   :=                FInputs[ flangechorus_dry     ];
      FWetMix   :=                FInputs[ flangechorus_wet     ];
      FFeedBack := Clip         ( FInputs[ flangechorus_feedback], -0.999, 0.999);

      OldRate  := FRate;
      OldSweep := FSweep;
      OldDelay := FDelay;

      FRate  := FInputs[ flangechorus_rate ];
      FSweep := FInputs[ flangechorus_sweep];
      FDelay := FInputs[ flangechorus_delay];

      if
        ( FRate  <> OldRate ) or
        ( FSweep <> OldSweep) or
        ( FDelay <> OldDelay)
      then Setup;
    end;

    procedure   TModFlangeChorus.DoTick; // override;
    var
      outval : TSignal;
      aSample: TSignal;
    begin
      aSample := FInputs[ flangechorus_in];

      // interpolate from the 2 read values

      outval :=
        ( Buf[ ep1] * wlSweep.Lo +
          Buf[ ep2] * ( 65536.0 - wlSweep.Lo)
        ) / 65536.0;

      // develop and output final output mix

      outval := Normalize( outval * FWetMix + aSample * FDryMix);
      FOutputs[ flangechorus_out] := SignalToMute( FInputs[ flangechorus_mute]) * outval;

      // update ptrs

       // store finished input plus feedback - if not frozen
      if not FFrozen
      then Buf[ fp] := aSample + outval * FFeedback;

      fp  := ( fp + 1) and ( DELAY_BUF_SIZE - 1);
      Inc( Integer( wlSweep), step);
      ep1 := ( fp + wlSweep.Hi) and ( DELAY_BUF_SIZE - 1);
      ep2 := ( ep1 - 1) and ( DELAY_BUF_SIZE - 1);
      if wlSweep.Hi > max_sweep       // see if we hit top of sweep
      then step := - step             // reverse sweep
      else if wlSweep.Hi < min_sweep  // or if we hit bottom of sweep
      then step := - step;            // reverse sweep
    end;

    procedure   TModFlangeChorus.SampleRateChanged; // override;
    begin
      Setup;
    end;

    function    TModFlangeChorus.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        flangechorus_in       : Result := 'in';
        flangechorus_mute     : Result := 'mute';
        flangechorus_freeze   : Result := 'freeze';
        flangechorus_dry      : Result := 'dry';
        flangechorus_wet      : Result := 'wet';
        flangechorus_feedback : Result := 'feedback';
        flangechorus_rate     : Result := 'rate';
        flangechorus_sweep    : Result := 'sweep';
        flangechorus_delay    : Result := 'delay';
        flangechorus_freezein : Result := 'freezein';
        else                    Result := NO_INPUT;
      end;
    end;

    function    TModFlangeChorus.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        flangechorus_out : Result := 'out';
        else               Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModPitchChange = class( TMod)
  // This is a copy from some archaic older FX code I had (originally Turbo Pascal era)
  private
    FDryMix   : TSignal;
    FWetMix   : TSignal;
    FFeedback : TSignal;
    FRate     : TSignal;
    FDepth    : TSignal;
    FDelay    : TSignal;
    FFrozen   : Boolean;
  private
    fp         : Integer;
    ep1        : Integer;
    ep2        : Integer;
    ep3        : Integer;
    ep4        : Integer;
    idepth     : Integer;
    idelay     : Integer;
    min_sweep  : Integer;
    max_sweep  : Integer;
    step       : Word;
    xfade      : Integer;
    xfade_cnt  : Integer;
    active     : Integer;
    active_cnt : Integer;
    sweep_up   : Boolean;
    chanA      : Boolean;
    blendA     : TSignal;
    blendB     : TSignal;
    fadeA      : PFadeTable;
    fadeB      : PFadeTable;
    wlSweep    : TLongWord;
    Buf        : array[ 0 .. DELAY_BUF_SIZE] of TSignal;
    Fade_Out   : TFadeTable;
    Fade_In    : TFadeTable;
  private
}

const

  pitchchange_in       = 0;
  pitchchange_mute     = 1;
  pitchchange_freeze   = 2;
  pitchchange_dry      = 3;
  pitchchange_wet      = 4;
  pitchchange_feedback = 5;
  pitchchange_rate     = 6;
  pitchchange_depth    = 7;
  pitchchange_delay    = 8;
  pitchchange_freezein = 9;

  pitchchange_out      = 0;

    procedure   TModPitchChange.Setup;
      //  pitch_change - original comments
      //
      //  dry_mix     mix of unaffected signal ( -0.999 to 0.999)
      //  wet_mix     mix of affected signal ( -0.999 - 0.999)
      //  feedback    amount of recirculation ( -0.9 - 0.9)
      //  rate        amount of pitch change ( see table below for values)
      //  depth       sweep range in millisecs for generating pitch shift
      //  delay       fixed additional delay

      //  Semitones      Up              Down
      //      1       0.059463        -0.056126
      //      2       0.122462        -0.109101
      //      3       0.189207        -0.159104
      //      4       0.259921        -0.206299
      //      5       0.334840        -0.250846
      //      6       0.414214        -0.292893
      //      7       0.498307        -0.332580
      //      8       0.587401        -0.370039
      //      9       0.681793        -0.405396
      //      10      0.781797        -0.438769
      //      11      0.887749        -0.470268
      //      12      1.000000        -0.500000
    begin
      sweep_up  := FRate > 0;
      idepth    := Round( FDepth * System_Rate / 1000);
      idelay    := Round( FDelay * System_Rate / 1000);
      max_sweep := DELAY_BUF_SIZE - 2 - idelay;
      min_sweep := max_sweep - idepth;
      if min_sweep < 0
      then min_sweep := 0;
      active    := max_sweep - min_sweep - Round( xfade * FRate) - 2;

      // init store and read ptrs to known value, chanA active 1st

      fp                := 0;
      ep3               := 0;
      ep4               := 0;
      xfade_cnt         := 0;
      Integer( wlSweep) := 0;

      active_cnt := active;
      blendA     := 1.0;
      blendB     := 0.0;
      fadeA      := @ fade_out;
      fadeB      := @ fade_in;
      chanA      := True;
    end;

//   public

    constructor TModPitchChange.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
      Setup;
    end;

    procedure   TModPitchChange.SetDefaults; // override;
    var
      i : Integer;
    begin
      FDezipperMap := [
        pitchchange_mute    ,
        pitchchange_dry     ,
        pitchchange_wet     ,
        pitchchange_feedback
      ];

      xfade := Round( 12.0 * System_Rate / 1000);
      if xfade > MAX_XTAB
      // then raise( ECrossfade.Create( 'Can''t do pitch change crossfade at this sample rate'));
      then xfade := MAX_XTAB;

      // build the crossfade lookup tables

      for i := 0 to xfade - 1
      do begin
        fade_in [ i] := cos( i * HALF_PI / xfade);
        fade_out[ i] := sin( i * HALF_PI / xfade);
      end;
    end;

    procedure   TModPitchChange.DoSlowTick; // override;
    var
      OldRate  : TSignal;
      OldDepth : TSignal;
      OldDelay : TSignal;
      rStep    : TSignal;
    begin
      FFrozen   := SignalToLogic( FInputs[ pitchchange_freeze  ]) xor SignalToLogic( FInputs[ pitchchange_freezein]);
      FDryMix   :=                FInputs[ pitchchange_dry     ];
      FWetMix   :=                FInputs[ pitchchange_wet     ];
      FFeedBack := Clip         ( FInputs[ pitchchange_feedback], -0.999, 0.999);

      OldRate  := FRate;
      OldDepth := FDepth;
      OldDelay := FDelay;

      FRate  := FInputs[ pitchchange_rate ];
      FDepth := FInputs[ pitchchange_depth];
      FDelay := FInputs[ pitchchange_delay];

      if FRate <> OldRate
      then begin
        rStep := FRate * 65536.0;
        while rStep < 0
        do rStep := rStep + 65536.0;
        if rStep > 65535.0
        then rStep := 65535.0;
        step := Round( rStep);
        sweep_up := FRate > 0;

        if sweep_up
        then begin
          ep1 := min_sweep;
          ep2 := min_sweep;
        end
        else begin
          ep1 := max_sweep;
          ep2 := max_sweep;
        end;
      end;

      if
        ( FDepth <> OldDepth) or
        ( FDelay <> OldDelay)
      then Setup;
    end;

    procedure   TModPitchChange.DoTick; // override;
    var
      outval  : TSignal;
      comp    : TSignal;
      aSample : TSignal;
    begin
      aSample := FInputs[ pitchchange_in];

      // messy expression to interpolate from both pairs of read ptrs

      comp   := 65536.0 - wlSweep.Lo;
      outval :=
        (( Buf[ ep1] * wlSweep.Lo + Buf[ ep2] * comp) * blendA +
         ( Buf[ ep3] * wlSweep.Lo + Buf[ ep4] * comp) * blendB) / 65536.0;


      // store finished input plus feedback

      if not FFrozen
      then Buf[ fp] := aSample + outval * FFeedBack;

      // develop final output mix

      outval := Normalize( outval * FWetMix + aSample * FDryMix);
      FOutputs[ pitchchange_out] := SignalToMute( FInputs[ pitchchange_mute]) * outval;

      // see if crossfade active

      if xfade_cnt <> 0
      then begin
        Dec( xfade_cnt);
        blendA := fadeA^[ xfade_cnt];
        blendB := fadeB^[ xfade_cnt];
      end;

      // update store ptr
      fp := ( fp + 1) and ( DELAY_BUF_SIZE - 1);

      // see which direction
      if sweep_up
      then begin

        // update sweep
        Inc( Integer( wlSweep), step);

        // always inc at least once
        ep2 := ep1;
        ep1 := ( ep1 + 1) and ( DELAY_BUF_SIZE - 1);
        ep4 := ep3;
        ep3 := ( ep3 + 1) and ( DELAY_BUF_SIZE - 1);

        // if sweep didn't overflow, we're done
        if wlSweep.Hi = 0
        Then Exit;

        // sweep overflowed, inc again
        ep2 := ep1;
        ep1 := ( ep1 + 1) and ( DELAY_BUF_SIZE - 1);
        ep4 := ep3;
        ep3 := ( ep3 + 1) and ( DELAY_BUF_SIZE - 1);
        wlSweep.Hi := 0;

        // see if it's time to switch over to other delay channel
        if active_cnt = 0
        then begin
          xfade_cnt  := xfade;        // initiate crossfade
          active_cnt := active;       // start counter on new channel
          if chanA
          then begin                  // A has been active, go to B
            chanA := False;
            ep3   := ( fp + min_sweep) and ( DELAY_BUF_SIZE - 1);
            fadeA := @ fade_out;
            fadeB := @ fade_in;
          end
          else begin
            chanA := True;
            ep1   := ( fp + min_sweep) and ( DELAY_BUF_SIZE - 1);
            fadeA := @ fade_in;
            fadeB := @ fade_out;
          end;
        end
        else Dec( active_cnt);
      end
      else begin                              // do downward sweep
        Inc( Integer( wlSweep), step);        // update sweep

        // if sweep didn't overflow, inc ptrs, that's all
        if wlSweep.Hi = 0
        then begin
          ep2 := ep1;
          ep1 := ( ep1 + 1) and ( DELAY_BUF_SIZE - 1);
          ep4 := ep3;
          ep3 := ( ep3 + 1) and ( DELAY_BUF_SIZE - 1);
          Exit;
        End;


        // sweep overflowed, check on stuff but skip ptr inc
        wlSweep.Hi := 0;


        // see if it's time to switch over to other delay channel
        if active_cnt = 0
        then begin
          xfade_cnt  := xfade;
          active_cnt := active;
          if chanA
          then begin                  // A has been active, go to B
            chanA := False;
            ep3   := ( fp + max_sweep) and ( DELAY_BUF_SIZE - 1);
            fadeA := @ fade_out;
            fadeB := @ fade_in;
          end
          else begin
            chanA := True;
            ep1   := ( fp + max_sweep) and ( DELAY_BUF_SIZE - 1);
            fadeA := @ fade_in;
            fadeB := @ fade_out;
          end;
        end
        else Dec( active_cnt);
      end; // end down sweep
    end;

    procedure   TModPitchChange.SampleRateChanged; // override;
    begin
      Setup;
    end;

    function    TModPitchChange.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pitchchange_in       : Result := 'in';
        pitchchange_mute     : Result := 'mute';
        pitchchange_freeze   : Result := 'freeze';
        pitchchange_dry      : Result := 'dry';
        pitchchange_wet      : Result := 'wet';
        pitchchange_feedback : Result := 'feedback';
        pitchchange_rate     : Result := 'rate';
        pitchchange_depth    : Result := 'depth';
        pitchchange_delay    : Result := 'delay';
        pitchchange_freezein : Result := 'freezein';
        else                   Result := NO_INPUT;
      end;
    end;

    function    TModPitchChange.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        pitchchange_out : Result := 'out';
        else              Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEuclids = class( TMod)
  private
  public
}

const

  euclids_trig       = 0;
  euclids_pattern    = 1;
  euclids_patternmod = 2;
  euclids_pulses     = 3;
  euclids_steps      = 4;

  euclids_out        = 0;


    constructor TModEuclids.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModEuclids.SetDefaults; // override;
    begin
      FSteps    := 16;
      FPulses   := 7;
      FOldTrig  := False;
      FPosition := 0;
      FPattern  := EuclideanRhythm( FPulses, FSteps);
    end;

    procedure   TModEuclids.DoSlowTick; // override;
    var
      aSteps  : Integer;
      aPulses : Integer;
      aValue  : TSignal;
    begin
      aSteps  := Clip( SignalToInt( FInputs[ euclids_steps]), 1, 256);
      aValue  := aSteps * ( FInputs[ euclids_pattern] * FInputs[ euclids_patternmod] + FInputs[ euclids_pulses]);
      aPulses := Clip(( SignalToInt( aValue)), 0, aSteps);
      if ( aSteps <> FSteps) or ( aPulses <> FPulses)
      then begin
        FPattern := EuclideanRhythm( aPulses, aSteps);
        FSteps   := aSteps;
        FPulses  := aPulses;
      end;
      FPosition := FPosition mod FSteps;
    end;

    procedure   TModEuclids.DoTick; // override;
    var
      aTrig : Boolean;
      anOut : Boolean;
    begin
      aTrig := SignalToLogic( FInputs[ euclids_trig]);
      if aTrig <> FOldTrig
      then begin
        FOldTrig := aTrig;
        if aTrig
        then begin
          FPosition := ( FPosition + 1) mod FSteps;
          anOut     := FPosition in FPattern;
        end
        else anOut := False;
        FOutputs[ euclids_out] := LogicToSignal( anOut);
      end;
    end;

    procedure   TModEuclids.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutputs[ euclids_out]      ], AppLocale));
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, LogicToSignal( not FOldTrig)], AppLocale));
    end;

    function    TModEuclids.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        euclids_trig       : Result := 'trig';
        euclids_pattern    : Result := 'pattern';
        euclids_patternmod : Result := 'patternmod';
        euclids_pulses     : Result := 'pulses';
        euclids_steps      : Result := 'steps';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModEuclids.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        euclids_out : Result := 'out';
        else          Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModXYScope = class( TMod)
  private
    FHistory   : Integer;
    FDataQueue : TSignalFifo;
  public
    property    History: Integer read FHistory write SetHistory;
  private
    procedure   SetHistory( aValue: Integer);
  public
}

const

  xyscope_inx     = 0;
  xyscope_iny     = 1;
  xyscope_levx    = 2;
  xyscope_levy    = 3;
  xyscope_lockxy  = 5;

  xyscope_history = 0;


    procedure   TModXYScope.SetHistory( aValue: Integer);
    var
      OldData : TSignalFifo;
      NewData : TSignalFifo;
    begin
      if ( aValue <> FHistory) and ( aValue > 0)
      then begin
        OldData := FDataQueue;
        NewData := TSignalFifo.Create( aValue);
        NewData.CountOverflows := False;
        FHistory   := 0;
        FDataQueue := NewData;
        FHistory   := aValue;
        FreeAndNil( OldData);
      end;
    end;

    procedure   TModXYScope.AddPoint( anX, anY: TSignal);
    var
      aPair : TSignalPair;
    begin
      aPair.Left  := anX;
      aPair.Right := anY;
      if Assigned( FDataQueue) and ( FHistory > 0)
      then FDataQueue.Append( aPair);
    end;

//  public

    constructor TModXYScope.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModXYScope.SetDefaults; // override;
    begin
      inherited;
      History := 500;
    end;

    procedure   TModXYScope.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'history')
      then History := SignalToInt( aValue)
      else inherited;
    end;

    function    TModXYScope.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'history')
      then Result := History
      else Result := inherited;
    end;

    procedure   TModXYScope.DoTick; // override;
    begin
      AddPoint( FInputs[ xyscope_inx] * FInputs[ xyscope_levx], FInputs[ xyscope_iny] * FInputs[ xyscope_levy]);
    end;

    procedure   TModXYScope.GatherXYData( const aPrefix: string; const aCallback: TXYDataHandler); // override;
    begin
      if ( FHistory > 0) and Assigned( FDataQueue) { and ( FDataQueue.FillCount = FDataQueue.Size) }
      then aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FDataQueue);
    end;

    function    TModXYScope.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        xyscope_inx     : Result := 'inx';
        xyscope_iny     : Result := 'iny';
        xyscope_levx    : Result := 'levx';
        xyscope_levy    : Result := 'levy';
        xyscope_lockxy  : Result := 'lockxy';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModXYScope.OutputName( anIndex: Integer): string; // override;
    begin
      Result := NO_OUTPUT;
    end;

    function    TModXYScope.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        xyscope_history : Result := 'history';
        else              Result := NO_INTERNAL;
      end;
    end;


{ ========
  TFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FCount        : Integer;
    FFTWInBuffer  : Pfftw_real;
    FFTWOutBuffer : Pfftw_complex;
    FFTWPlan      : fftw_plan;
  end;
  TFFTDataArray = array of TFFTData;
}

//  public

    procedure   TFFTData.Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    var
      i : Integer;
    begin
      FWindow       := aWindow;
      FBinCount     := aBinCount;
      FIndex        := aStartIndex;
      FFTWInBuffer  := fftw_malloc( aBinCount * SizeOf( Tfftw_real   ));
      FFTWOutBuffer := fftw_malloc( aBinCount * SizeOf( Tfftw_complex));
      FFTWPlan      := fftw_plan_dft_r2c_1d( aBinCount, FFTWInBuffer, FFTWOutBuffer, FFTW_MEASURE);
      for i := 0 to aBinCount - 1
      do begin
        Pfftw_complex_array( FFTWOutBuffer)^[ i].re.v := 0;
        Pfftw_complex_array( FFTWOutBuffer)^[ i].im.v := 0;
        Pfftw_real_array   ( FFTWInBuffer )^[ i].v    := 0;
      end;
    end;

    procedure   TFFTData.Finalize;
    begin
      if Assigned( FFTWInBuffer)
      then fftw_free( FFTWInBuffer);
      if Assigned( FFTWOutBuffer)
      then fftw_free( FFTWOutBuffer);
      if Assigned( FFTWPlan)
      then fftw_destroy_plan( FFTWPlan);
    end;

    procedure   TFFTData.Tick( anInValue: TSignal; var anOutValue: TComplexSignal);
    begin
      if FIndex >= FBinCount
      then begin
        FIndex := 0;
        fftw_execute( FFTWPlan);
      end;
      if FIndex >= 0
      then begin
        Pfftw_real_array( FFTWInBuffer)[ FIndex].v := anInValue * FWindow[ FIndex];
        anOutValue.Real := anOutValue.Real + Pfftw_complex_array( FFTWOutBuffer)^[ FIndex].re.v;
        anOutValue.Imag := anOutValue.Imag + Pfftw_complex_array( FFTWOutBuffer)^[ FIndex].im.v;
      end;
      Inc( FIndex);
    end;


{ ========
  TFFTHandler = class
  private
    FBinCount : Integer;
    FWindow   : TFFTWindow;
    FData     : TFFTDataArray;
  public
    property    WindowCount: Integer read GetWindowCount;
  private
}

    function    TFFTHandler.GetWindowCount: Integer;
    begin
      Result := Length( FData);
    end;

//  public

    constructor TFFTHandler.Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      i : Integer;
    begin
      Assert( aBinCount     > 0);
      Assert( aWindowCount > 0);
      inherited Create;
      FDestroying := False;
      FBinCount    := aBinCount;
      FWindow     := aWindowType.Create( aBinCount);
      for i := 0 to aWindowCount - 1
      do begin
        SetLength( FData, WindowCount + 1);
        FData[ WindowCount - 1].Initialize( FBinCount, - (( i * FBinCount) div aWindowCount), FWindow);
      end;
    end;

    destructor  TFFTHandler.Destroy; // override;
    begin
      FDestroying := True;
      while WindowCount > 0
      do begin
        FData[ WindowCount - 1].Finalize;
        SetLength( FData, WindowCount - 1);
      end;
      FreeAndNil( FWindow);
      inherited;
    end;

    procedure   TFFTHandler.Tick( anInValue: TSignal; var anOutValue: TComplexSignal);
    var
      i : Integer;
    begin
      anOutValue.Real := 0;
      anOutValue.Imag := 0;
      if not FDestroying
      then begin
        for i := 0 to WindowCount - 1
        do FData[ i].Tick( anInValue, anOutValue);
        anOutValue.Real := anOutValue.Real / WindowCount;
        anOutValue.Imag := anOutValue.Imag / WindowCount;
      end;
    end;


{ ========
  TIFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FCount        : Integer;
    FFTWInBuffer  : Pfftw_complex;
    FFTWOutBuffer : Pfftw_real;
    FFTWPlan      : fftw_plan;
  end;
  TIFFTDataArray = array of TIFFTData;
}

//  public

    procedure   TIFFTData.Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    var
      i : Integer;
    begin
      FWindow       := aWindow;
      FBinCount      := aBinCount;
      FIndex        := aStartIndex;
      FFTWInBuffer  := fftw_malloc( aBinCount * SizeOf( Tfftw_complex));
      FFTWOutBuffer := fftw_malloc( aBinCount * SizeOf( Tfftw_real   ));
      FFTWPlan      := fftw_plan_dft_c2r_1d( aBinCount, FFTWInBuffer, FFTWOutBuffer, FFTW_MEASURE);
      for i := 0 to aBinCount - 1
      do begin
        Pfftw_complex_array( FFTWInBuffer )^[ i].re.v := 0;
        Pfftw_complex_array( FFTWInBuffer )^[ i].im.v := 0;
        Pfftw_real_array   ( FFTWOutBuffer)^[ i].v    := 0;
      end;
    end;

    procedure   TIFFTData.Finalize;
    begin
      if Assigned( FFTWInBuffer)
      then fftw_free( FFTWInBuffer);
      if Assigned( FFTWOutBuffer)
      then fftw_free( FFTWOutBuffer);
      if Assigned( FFTWPlan)
      then fftw_destroy_plan( FFTWPlan);
    end;

    procedure   TIFFTData.Tick( anInValue: TComplexSignal; var anOutValue: TSignal);
    begin
      if FIndex >= FBinCount
      then begin
        FIndex := 0;
        fftw_execute( FFTWPlan);
      end;
      if FIndex >= 0
      then begin
        Pfftw_complex_array( FFTWInBuffer)[ FIndex].re.v := anInValue.Real;
        Pfftw_complex_array( FFTWInBuffer)[ FIndex].im.v := anInValue.Imag;
        anOutValue := anOutValue  + FWindow[ FIndex] * Pfftw_real_array( FFTWOutBuffer)^[ FIndex].v;
      end;
      Inc( FIndex);
    end;


{ ========
  TIFFTHandler = class
  private
    FBinCount : Integer;
    FWindow   : TFFTWindow;
    FData     : TIFFTDataArray;
  public
    property    Count: Integer read GetCount;
  private
}

    function    TIFFTHandler.GetCount: Integer;
    begin
      Result := Length( FData);
    end;

//  public

    constructor TIFFTHandler.Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      i : Integer;
    begin
      Assert( aBinCount     > 0);
      Assert( aWindowCount > 0);
      inherited Create;
      FDestroying := False;
      FBinCount    := aBinCount;
      FWindow     := aWindowType.Create( aBinCount);
      for i := 0 to aWindowCount - 1
      do begin
        SetLength( FData, Count + 1);
        FData[ Count - 1].Initialize( FBinCount, ( i * FBinCount) div aWindowCount, FWindow);
      end;
    end;

    destructor  TIFFTHandler.Destroy; // override;
    begin
      FDestroying := True;
      while Count > 0
      do begin
        FData[ Count - 1].Finalize;
        SetLength( FData, Count - 1);
      end;
      FreeAndNil( FWindow);
      inherited;
    end;

    procedure   TIFFTHandler.Tick( anInValue: TComplexSignal; var anOutValue: TSignal);
    var
      i : Integer;
    begin
      anOutValue := 0;
      if not FDestroying
      then begin
        for i := 0 to Count - 1
        do FData[ i].Tick( anInValue, anOutValue);
        anOutValue  := anOutValue  / Count;
      end;
    end;


{ ========
  TModBaseFFT = class( TMod)
  private
    FBinCount    : Integer;
    FWindowCount : Integer;
    FWindowType  : TFFTWindowClass;
    FHandler     : TModBaseFFT;
  public
    property    BinCount : Integer read FBinCount   write SetBinCount;
  private
}

const

  fft_bincount = 0;


    procedure   TModBaseFFT.SetNewHandler( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      OldHandler : TFFTHandler;
      NewHandler : TFFTHandler;
    begin
      OldHandler := FHandler;
      if aBinCount > 0
      then NewHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
      else NewHandler := nil;
      FHandler := NewHandler;
      FreeAndNil( OldHandler);
    end;

    procedure   TModBaseFFT.SetBinCount( aValue: Integer);
    begin
      if aValue <> FBinCount
      then begin
        FBinCount := 0;
        SetNewHandler( FWindowCount, aValue, FWindowType);
        FBinCount := aValue;
      end;
    end;

  // public

    destructor  TModBaseFFT.Destroy; // override;
    begin
      FreeAndNil( FHandler);
      inherited;
    end;

    procedure   TModBaseFFT.SetDefaults; // override;
    begin
      inherited;
      FWindowType  := TFFTHammingWindow; // Defaults to Hann coefficients
      FWindowCount := 2;
      BinCount      := DEF_FFT_SIZE;
    end;

    procedure   TModBaseFFT.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'bincount')
      then BinCount := SignalToInt( aValue)
      else inherited;
    end;

    function    TModBaseFFT.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'bincount')
      then Result := BinCount
      else Result := inherited;
    end;

    function    TModBaseFFT.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        fft_bincount : Result := 'bincount';
        else           Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModBaseIFFT = class( TMod)
  private
    FBinCount    : Integer;
    FWindowCount : Integer;
    FWindowType  : TFFTWindowClass;
    FHandler     : TIFFTHandler;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  private
}

const

  ifft_bincount = 0;

    procedure   TModBaseIFFT.SetNewHandler( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      OldHandler : TIFFTHandler;
      NewHandler : TIFFTHandler;
    begin
      OldHandler := FHandler;
      if aBinCount > 0
      then NewHandler := TIFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
      else NewHandler := nil;
      FHandler := NewHandler;
      FreeAndNil( OldHandler);
    end;

    procedure   TModBaseIFFT.SetBinCount( aValue: Integer);
    begin
      if aValue <> FBinCount
      then begin
        FBinCount := 0;
        SetNewHandler( FWindowCount, aValue, FWindowType);
        FBinCount := aValue;
      end;
    end;

  // public

    destructor  TModBaseIFFT.Destroy; // override;
    begin
      FreeAndNil( FHandler);
      inherited;
    end;

    procedure   TModBaseIFFT.SetDefaults; // override;
    begin
      inherited;
      FWindowType  := TFFTHammingWindow; // Defaults to Hann coefficients
      FWindowCount := 2;
      BinCount      := DEF_FFT_SIZE;
    end;

    procedure   TModBaseIFFT.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'bincount')
      then BinCount := SignalToInt( aValue)
      else inherited;
    end;

    function    TModBaseIFFT.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'bincount')
      then Result := BinCount
      else Result := inherited;
    end;

    function    TModBaseIFFT.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        ifft_bincount : Result := 'bincount';
        else            Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModBaseFFTIFFT = class( TMod)
  private
    FBinCount    : Integer;
    FWindowCount : Integer;
    FWindowType  : TFFTWindowClass;
    FInHandler   : TFFTHandler;
    FOutHandler  : TIFFTHandler;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  private
}

const

  fftifft_bincount = 0;

    procedure   TModBaseFFTIFFT.SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      OldInHandler  : TFFTHandler;
      NewInHandler  : TFFTHandler;
      OldOutHandler : TIFFTHandler;
      NewOutHandler : TIFFTHandler;
    begin
      OldInHandler := FInHandler;
      if aBinCount > 0
      then NewInHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
      else NewInHandler := nil;
      FInHandler := NewInHandler;
      FreeAndNil( OldInHandler);
      OldOutHandler := FOutHandler;
      if aBinCount > 0
      then NewOutHandler := TIFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
      else NewOutHandler := nil;
      FOutHandler := NewOutHandler;
      FreeAndNil( OldOutHandler);
    end;

    procedure   TModBaseFFTIFFT.SetBinCount( aValue: Integer);
    begin
      if aValue <> FBinCount
      then begin
        FBinCount := 0;
        SetNewHandlers( FWindowCount, aValue, FWindowType);
        FBinCount := aValue;
      end;
    end;

  // public

    destructor  TModBaseFFTIFFT.Destroy; // override;
    begin
      FreeAndNil( FInHandler);
      FreeAndNil( FOutHandler);
      inherited;
    end;

    procedure   TModBaseFFTIFFT.SetDefaults; // override;
    begin
      inherited;
      FWindowType  := TFFTHammingWindow; // Defaults to Hann coefficients
      FWindowCount := 2;
      BinCount      := DEF_FFT_SIZE;
    end;

    procedure   TModBaseFFTIFFT.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'bincount')
      then BinCount := SignalToInt( aValue)
      else inherited;
    end;

    function    TModBaseFFTIFFT.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'bincount')
      then Result := BinCount
      else Result := inherited;
    end;

    function    TModBaseFFTIFFT.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        fftifft_bincount : Result := 'bincount';
        else               Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModFFT = class( TModBaseFFT)
  public
}

const

  fft_in      = 0;

  fft_re      = 0;
  fft_im      = 1;

    constructor TModFFT .Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModFFT.DoTick; // override;
    var
      anOutValue : TComplexSignal;
    begin
      if Assigned( FHandler)
      then begin
        FHandler.Tick( FInputs[ fft_in], anOutValue);
        FOutputs[ fft_re ] := anOutValue.Real;
        FOutputs[ fft_im ] := anOutValue.Imag;
      end;
    end;

    function    TModFFT.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        fft_in : Result := 'in';
        else     Result := NO_INPUT;
      end;
    end;

    function    TModFFT.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        fft_re : Result := 're';
        fft_im : Result := 'im';
        else     Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModIFFT = class( TModBaseIFFT)
  public
}

const

  ifft_re      = 0;
  ifft_im      = 1;

  ifft_out     = 0;

    constructor TModIFFT.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModIFFT.DoTick; // override;
    var
      anOutValue : TSignal;
    begin
      if Assigned( FHandler)
      then begin
        FHandler.Tick( MakeComplexSignal( FInputs[ ifft_re], FInputs[ ifft_im]), anOutValue);
        FOutputs[ ifft_out ] := anOutValue / ( 0.25 * BinCount); // 0.25 to compensate for window losses .. expeted about 0.64 tho ...
      end;
    end;

    function    TModIFFT.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        ifft_re : Result := 're';
        ifft_im : Result := 'im';
        else      Result := NO_INPUT;
      end;
    end;

    function    TModIFFT.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        ifft_out : Result := 'out';
        else       Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModVocoder = class( TModBaseFFTIFFT)
  private
    FBinCount         : Integer;
    FWindowCount      : Integer;
    FModulatorHandler : TFFTHandler;
    FCarrierHandler   : TFFTHandler;
    FOutHandler       : TIFFTHandler;
  private
}

const

  vocoder_modulator = 0;
  vocoder_carrier   = 1;
  vocoder_mute      = 2;

  vocoder_out       = 0;

    procedure   TModVocoder.SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      OldModulatorHandler  : TFFTHandler;
      NewModulatorHandler  : TFFTHandler;
      OldCarrierHandler    : TFFTHandler;
      NewCarrierHandler    : TFFTHandler;
      OldOutHandler        : TIFFTHandler;
      NewOutHandler        : TIFFTHandler;
    begin
      try
        OldModulatorHandler := FModulatorHandler;
        if aBinCount > 0
        then NewModulatorHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewModulatorHandler := nil;
        FModulatorHandler := NewModulatorHandler;
        FreeAndNil( OldModulatorHandler);

        OldCarrierHandler := FCarrierHandler;
        if aBinCount > 0
        then NewCarrierHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewCarrierHandler := nil;
        FCarrierHandler := NewCarrierHandler;
        FreeAndNil( OldCarrierHandler);

        OldOutHandler := FOutHandler;
        if aBinCount > 0
        then NewOutHandler := TIFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewOutHandler := nil;
        FOutHandler := NewOutHandler;
        FreeAndNil( OldOutHandler);
      except
        on E: Exception
        do FftwWarning( 'Vocoder');
      end;
    end;

    procedure   TModVocoder.SetBinCount( aValue: Integer);
    begin
      if aValue <> FBinCount
      then begin
        FBinCount := 0;
        SetNewHandlers( FWindowCount, aValue, TFFTHammingWindow);
        FBinCount := aValue;
      end;
    end;

//  public

    constructor TModVocoder.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModVocoder.SetDefaults; // override;
    const
      Channels = 2048 + 1;                     // Preferrably of form 2 ^ n + 1
    begin
      FDezipperMap := [
        vocoder_mute
      ];
      FWindowCount := 4;
      BinCount      := ( Channels - 1) * 2;
    end;

    procedure   TModVocoder.DoTick; // override;
    var
      aModulatorValue : TComplexSignal;
      aCarrierValue   : TComplexSignal;
      aChValue        : TComplexSignal;
      anAmplitude     : TSignal;
      anOutValue      : TSignal;
    begin
      if Assigned( FModulatorHandler) and Assigned( FCarrierHandler) and Assigned( FOutHandler)
      then begin
        FModulatorHandler.Tick( FInputs[ vocoder_modulator], aModulatorValue);
        FCarrierHandler  .Tick( FInputs[ vocoder_carrier  ], aCarrierValue  );

        anAmplitude   := Sqrt( Sqr( aModulatorValue.Real) + Sqr( aModulatorValue.Imag)) / BinCount;
        aChValue.Real := aCarrierValue.Real * anAmplitude;
        aChValue.Imag := aCarrierValue.Imag * anAmplitude;

        FOutHandler.Tick( aChValue, anOutValue);
        FOutputs[ vocoder_out] := SignalToMute( FInputs[ vocoder_mute]) * anOutValue;
      end;
    end;

    function    TModVocoder.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        vocoder_modulator : Result := 'modulator';
        vocoder_carrier   : Result := 'carrier';
        vocoder_mute      : Result := 'mute';
        else                Result := NO_INPUT;
      end;
    end;

    function    TModVocoder.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        vocoder_out : Result := 'out';
        else          Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModFilterBank = class( TMod)
  private
    FBinCount     : Integer;
    FWindowCount  : Integer;
    FInHandler    : TFFTHandler;
    FOutHandler   : TIFFTHandler;
    FCoefficients : TSignalArray;
    FCurrCoeff    : Integer;
    FFrequencies  : TSignalArray;
  public
    property    BinCount: Integer read FBinCount write SetBinCount;
  private
}

const

  filterbank_in     =  0;
  filterbank_fader1 =  1;
  filterbank_fader2 =  2;
  filterbank_fader3 =  3;
  filterbank_fader4 =  4;
  filterbank_fader5 =  5;
  filterbank_fader6 =  6;
  filterbank_fader7 =  7;
  filterbank_fader8 =  8;
  filterbank_fader9 =  9;
  filterbank_fader10 = 10;
  filterbank_fader11 = 11;
  filterbank_fader12 = 12;
  filterbank_fader13 = 13;
  filterbank_fader14 = 14;
  filterbank_fader15 = 15;
  filterbank_fader16 = 16;
  filterbank_fader17 = 17;

  filterbank_out     =  0;


    procedure   TModFilterBank.SetNewHandlers( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      OldInHandler  : TFFTHandler;
      NewInHandler  : TFFTHandler;
      OldOutHandler : TIFFTHandler;
      NewOutHandler : TIFFTHandler;
    begin
      try
        OldInHandler := FInHandler;
        if aBinCount > 0
        then NewInHandler := TFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewInHandler := nil;
        FInHandler := NewInHandler;
        FreeAndNil( OldInHandler);

        OldOutHandler := FOutHandler;
        if aBinCount > 0
        then NewOutHandler := TIFFTHandler.Create( aWindowCount, aBinCount, aWindowType)
        else NewOutHandler := nil;
        FOutHandler := NewOutHandler;
        FreeAndNil( OldOutHandler);

        SetLength( FCoefficients, aBinCount);
        FCurrCoeff := 0;
      except
        on E: Exception
        do FftwWarning( 'FilterBank');
      end;
    end;

    procedure   TModFilterBank.SetBinCount( aValue: Integer);
    begin
      if aValue <> FBinCount
      then begin
        FBinCount := 0;
        SetNewHandlers( FWindowCount, aValue, TFFTHammingWindow); // Hamming defaults to a Hann window
        FBinCount := aValue;
      end;
    end;

//  public

    constructor TModFilterBank.Create( aParent: TSynthPatch; const aName: string); // override;
    var
      Freq : TSignal;
      i    : Integer;
    begin
      inherited;
      SetLength( FFrequencies, 17);
      Freq := 40 / Sqrt( 2);               // about 28.3 Hz
      for i := 0 to 16
      do begin
        FFrequencies[ i] := Freq;
        Freq             := Freq * Sqrt( 2);
      end;
      FIsFast := True;
      FIsSlow := True;
    end;

    procedure   TModFilterBank.SetDefaults; // override;
    begin
      FDezipperMap := [
        filterbank_fader1,
        filterbank_fader2,
        filterbank_fader3,
        filterbank_fader4,
        filterbank_fader5,
        filterbank_fader6,
        filterbank_fader7,
        filterbank_fader8,
        filterbank_fader9,
        filterbank_fader10,
        filterbank_fader11,
        filterbank_fader12,
        filterbank_fader13,
        filterbank_fader14,
        filterbank_fader15,
        filterbank_fader16,
        filterbank_fader17
      ];
      FWindowCount := 8;
      BinCount     := 4096;
    end;

    procedure   TModFilterBank.DoSlowTick; // override;
    // Need to smear out the faders over an array of values
    // to be used to multiply the FFT'd input values with to get an output value
    // Use rectuamgular smear for now ... also .. the top half of the fft result
    // represents negative frequencies ... so ... the top half of the Coefficients
    // array shuld do that too .. the order is inverted tho ...
    var
      aHop  : TSignal;
      aFreq : TSignal;
      i     : Integer;
      p     : Integer;
    begin
      if BinCount > 0
      then begin
        aHop  := ( 0.5 * System_Rate) / BinCount;
        aFreq := FFrequencies[ 0];
        p     := 0;

        for i := 0 to 16
        do begin
          while ( aFreq <= FFrequencies[ i]) and ( p <= BinCount div 2)
          do begin
            FCoefficients[ p] := FInputs[ filterbank_fader1 + i];
            aFreq := aFreq + aHop;
            Inc( p);
          end;
          if p >= BinCount div 2
          then Break;
        end;

        // Fill out possible reaminder of first half
        while p <= BinCount div 2
        do begin
          FCoefficients[ p] := FInputs[ filterbank_fader17];
          Inc( p);
        end;

        // Set up 2nd half as a mirror image of the first half
        while p < BinCount
        do begin
          FCoefficients[ p] := FCoefficients[ BinCount - p - 1];
          Inc( p);
        end;
      end;
    end;

    procedure   TModFilterBank.DoTick; // override;
    var
      anInValue  : TComplexSignal;
      anOutValue : TSignal;
    begin
      if Assigned( FInHandler) and Assigned( FOutHandler)
      then begin
        FInHandler.Tick( FInputs[ filterbank_in], anInValue);

        anInValue.Real := anInValue.Real * FCoefficients[ FCurrCoeff];
        anInValue.Imag := anInValue.Imag * FCoefficients[ FCurrCoeff];
        FCurrCoeff     := ( FCurrCoeff + 1) mod BinCount;

        FOutHandler.Tick( anInValue, anOutValue);
        FOutputs[ filterbank_out] := SignalToMute( FInputs[ vocoder_mute]) * anOutValue;
      end;
    end;

    function    TModFilterBank.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        filterbank_in      : Result := 'in';
        filterbank_fader1  : Result := 'fader1';
        filterbank_fader2  : Result := 'fader2';
        filterbank_fader3  : Result := 'fader3';
        filterbank_fader4  : Result := 'fader4';
        filterbank_fader5  : Result := 'fader5';
        filterbank_fader6  : Result := 'fader6';
        filterbank_fader7  : Result := 'fader7';
        filterbank_fader8  : Result := 'fader8';
        filterbank_fader9  : Result := 'fader9';
        filterbank_fader10 : Result := 'fader10';
        filterbank_fader11 : Result := 'fader11';
        filterbank_fader12 : Result := 'fader12';
        filterbank_fader13 : Result := 'fader13';
        filterbank_fader14 : Result := 'fader14';
        filterbank_fader15 : Result := 'fader15';
        filterbank_fader16 : Result := 'fader16';
        filterbank_fader17 : Result := 'fader17';
        else                 Result := NO_INPUT;
      end;
    end;

    function    TModFilterBank.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        filterbank_out : Result := 'out';
        else             Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModTalkie = class( TMod)
  private
    FTalkie     : TTalkie;
    FDestroying : Boolean;
    FStart      : Boolean;
    FOldBankSel : TSignal;
    FBank       : TSignal;
    FPhrase     : TSignal;
    FSampleRate : TSignal;
    FFrameRate  : TSignal;
    FSpeaking   : Boolean;
    FDidRun     : Boolean;
  private
}

const

  talkie_start    =  0;
  talkie_phrase   =  1;
  talkie_mute     =  2;
  talkie_fr       =  3;
  talkie_frmod    =  4;
  talkie_frmodamt =  5;
  talkie_sr       =  6;
  talkie_srmod    =  7;
  talkie_srmodamt =  8;
  talkie_bank     =  9;
  talkie_banksel  = 10;
  talkie_fm       = 11;

  talkie_out      =  0;
  talkie_rdy      =  1;


    procedure   TModTalkie.TalkStarted( aSender: TObject);
    begin
      FSpeaking := True;
      FDidRun   := True;
    end;

    procedure   TModTalkie.TalkDone( aSender: TObject);
    begin
      FSpeaking := False;
    end;

    procedure   TModTalkie.Say;
    var
      aPhraseSets : TTalkiePhraseSets;
      aBankMax    : Integer;
      aPhraseMax  : Integer;
      TheBank     : Integer;
      ThePhrase   : Integer;
    begin
      if Assigned( FTalkie)
      then begin
        aPhraseSets := TalkiePhraseSets;
        if Assigned( aPhraseSets)
        then begin
          aBankMax   := aPhraseSets.Count - 1;
          TheBank    := Clip( SignalToInt( aBankMax * FBank), 0, aBankMax);
          aPhraseMax := aPhraseSets.PhraseSet[ TheBank].Count - 1;
          ThePhrase  := Clip( SignalToInt(  aPhraseMax * FPhrase), 0, aPhraseMax);
          FTalkie.Say( FFrameRate, FSampleRate, aPhraseSets.PhraseSet[ TheBank].Data[ ThePhrase]);
        end;
      end;
    end;

    procedure   TModTalkie.NewFramRate( aValue: TSignal);
    begin
      if ( aValue <> FFrameRate) and Assigned( FTalkie)
      then begin
        FFrameRate        := aValue;
        FTalkie.FrameRate := aValue;
      end;
    end;

    procedure   TModTalkie.NewSampleRate( aValue: TSignal);
    begin
      if ( aValue <> FSampleRate) and Assigned( FTalkie)
      then begin
        FSampleRate        := aValue;
        FTalkie.SampleRate := aValue;
      end;
    end;

//  public

    constructor TModTalkie.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FFrameRate  :=   40.0; // Frame  rate in frames/s, processing rate of speech frames
      FSampleRate := 8000.0; // Sample rate in samples/s for speech, Talkie will rate convert this to System_Rate
      FTalkie := TTalkie.Create;
      FTalkie.OnStarted := TalkStarted;
      FTalkie.OnDone    := TalkDone;
      FIsFast := True;
      FIsSlow := False;
    end;

    destructor  TModTalkie.Destroy; // override;
    var
      FOldTalkie : TTalkie;
    begin
      FDestroying := True;
      FOldTalkie  := FTalkie;
      FTalkie     := nil;
      FOldTalkie.Stop;
      FOldTalkie.OnStarted := nil;
      FOldTalkie.OnDone    := nil;
      FreeAndNil( FOldTalkie);
      inherited;
    end;

    procedure   TModTalkie.SetDefaults; // override;
    begin
      inherited;
      FStart  := False;
      FDidRun := False;
      FBank   := -1;
      FPhrase := -1;
      FInputs[ talkie_fm     ] := 0.0; // Set FM amount to neutral to make speed work ...
      FInputs[ talkie_banksel] := -1;  // Mark input as unused
    end;

    procedure   TModTalkie.DoTick; // override;
    var
      Start      : Boolean;
      BankSel    : TSignal;
      FrameRate  : TSignal;
      SampleRate : TSignal;
      FmFactor   : TSignal;
    begin
      if not FDestroying and Assigned( FTalkie)
      then begin
        Start      := SignalToLogic(    FInputs[ talkie_start  ]);
        BankSel    :=                   FInputs[ talkie_banksel];
        FBank      :=                   FInputs[ talkie_bank   ];
        FPhrase    :=                   FInputs[ talkie_phrase ];
        FmFactor   := LookupFrequency(  FInputs[ talkie_fm]) / LookupFrequency( 0);
        FrameRate  :=      80.0 * Clip( FInputs[ talkie_fr] + FInputs[ talkie_frmod] * FInputs[ talkie_frmodamt], 0, 1);
        SampleRate := Clip( 16000.0 * ( FInputs[ talkie_sr] + FInputs[ talkie_srmod] * FInputs[ talkie_srmodamt]) * FmFactor, 0, System_Rate div 2);

        if ( BankSel >= 0) and ( BankSel <> FOldBankSel) then FBank := Clip( BankSel, 0, 1);
        if FrameRate  <> FFrameRate                      then NewFramRate  ( FrameRate );
        if SampleRate <> FSampleRate                     then NewSampleRate( SampleRate);

        if ( not FStart and Start)
        then Say;

        FStart      := Start;
        FOldBankSel := BankSel;

        FOutputs[ talkie_out] := SignalToMute( FInputs[ talkie_mute]) * FTalkie.Tick;
        FOutputs[ talkie_rdy] := LogicToSignal( FDidRun and not FSpeaking);
      end
      else begin
        FSpeaking             := False;
        FStart                := False;
        FOutputs[ talkie_out] := SignalToMute( FInputs[ talkie_mute]);
        FOutputs[ talkie_rdy] := LogicToSignal( False);
      end;
      ResetFlag := False;
    end;

    procedure   TModTalkie.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/runs/%f', [ aPrefix, Name, LogicToSignal( FSpeaking)], AppLocale));
      aCallback( Self, Format( '%s/%s/rdy/%f' , [ aPrefix, Name, FOutputs[ talkie_rdy]]    , AppLocale));
    end;

    function    TModTalkie.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        talkie_start    : Result := 'start';
        talkie_phrase   : Result := 'phrase';
        talkie_mute     : Result := 'mute';
        talkie_fr       : Result := 'fr';
        talkie_frmod    : Result := 'frmod';
        talkie_frmodamt : Result := 'frmodamt';
        talkie_sr       : Result := 'sr';
        talkie_srmod    : Result := 'srmod';
        talkie_srmodamt : Result := 'srmodamt';
        talkie_bank     : Result := 'bank';
        talkie_banksel  : Result := 'banksel';
        talkie_fm       : Result := 'fm';
        else              Result := NO_INPUT;
      end;
    end;

    function    TModTalkie.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        talkie_out : Result := 'out';
        talkie_rdy : Result := 'rdy';
        else         Result := NO_OUTPUT;
      end;
    end;


end.
